#ifndef _ARTE_TABLE_HH_
#define _ARTE_TABLE_HH_
#define _ARTE_HH_
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   09. Mar 1997 by C. Borgmeier
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: This file is part of the C++ interface for Arte
//              the HERA-B reconstruction and analysis tool
//
//              This is the definition of the ArtePointer class.
//
//              ArtePointer allows access to the Arte data tables in
//              a pointer-like way.
//
// changes:
// 09 Mar 1997 creation by Christoph Borgmeier
// 08 Nov 2000 (TG) added constructor to ArteTableIterator
//
// *****************************************************************************

#include <string>
#include <iterator>
#include "arte/PointerBase.hh"

#ifndef NO_ARTE
extern "C" void mmbrel_( const int& link,
			 const ArtePointerBase& e1,
			 const ArtePointerBase& e2 );
extern "C" void mmdrop_( int &link );
extern "C" void mmbook_( int &link, const int& );
#endif

// template <class T> class ArtePointer;
template <class T> class ArteTable;
template <class T> class ArteTableIterator;

//  #if defined(__CINT__) || defined(CINTOBJECT)
//  class Pointer {};
//  #endif

// #warning ---------> Reading ArtePointer!!

//==============================================================================
// ArtePointer
//==============================================================================
template <class T>
class ArtePointer : public ArtePointerBase {
  
//  #ifdef USE_GATE
//    friend void operator<< <>( TBuffer& buf, const ArtePointer<T> ap );
//  #endif
  friend class ArteTable<T>;
  friend class ArteTableIterator<T>;
private:
//  #if !defined(__CINT__) && !defined(CINTOBJECT)
//    class Pointer {};
//  #endif
public:
  typedef ArtePointer<T>    self;
  typedef T                 value_type;
  typedef value_type*       pointer;
  typedef value_type&       reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif

  ArtePointer( const ArtePointer<T>& index ): ArtePointerBase( index ) {}  
  ArtePointer(): ArtePointerBase( 0 ) {}
  
  inline reference operator*() const;
  inline T* operator->() const;

//    inline void setInt( unsigned int index )
//      { *static_cast<ArtePointerBase*>(this) = ArtePointerBase( index ); }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1) || defined(CINTOBJECT)
  ArtePointer( unsigned int index ): ArtePointerBase( index ) {}

  inline ArtePointer<T>& operator++() { increment(); return *this; }
  inline ArtePointer<T>  operator++(int) {
    ArtePointer<T> tmp = *this;
    increment(); return tmp;
  }

#ifndef NO_ARTE
  /// Returns the number of entries in the corresponding Arte table.
  inline static unsigned int nrow();
  
  inline static self append(const_reference e);

  inline static void*     getLink();
  inline static int&      getIntLink();
  inline static ArteLink& getLinkLink();

  // Return the width of the Arte table.
  inline static int ncol();
#endif /* NO_ARTE */

public:
#else // ARTE_CXX_COMPAT
//  #ifndef __MAKECINT__
//    ArtePointer( Pointer* ): ArtePointerBase( 0 ) {}

//    inline bool operator!=( Pointer* ) const
//      { return ArtePointerBase::operator!=( 0 ); }
//  #endif
protected:
#endif /* COMPAT */
  unsigned int index() const { return ArtePointerBase::index(); }

public:
  // by T.Glebe
  inline bool operator< ( const ArtePointer<T>& rhs) const
    { return index() < rhs.index(); }

  inline int  operator- ( const ArtePointer<T>& rhs) const
    { return index() - rhs.index(); }

  inline bool operator!=( const ArtePointer<T>& rhs ) const
    { return ArtePointerBase::operator!=( rhs ); }

  inline bool operator==( const ArtePointer<T>& rhs ) const
    { return !operator!=( rhs ); }

private:
  inline void* dereference() const;
  static int* getBase();

#ifdef USE_ROOT
  ClassDefT(ArtePointer<T>,1)
#endif
}; // End of class ArtePointer

#ifdef USE_ROOT
ClassDefT2(ArtePointer,T)
#endif

#ifdef USE_ROOT
template<class T>
inline void operator<<( TBuffer& buf, const ArtePointer<T> ap )
{
  buf << ap.index();
}

//  template<class T>
//  inline void operator>>( TBuffer& buf, ArtePointer<T>& ap )
//  {
//    int idx;
//    buf >> idx;
//    ap.setInt( idx );
//  }


template<class T>
inline void operator>>( TBuffer& buf, ArteTableIterator<T>& ap )
{
  ArtePointer<T> p;
  int idx;
  buf >> idx;
  p.setInt( idx );
  ap = p;
}
#endif

//==============================================================================
// ArteTableIterator
//==============================================================================
template <class T>
class ArteTableIterator {
  friend class ArteTable<T>;
public:
  typedef ArteTableIterator<T> self;
  typedef ArtePointer<T> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif
  typedef std::random_access_iterator_tag iterator_category;
  typedef ptrdiff_t difference_type;

  ArteTableIterator() { ptr.setInt( 1 ); }

  ArteTableIterator( const ArtePointer<T>& rhs );

  ArteTableIterator( const ArteTableIterator<T>& rhs ) :
    ptr( rhs.ptr ) {}

  inline reference operator*() { return ptr; }

#ifdef __CINT__
  inline const reference operator*() const { return ptr; }
#else
  inline const_reference operator*() const { return ptr; }
#endif

  self& operator++() {
    ptr.increment();
    return *this;
  }

  self operator++(int) {
    self tmp = *this;
    ptr.increment();
    return tmp;
  }

  self& operator--() {
    ptr.decrement();
    return *this;
  }

  self operator--(int) {
    self tmp = *this;
    ptr.decrement();
    return tmp;
  }

  inline self& operator+=( int offset ) { ptr+=offset; return *this; }
  inline self& operator-=( int offset ) { ptr-=offset; return *this; }

  // added by T.Glebe
  inline int operator!=( const ArteTableIterator<T>& rhs ) const
    { return ptr!=rhs.ptr; }

  inline int operator==( const ArteTableIterator<T>& rhs ) const
    { return ptr==rhs.ptr; }

  inline int operator-( const ArteTableIterator<T>& rhs ) const
    { return ptr -  rhs.ptr; }

  inline ArteTableIterator<T> operator+( int rhs ) const {
    return ArteTableIterator<T>(*this) += rhs;
  }

  //    {  return operator+(*this,rhs); }

  inline ArteTableIterator<T> operator-( int rhs ) const
    {  return ArteTableIterator<T>(*this) -= rhs; }

  //    {  return operator+(*this,-rhs); }


#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  operator ArtePointer<T>() const { return ptr; }
  inline int valid() const;
  ArtePointer<T> next();
  inline static unsigned int nrow();
  inline T* operator->() const;
#else
protected:
#endif
  unsigned int index() const { return ptr.index(); }
  ArteTableIterator( unsigned int i )
    : ptr(*( ArteTable<T>::begin() + (i - 1) )) {}

private:
  ArtePointer<T> ptr;
}; // End of class ArteTableIterator


template <class T>
inline std::random_access_iterator_tag
iterator_category(const ArteTableIterator<T>&) {
  return std::random_access_iterator_tag();
}


template <class T>
inline ptrdiff_t*
distance_type(const ArteTableIterator<T>&) {
  return 0;
}


//  template <class T>
//  inline int operator-( const ArteTableIterator<T>& lhs,
//  		      const ArteTableIterator<T>& rhs )
//  {
//    return operator-( lhs.ptr, rhs.ptr );
//  }


//==============================================================================
// ArteTable - STL compatible container for Arte tables
//==============================================================================
template <class T>
class ArteTable: public TObject {
  //  friend class ArtePointer<T>;
public:
  typedef ArtePointer<T>    value_type;
  typedef value_type&       reference;
#ifndef __CINT__
  typedef const value_type& const_reference;
#endif
  typedef size_t            size_type;
  typedef ptrdiff_t         difference_type;

  typedef ArteTableIterator<T> iterator;
  
//   this thing does not go through egcs 1.0
//   typedef reverse_bidirectional_iterator<const_iterator, value_type,
//                                          const_reference, difference_type>
//       const_reverse_iterator;

#ifndef __CINT__
  // this can become too long for CINT
  typedef std::reverse_bidirectional_iterator<iterator, value_type, reference,
    difference_type> reverse_iterator;
#endif /* __CINT__ */
public:
  static iterator begin();
  static iterator end();

  static unsigned int size();
  static ArtePointer<T> push_back(const T& e);
  static ArtePointer<T> allocate( unsigned int n );
#ifndef NO_ARTE
  static void book( int num = 1 )
    { mmbook_( *static_cast<int*>(getLink()), num ); }
  static bool booked() { return *static_cast<int*>(getLink())!=0; }
  static void free() { mmdrop_( *static_cast<int*>(getLink()) ); }
  static void clear() { mmdrop_( *static_cast<int*>(getLink()) ); }
#endif
  static std::string getName();

  ArteTable() {};
  
public:
#ifdef USE_ROOT
  virtual void Print( Option_t* = "" );
#endif

  static bool isArteTable();
  
#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  inline static unsigned int nrow() { return size(); }
private:
#endif

  static void* getLink();

#ifndef NO_ARTE
private:
  static int ncol() { return gcbank_.iq[*static_cast<int*>(getLink())]; }
  // Return the width of the Arte table
#endif
#if defined(USE_GATE) && defined(USE_ROOT)
  ClassDefT(ArteTable<T>,1)
#endif
}; // end of class ArteTable


#if defined(USE_GATE) && defined(USE_ROOT)
ClassDefT2(ArteTable,T)
#endif


template<class T>
inline ArteTableIterator<T>::ArteTableIterator( const ArtePointer<T>& p )
  : ptr( p )
{
}
  
  
//==============================================================================
// ArteRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteRelationPointer: public ArtePointerBase {
  typedef ArteRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;

  friend class ArteRelVector<Relation>;
public:
  ArteRelationPointer(): ArtePointerBase( 0 ) {}
  ArteRelationPointer( const ArteRelationPointer<Relation,Result>& arg )
    : ArtePointerBase( arg ) {}
}; // end of ArteRelationPointer


template<class Result> class ArteRelationIterator;


//==============================================================================
// ArteInvOneRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteInvOneRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteInvOneRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef ptrdiff_t difference_type;

public:  
  iterator begin() const;
  iterator end() const;
  
  inline static ArteRelVector<Relation>& getLink();
#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  inline static ArteRelVector<Relation>& getLinkLink()
    { return getLink(); }
#endif
}; // end of ArteInvOneRelationPointer


//==============================================================================
// ArteManRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteManRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteManRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef ptrdiff_t difference_type;

public:
  iterator begin() const;
  iterator end() const;

  inline static ArteRelVector<Relation>& getLink();
#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  inline static ArteRelVector<Relation>& getLinkLink()
    { return getLink(); }
#endif
}; // end of ArteManRelationPointer


//==============================================================================
// ArteInvManRelationPointer
//==============================================================================
template<class Relation, class Result>
class ArteInvManRelationPointer: public ArteRelationPointer<Relation,Result> {
  typedef ArteInvManRelationPointer<Relation,Result> self;
  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef ArteRelationIterator<Result> iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef ptrdiff_t difference_type;

public:
  iterator begin() const;
  iterator end() const;

  inline static ArteRelVector<Relation>& getLink();
}; // end of ArteInvManRelationPointer


//==============================================================================
// ArteRelationIterator
//==============================================================================
template<class Result>
class ArteRelationIterator: public ArtePointerBase {
public:
  typedef ArteRelationIterator<Result> self;

  typedef ArtePointer<Result> value_type;
  typedef value_type* pointer;
  typedef value_type& reference;
  typedef self iterator;
  typedef std::forward_iterator_tag iterator_category;
  typedef int difference_type; // CINT does not know ptrdiff_t

  struct ArteOneToManyRelation {
    ArteOneToManyRelation(): next( ArtePointer<Result>(0) ) {}
    ArtePointer<Result> el;
    ArtePointerBase next;
  };
  
  struct ArteManyToManyRelation {
//      ArteManyToManyRelation()
//        : next(     ArtePointer<Result>( 0 ) ),
//    	othernext(ArtePointer<Result>( 0 ) ) {}
    ArteManyToManyRelation()
      : next      ( ArtePointer<Result>() ),
  	othernext ( ArtePointer<Result>() ) {}

    ArtePointer<Result> el;
    ArtePointerBase next;
    int selfel;  // we do not know the type
    ArtePointerBase othernext;
  }; // end of ArteManyToManyRelation

  struct ArteBackwardsManyToManyRelation {
//      ArteBackwardsManyToManyRelation()
//        : othernext( ArtePointer<Result>(0) ),
//  	next(      ArtePointer<Result>(0) ) {}
    ArteBackwardsManyToManyRelation()
      : othernext ( ArtePointer<Result>() ),
	next      ( ArtePointer<Result>() ) {}
    int selfel;  // we do not know the type
    ArtePointerBase othernext;
    ArtePointer<Result> el;
    ArtePointerBase next;
  }; // end of ArteBackwardsManyToManyRelation
  
public:
  ArteRelationIterator();

  template<class Relation>
  ArteRelationIterator( const ArteInvOneRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 2 )
    {}
  
  template<class Relation>
  ArteRelationIterator( const ArteManRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 3 )
    {}
  
  template<class Relation>
  ArteRelationIterator( const ArteInvManRelationPointer<Relation,Result>& x )
    : ArtePointerBase( x ),
      relation( &x.getLink() ),
      relType( 4 )
    {}
  
  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteInvOneRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 2;
      return *this;
    }

  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteManRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 3;
      return *this;
    }

  template<class Relation>
  const ArteRelationIterator<Result>&
  operator=( const ArteInvManRelationPointer<Relation,Result>& x )
    {
      *(ArtePointerBase*)this = x;
      relation = &x.getLink();
      reltype  = 4;
      return *this;
    }

  ArtePointer<Result> operator*() const { return lookup(); }
  // Dereferences the pointer.

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  operator ArtePointer<Result>() const { return lookup(); }
#endif

  self& operator++() {
    followLink();
    return *this;
  }
  self operator++(int) {
    self tmp = *this;
    followLink();
    return tmp;
  }

  bool operator!=( const ArteRelationIterator<Result>& rhs )
    { return ArtePointerBase::operator!=( rhs ); }

  bool operator==( const ArteRelationIterator<Result>& rhs )
    { return ArtePointerBase::operator==( rhs ); }

#if !defined(ARTE_CXX_COMPAT_MODE) || (ARTE_CXX_COMPAT_MODE==1)
  int valid() const { return ArtePointerBase::operator!=(0); }
  void next() { followLink(); }
#endif
  
private:
  void* getBase() const
    {
#ifndef NO_ARTE
      return gcbank_.iq + (*relation).getInt() + 2;
#else
      cerr << "wrong getBase called" << endl;
      return 0;
#endif
    }

  ArteManyToManyRelation getManyToMany() const {
    ArteManyToManyRelation* rp
      = static_cast<ArteManyToManyRelation*>( getBase() );
    return rp[index()-1];      
  }
  
  ArteBackwardsManyToManyRelation getBackwardsManyToMany() const {
    ArteBackwardsManyToManyRelation* rp
      = static_cast<ArteBackwardsManyToManyRelation*>( getBase() );
    return rp[index()-1];
  }
  
  ArteOneToManyRelation getOneToMany() const {
    ArteOneToManyRelation* rp
      = static_cast<ArteOneToManyRelation*>( getBase() );
    return rp[index()-1];
  }

protected:
  ArtePointer<Result> lookup() const {
    switch( relType ) {
    case 2: { // OneToMany
      ArteOneToManyRelation rel = getOneToMany();
      return rel.el;
    }
    case 3: { // ManyToMany
      ArteManyToManyRelation rel = getManyToMany();
      return rel.el;
    }
    case 4:
    default: { // BackwardsManyToMany
      ArteBackwardsManyToManyRelation rel = getBackwardsManyToMany();
      return rel.el;
    }
    } // switch
  } // lookup()


  void followLink() {
    if( index()!= 0 ) {
      switch( relType ) {
	
      case 2: { // OneToMany
	ArteOneToManyRelation rel = getOneToMany();
	//indexReference() = rel.next.index();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      case 3: { // ManyToMany
	ArteManyToManyRelation rel = getManyToMany();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      case 4:
      default: { // ManyToMany
	ArteBackwardsManyToManyRelation rel
	  = getBackwardsManyToMany();
	*static_cast<ArtePointerBase*>(this) = rel.next;
	break;
      }
      } // switch
    } // if index()
  } // followLink()    

private:
  ArteLink* relation;
  int relType;
}; // End of class ArteRelationIterator


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvOneRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteManRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvManRelationPointer<Relation,Result>::begin() const
{
  return ArteRelationIterator<Result>( *this );
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvManRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteManRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}


template<class Relation, class Result>
inline ArteRelationIterator<Result>
ArteInvOneRelationPointer<Relation,Result>::end() const
{
  return ArteRelationIterator<Result>();
}
    
#endif /* _ARTE_TABLE_HH_ */
