// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   08. Jun 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Photon candidate representation class
//
// changes:
// 08 Jun 2000 (TG) creation from C. Borgmeiers EcalHit class
// 17 Nov 2000 (TG) modified header files
// 20 Nov 2000 (TG) added namespace CLUE
// 11 Jan 2001 (TG) removed dependency from Point3D
// 02 Apr 2001 (TG) Tx, Ty, Wid added
// 05 Apr 2001 (TG) removed Tx, Ty, added x0vec
// 11 Jul 2001 (TG) added TrackIf inheritance, added RelStreamer, COV_, CINV_,
//                  added CTOR, Streamer, print_cov, FillCINV
// 12 Jul 2001 (TG) added cov_x_tx() cov_y_ty(), Zorig, calc_track, calc_CINV
// 16 Jul 2001 (TG) replaced Cov[] by Eerr (removed error on x,y), removed print_cov()
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
// 08 Jan 2002 (TG) added PrepareForVertexing();
// 09 Jan 2002 (TG) return value of calc_track() changed to bool, improved
//                  calc_track() and print()
// 16 Jan 2002 (TG) added calc_track(VertexIf)
// 21 Jan 2002 (TG) calc_track: bug fix: new computation of covariance matrix,
//                  provided by Carsten Crauss.
// 22 Jan 2002 (TG) calc_track(VertexIf&): bug fix: new computation of
//                  covariance matrix which uses full vertex covariance matrix,
//                  provided by Carsten Crauss, added Xerr, Yerr, adopted Streamer()
// 16 Jul 2002 (WG) Add energy of cluster's central tower
//
// ********************************************************************
#include <iostream>
#include <iomanip>
#include <cmath>

#include "clue/cluearte.hh"
#include "clue/EcalHit.hh"
#include "interfaces/VertexIf.hh"

#ifdef USE_ROOT
#include <TPolyMarker3D.h>
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif

  /// ECAL z position
  //  static const float zECAL = 1348.3;
  
  //==============================================================================
  // Constructors
  //==============================================================================
  EcalHit::EcalHit() : 
    rtrack(),
    X     (0.),
    Y     (0.),
    Z     (0.),
    Esum  (0.),
    Xerr  (0.),
    Yerr  (0.),
    Eerr  (0.),
    Wid   (0),
    En_C_T(0),
    Asym_ (0),
    Flag_ (0),
    Section(ECAL_none), 
    valid_(false),
    Zorig (0.),
    X_    (0.),
    Y_    (0.),
    Z_    (0.),
    Tx_   (0.),
    Ty_   (0.),
    COV_  (0),
    CINV_ (0)
    {}

  EcalHit::EcalHit( const EcalHit& rhs) : 
    rtrack(rhs.rtrack),
    X     (rhs.X),
    Y     (rhs.Y),
    Z     (rhs.Z),
    Esum  (rhs.Esum),
    Xerr  (rhs.Xerr),
    Yerr  (rhs.Yerr),
    Eerr  (rhs.Eerr),
    Wid   (rhs.Wid),
    En_C_T(rhs.En_C_T),
    Asym_ (rhs.Asym_),
    Flag_ (rhs.Flag_),
    Section(rhs.Section),
    valid_(rhs.valid_),
    Zorig (rhs.Zorig),
    X_    (rhs.X_),
    Y_    (rhs.Y_),
    Z_    (rhs.Z_),
    Tx_   (rhs.Tx_),
    Ty_   (rhs.Ty_),
    COV_  (rhs.COV_),
    CINV_ (rhs.CINV_)
    {}

  EcalHit::~EcalHit() {}


  //==============================================================================
  // TrackIf methods
  //==============================================================================
  float EcalHit::x()        const { return X_; }
  float EcalHit::y()        const { return Y_; }
  float EcalHit::z()        const { return Z_; }
  float EcalHit::x(float z) const { return X_ + Tx_ * (z - Z_); }
  float EcalHit::y(float z) const { return Y_ + Ty_ * (z - Z_); }
  float EcalHit::tx()       const { return Tx_; }
  float EcalHit::ty()       const { return Ty_; }
  float EcalHit::p()        const { return Esum; }
  float EcalHit::pz()       const { return Esum / sqrt(square(Tx_)+square(Ty_)+1); }
  float EcalHit::pt()       const { return Esum * sqrt((square(Tx_)+square(Ty_))/(square(Tx_)+square(Ty_)+1)); }
  float EcalHit::phi()      const { return 180. + 57.29577957 * atan2(Ty_,Tx_); }
  float EcalHit::theta()    const { return acos(evec()[2]); }
  float EcalHit::eta()      const { return -log(tan(theta()/2.)); }
  int   EcalHit::charge()   const { return 0; }
  float EcalHit::energy(double mass) const { return Esum; }
  float EcalHit::xf(double mass)     const { return (pz() - Esum*0.99898)/0.93827; }
  float EcalHit::rap(double mass)    const { return 0.5*log((Esum+pz())/(Esum-pz())); }

  SVector<double,3> EcalHit::xvec() const { return SVector<double,3>(X_,Y_,Z_); }

  SVector<double,3> EcalHit::tvec() const { return SVector<double,3>(Tx_,Ty_,1); }

  SVector<double,3> EcalHit::evec() const {
    return SVector<double,3>(1./mag(tvec()) * tvec());
  }

  SVector<double,3> EcalHit::pvec() const {
    return SVector<double,3>(p()/mag(tvec()) * tvec());
  }

  float EcalHit::cov_x(double dz=0.)  const { 
    return COV_(0,0) + 2.*dz*cov_x_tx() + dz*dz * cov_tx();
  }

  float EcalHit::cov_y(double dz=0.)  const {
    return COV_(1,1) + 2.*dz*cov_y_ty() + dz*dz * cov_ty();
  }

  float EcalHit::cov_x_tx(double dz=0.) const {
    return COV_(0,2) + dz * cov_tx();
  }

  float EcalHit::cov_y_ty(double dz=0.) const {
    return COV_(1,3) + dz * cov_ty();
  }

  float EcalHit::cov_tx()             const { return COV_(2,2); }
  float EcalHit::cov_ty()             const { return COV_(3,3); }
  float EcalHit::cov_p()              const { return COV_(4,4); }

  const SMatrix<double,5>& EcalHit::COV() const { return COV_; }

  const SMatrix<double,5>& EcalHit::CINV() const { return CINV_; }

  void EcalHit::valid()         { valid_ = true; }
  void EcalHit::invalid()       { valid_ = false; }
  bool EcalHit::isValid() const { return (valid_ == false) ?
				    const_cast<EcalHit*>(this)->calc_track(SVector<double,6>(0.,0.,0.,50.,50.,50.)) : true; }


  //==============================================================================
  // dummy TrackIf functions
  //==============================================================================
  float EcalHit::chi2()     const { return 0; }
  unsigned short int EcalHit::ndf() const { return 0; }

  //==============================================================================
  // operator==
  //==============================================================================
  bool EcalHit::operator==( const TrackIf& rhs ) const {
    return &rhs == this;
  }

  //==============================================================================
  // collect
  //==============================================================================
  void EcalHit::collect(vector<TrackIf*>& c) const {
    c.push_back(const_cast<EcalHit*>(this));
  }

  //==============================================================================
  // calc_track
  //==============================================================================
  bool EcalHit::calc_track(const SVector<double,6>& x) {

    // avoid division by zero
    if( (x[2] - Z) == 0. ) return false;

    X_  = x[0];
    Y_  = x[1];
    Z_  = x[2];
    const double dz = 1/(Z - Z_);
    const double dx = X - X_;
    const double dy = Y - Y_;
    Tx_ = dx * dz;
    Ty_ = dy * dz;

    const double dz2 = square(dz);
    const double dz4 = square(dz2);
    const double sx1 = Xerr;         // ECAL x-error squared
    const double sy1 = Yerr;         // ECAL y-error squared
    const double sx2 = square(x[3]); // origin x-error squared
    const double sy2 = square(x[4]); // origin y-error squared
    const double sz2 = square(x[5]); // origin z-error squared

    // see file EcalCov.ps for computation of the covariance matrix
    COV_(0,0)             = sx1;                        // cov(x,x)
    COV_(1,0) = COV_(0,1) = 0.;                         // cov(x,y)
    COV_(2,0) = COV_(0,2) = sx1*dz;                     // cov(x,tx)
    COV_(3,0) = COV_(0,3) = 0.;                         // cov(x,ty)
    COV_(1,1)             = sy1;                        // cov(y,y)
    COV_(1,2) = COV_(2,1) = 0.;                         // cov(y,tx)
    COV_(1,3) = COV_(3,1) = sy1*dz;                     // cov(y,ty)
    COV_(2,2) = sx2*dz2 + square(dx)*sz2*dz4 + sx1*dz2; // cov(tx,tx)
    COV_(2,3) = COV_(3,2) = dx*sz2*dy*dz4;              // cov(tx,ty)
    COV_(3,3) = sy2*dz2 + square(dy)*sz2*dz4 + sy1*dz2; // cov(ty,ty)
    COV_(4,4)             = Eerr;                       // cov(E,E)

    return (valid_ = true);
  }

  //==============================================================================
  // calc_track
  //==============================================================================
  bool EcalHit::calc_track(const VertexIf& rhs) {

    // avoid division by zero
    if( (rhs.vz() - Z) == 0. ) return false;

    X_  = rhs.vx();
    Y_  = rhs.vy();
    Z_  = rhs.vz();
    const double dz = 1/(Z - Z_);
    const double dx = X - X_;
    const double dy = Y - Y_;
    Tx_ = dx * dz;
    Ty_ = dy * dz;


    const SMatrix<double,3>& vcov = rhs.VCOV();

    const double dz2  = square(dz);
    const double dz3  = dz2*dz;
    const double dz4  = square(dz2);
    const double sx1  = Xerr;      // ECAL x-error squared
    const double sy1  = Yerr;      // ECAL y-error squared
    const double sx2  = vcov(0,0); // cov(x,x) of origin
    const double sy2  = vcov(1,1); // cov(y,y) of origin
    const double sz2  = vcov(2,2); // cov(z,z) of origin
    const double cxy2 = vcov(0,1); // cov(x,y) of origin
    const double cxz2 = vcov(0,2); // cov(x,z) of origin
    const double cyz2 = vcov(1,2); // cov(y,z) of origin

    // see file EcalCov2.ps for computation of the covariance matrix
    COV_(0,0)             = sx1;                        // cov(x,x)
    COV_(1,0) = COV_(0,1) = 0.;                         // cov(x,y)
    COV_(2,0) = COV_(0,2) = sx1*dz;                     // cov(x,tx)
    COV_(3,0) = COV_(0,3) = 0.;                         // cov(x,ty)
    COV_(1,1)             = sy1;                        // cov(y,y)
    COV_(1,2) = COV_(2,1) = 0.;                         // cov(y,tx)
    COV_(1,3) = COV_(3,1) = sy1*dz;                     // cov(y,ty)
    COV_(2,2) = sx2*dz2 - dx*cxz2*dz3 - cxz2*dx*dz3 + square(dx)*sz2*dz4 +
      sx1*dz; // cov(tx,tx)
    COV_(2,3) = COV_(3,2) = cxy2*dz2 - dx*cyz2*dz3 - cxz2*dy*dz3 +
      dx*dy*sz2*dz4; // cov(tx,ty)
    COV_(3,3) = sy2*dz2 - dy*cyz2*dz3 - dy*cyz2*dz3 + square(dy)*sz2*dz4 +
      sy1*dz2;  // cov(ty,ty)
    COV_(4,4)             = Eerr;                       // cov(E,E)

    return (valid_ = true);
  }

  //==============================================================================
  // calc_CINV
  //==============================================================================
  bool EcalHit::calc_CINV() {
    CINV_ = COV_;
    return CINV_.sinvert();
  }

  //==============================================================================
  // propagate
  //==============================================================================
  bool EcalHit::propagate(const double zz) {

    const double dz = zz - Z_;
    if(dz == 0.) return true;

    Zorig = Z_;
    X_ += dz * Tx_;
    Y_ += dz * Ty_;
    Z_  = zz;

    COV_(0,0)             += 2*dz*COV_(0,2) + dz*dz * COV_(2,2);
    COV_(0,1) = COV_(1,0) += dz * (COV_(1,2) + COV_(0,3) + dz*COV_(2,3));
    COV_(1,1)             += 2.*dz*COV_(1,3) + dz*dz * COV_(3,3);
    COV_(0,2) = COV_(2,0) += dz * COV_(2,2);
    COV_(1,2) = COV_(2,1) += dz * COV_(2,3);
    COV_(0,3) = COV_(3,0) += dz * COV_(2,3);
    COV_(1,3) = COV_(3,1) += dz * COV_(3,3);
    COV_(0,4) = COV_(4,0) += dz * COV_(4,2);
    COV_(1,4) = COV_(4,1) += dz * COV_(4,3);

    return true;
  }


  //============================================================================
  // PrepareForVertexing
  //============================================================================
  bool EcalHit::PrepareForVertexing() {

    propagate(0);
    return calc_CINV();
  }

  //==============================================================================
  // print
  //==============================================================================
  std::ostream& EcalHit::print( std::ostream& os ) const {
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    if(valid_ == true) {
      os << " x: "    << setw(10) << setprecision(2) << X_
	 << " y:"     << setw(10)                    << Y_
	 << " z:"     << setw(10)                    << Z_
	 << " tx:"    << setw(10) << setprecision(3) << Tx_
	 << " ty:"    << setw(10)                    << Ty_;
    } else {
      os << " x_c: " << setw(10) << setprecision(2) << X
	 << " y_c: " << setw(10) << Y
	 << " z_c: " << setw(10) << Z;
    }
    os << " esum: " << setw(8) << setprecision(2) << Esum
       << " eerr: " << setw(8) << setprecision(2) << sqrt(Eerr)
       << " width: " << setw(2) << static_cast<unsigned int>(Wid) 
       << " E_Centr: " << setw(8) << setprecision(2) << En_C_T ;

    if(valid_ == true) {
      os << endl << COV_ << endl;
    }
    return os;
  }


  //==============================================================================
  // operator =
  //==============================================================================
  EcalHit& EcalHit::operator= (const EcalHit& rhs) {
    if( this == &rhs ) return *this;

    rtrack = rhs.rtrack;
    X      = rhs.X;
    Y      = rhs.Y;
    Z      = rhs.Z;
    Esum   = rhs.Esum;
    Eerr   = rhs.Eerr;
    Wid    = rhs.Wid;
    valid_ = rhs.valid_;
    Zorig  = rhs.Zorig;
    COV_   = rhs.COV_;
    CINV_  = rhs.CINV_;
    X_     = rhs.X_;
    Y_     = rhs.Y_;
    Z_     = rhs.Z_;
    Tx_    = rhs.Tx_;
    Ty_    = rhs.Ty_;
    En_C_T = rhs.En_C_T;
    Asym_  = rhs.Asym_;
    Flag_  = rhs.Flag_;
    Section= rhs.Section;

    return *this;
  }

#ifdef USE_ROOT
  //==============================================================================
  // RelStreamer
  //==============================================================================
  /// only availabe in ROOT: call Streamer of relation container
  void EcalHit::RelStreamer(TBuffer& R__b, const Version_t& StreamVersion) {
    rTrack().Streamer(R__b);
  }

  //==============================================================================
  // Print
  //==============================================================================
  void EcalHit::Print(Option_t *t) {
    print( std::cout );
    std::cout << endl;
  }


  //==============================================================================
  // Streamer
  //==============================================================================
  void EcalHit::Streamer(TBuffer &R__b) {
   // Stream an object of class EcalHit.

    if (R__b.IsReading()) {
      Version_t R__v = R__b.ReadVersion(); 
      TObject::Streamer(R__b);
      R__b >> X;
      R__b >> Y;
      R__b >> Z;
      R__b >> Esum;
      if (R__v > 2) {
	R__b >> Xerr;
	R__b >> Yerr;
      }
      if (R__v > 1) {
	R__b >> Eerr;
      }
      R__b >> Wid;
      if (R__v > 3) {
	R__b >> En_C_T;
	R__b >> Asym_;
	R__b >> Flag_;
	UChar_t theSect;
	R__b >> theSect;
	switch (theSect) {
	case 1:  Section = ECAL_inner ; break;
	case 2:  Section = ECAL_middle ; break;
	case 5:  Section = ECAL_outer ; break;
	default: Section = ECAL_none;
	}
      }
    } else {
      R__b.WriteVersion(EcalHit::IsA());
      TObject::Streamer(R__b);
      R__b << X;
      R__b << Y;
      R__b << Z;
      R__b << Esum;
      R__b << Xerr;
      R__b << Yerr;
      R__b << Eerr;
      R__b << Wid;
      R__b << En_C_T;
      R__b << Asym_;
      R__b << Flag_;
      R__b << static_cast<UChar_t>(Section);
    }
  }

  //==============================================================================
  // Root::Paint
  //==============================================================================
  // How to use: a) create canvas
  //             b) create view via:  TView* view = new TView(1)
  //             c) loop over hits and call Draw()
  void EcalHit::Paint(Option_t *option) {

    // in a root kumac, gPad->Clear() deletes this pointer.
    TPolyMarker3D* m = new TPolyMarker3D();
    
    m->SetPoint(0, x(), y(), z());
    m->SetMarkerSize(1);
    m->SetMarkerColor(4);
    m->SetMarkerStyle(3);

    m->Draw();
  }

  ClassImp(EcalHit)
#endif // USE_ROOT

#ifdef NAMESPACE
} // end of namespace CLUE
#endif
