// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for Monte-Carlo tracks.
//
// changes:
// 27 Jan 1997 creation and 1st implementation by I. Siccama and C. Borgmeier
// 03 Nov 2000 (TG) changed ArtePointers to references
// 20 Nov 2000 (TG) added namespace CLUE
// 11 Jan 2001 (TG) removed dependency from Point3D
// 13 Aug 2001 (TG) added VertexIf members
// 14 Aug 2001 (TG) added Mother, added Streamer(), added copy CTOR, added operator=()
// 15 Aug 2001 (TG) added Paint()
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
//
// *****************************************************************************
#include <iostream>
#include <iomanip>

#include "clue/cluearte.hh"
#include "clue/MC_Vertex.hh"
#include "clue/MC_Track.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  static const SMatrix<double,3> MCVCOV;

  //==============================================================================
  // Constructors
  //==============================================================================
  MC_Vertex::MC_Vertex() :
    out   (),
    X     (0.),
    Y     (0.),
    Z     (0.),
    Mother(0) {}

  MC_Vertex::MC_Vertex( const MC_Vertex& rhs) :
    out   (rhs.out),
    X     (rhs.X),
    Y     (rhs.Y),
    Z     (rhs.Z),
    Mother(rhs.Mother) {}

  MC_Vertex::~MC_Vertex() {}


  //==============================================================================
  // operator=
  //==============================================================================
  MC_Vertex& MC_Vertex::operator=(const MC_Vertex& rhs) {
    if(this == &rhs) return *this;

    out    = rhs.out;
    X      = rhs.X;
    Y      = rhs.Y;
    Z      = rhs.Z;
    Mother = rhs.Mother;

    return *this;
  }

  //==============================================================================
  // VertexIf interface methods
  //==============================================================================
  float MC_Vertex::vx()   const { return X; }
  float MC_Vertex::vy()   const { return Y; }
  float MC_Vertex::vz()   const { return Z; }
  float MC_Vertex::chi2() const { return 0.; }
  unsigned short MC_Vertex::ndf() const { return 0; }
  float MC_Vertex::prob() const { return 1.; }
  unsigned short int MC_Vertex::ntracks() const { return out.size(); }

  const TrackIf* MC_Vertex::track(unsigned int i) const {
    return (i>static_cast<unsigned int>(out.size())) ? 0 : &(*out[i]); 
  }
  const TrackIf*& MC_Vertex::track(unsigned int i) {
    static const TrackIf* nullptr = 0;
    cerr << "Modification of a Relation is not allowed!" << endl;
    return nullptr;
  }

  SVector<double,3> MC_Vertex::vpos() const { return SVector<double,3>(X,Y,Z); }
  const SMatrix<double,3>& MC_Vertex::VCOV() const { return MCVCOV; }

  float MC_Vertex::vtx_cov_x()   const { return 0.; }
  float MC_Vertex::vtx_cov_y()   const { return 0.; }
  float MC_Vertex::vtx_cov_z()   const { return 0.; }
  

  //==============================================================================
  // print
  //==============================================================================
  std::ostream& MC_Vertex::print( std::ostream& os ) const {
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    os <<  "x:" << setw(10) << setprecision(2) << X
       << " y:" << setw(10) << setprecision(2) << Y
       << " z:" << setw(10) << setprecision(2) << Z;
    if(Mother > 0) {
      os << ", moth: " << setw(4) << mother();
    }
    if(out.size() > 0) {
      os << " outgoing tracks: ";
      copy(out.begin(), out.end(), ostream_iterator<ArtePointer<MC_Track> >(os, " "));
    }
    return os;
  }


#ifdef USE_ROOT
  //==============================================================================
  // RelStreamer
  //==============================================================================
  void MC_Vertex::RelStreamer(TBuffer& R__b, const Version_t& StreamVersion) {
    Out().Streamer(R__b);
  }

  //==============================================================================
  // Streamer
  //==============================================================================
  void MC_Vertex::Streamer(TBuffer &R__b) {
    // Stream an object of class MC_Vertex.
    
    if (R__b.IsReading()) {
      const Version_t R__v = R__b.ReadVersion(); 
      TObject::Streamer(R__b);
      R__b >> X;
      R__b >> Y;
      R__b >> Z;
      if (R__v > 1) { 
	R__b >> Mother;
      }
    } else {
      R__b.WriteVersion(MC_Vertex::IsA());
      TObject::Streamer(R__b);
      R__b << X;
      R__b << Y;
      R__b << Z;
      R__b << Mother;
    }
  }

  //==============================================================================
  // Print
  //==============================================================================
  void MC_Vertex::Print(Option_t *) {
    print( std::cout );
    std::cout << endl;
  }

#include "geometry/Point3D.hh"
  //==============================================================================
  // Root::Paint
  //==============================================================================
  // How to use: a) create canvas 
  //             b) create view via:  TView* view = new TView(1)
  //             c) loop over tracks and call Draw()
  // 
  void MC_Vertex::Paint(Option_t *option) {
    
    Point3D thisPoint(vx(), vy(), vz());
    thisPoint.Paint();
  }

  ClassImp(MC_Vertex)
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif
