// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Reconstructed Track class
//
// changes:
// 27 Jan 1997 created by I. Siccama, C. Borgmeier
// 31 Okt 2000 (TG) removed constructors using FORTRAN-Vt interface calls
// 31 Okt 2000 (TG) removed functions using CLHEP: merge(), getPar(), setPar(), getCov()
// 20 Nov 2000 (TG) added namespace CLUE
// 27 Dec 2000 (TG) end() optimization
// 12 Jan 2001 (TG) removed propagate() member function, print() changed
// 12 Jan 2001 (TG) removed GetLorentzVector(), mass() and massDiff() functions
// 15 Mar 2001 (TG) EoP() added
// 16 Mar 2001 (TG) print_cov() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 29 Jun 2001 (TG) moved charge(), pz() from .icc file to this place, removed
//                  xfabs(), moved energy(mass), xf(mass) from RecoTrack to here
// 05 Jul 2001 (TG) collect() added
// 09 Jul 2001 (TG) lock(), locked(), unlock() replaced by valid(), invalid(), isValid()
// 10 Jul 2001 (TG) operator==() added
// 11 Jul 2001 (TG) moved some member function bodies from .hh file to here,
//                  EoP(): changed esum() call to energy()
// 11 Jul 2001 (TG) removed ecal(), set_ecal(), Ecal, added Streamer()
// 16 Jul 2001 (TG) removed EoP()
// 31 Aug 2001 (TG) added PrepareForVertexing()
// 13 Sep 2001 (TG) removed call to sPinvToP() in PrepareForVertexing(), we
//                  will now write already in cluearte sigma(p) to avoid confusions
// 09 Jan 2002 (TG) changed return type of propagate() to bool
// 14 Jan 2002 (TG) added TrackIf constructor
//
// *****************************************************************************
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <cmath>

#include "clue/cluearte.hh"
#include "clue/RecoSegment.hh"
#include "clue/EcalHit.hh"
#include "smatrix/Prob.hh"
#include "smatrix/SVector.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors
  //==============================================================================
  RecoSegment::RecoSegment() :
    pmatch (),
    X     (0),
    Y     (0),
    Z     (0),
    Zorig (Z),
    Tx    (0),
    Ty    (0),
    P     (0),
    Chi2  (0),
    Charge(0),
    Cmp   (0),
    Rtra  (0),
    Nf    (0),
    valid_(true),
    COV_  (0),
    CINV_ (0) {
    for(unsigned int i=0; i<15; ++i) Cov[i] = 0.;
    //cout << "default constructing " << this << endl;
  }
  

  RecoSegment::RecoSegment( const RecoSegment& rhs ) :
    pmatch (rhs.pmatch ),
    X      (rhs.X      ),
    Y      (rhs.Y      ),
    Z      (rhs.Z      ),
    Zorig  (rhs.Zorig  ),
    Tx     (rhs.Tx     ),
    Ty     (rhs.Ty     ),
    P      (rhs.P      ),
    Chi2   (rhs.Chi2   ),
    Charge (rhs.Charge ),
    Cmp    (rhs.Cmp    ),
    Rtra   (rhs.Rtra   ),
    Nf     (rhs.Nf     ),
    valid_ (rhs.valid_),
    COV_   (rhs.COV_   ),
    CINV_  (rhs.CINV_) {
    for(unsigned int i=0; i<15; ++i) Cov[i] = rhs.Cov[i];
    
    //cout << "copy constructing " << this << " from " << &rhs << endl;
  }


  RecoSegment::RecoSegment(const TrackIf& rhs) :
    pmatch (),
    X     (rhs.x()),
    Y     (rhs.y()),
    Z     (rhs.z()),
    Zorig (Z),
    Tx    (rhs.tx()),
    Ty    (rhs.ty()),
    P     (rhs.p()),
    Chi2  (rhs.chi2()),
    Charge(rhs.charge()),
    Cmp   (0),
    Rtra  (0),
    Nf    (rhs.ndf()),
    valid_(true),
    COV_  (rhs.COV()),
    CINV_ (rhs.CINV()) 
    {
    }
  
  
  RecoSegment::~RecoSegment() {}


  //==============================================================================
  // operator==
  //==============================================================================
  bool RecoSegment::operator==( const TrackIf& rhs ) const {
    return &rhs == this;
  }

  //==============================================================================
  // TrackIf methods
  //==============================================================================
  float RecoSegment::x()        const { return X; }
  float RecoSegment::y()        const { return Y; }
  float RecoSegment::x(float z) const { return X + Tx * (z - Z); }
  float RecoSegment::y(float z) const { return Y + Ty * (z - Z); }
  float RecoSegment::z()        const { return Z; }
  float RecoSegment::zorig()    const { return Zorig; }
  float RecoSegment::tx()       const { return Tx; }
  float RecoSegment::ty()       const { return Ty; }
  float RecoSegment::p()        const { return P; }
  float RecoSegment::pt()       const { return P/sqrt(1.+1./(Tx*Tx+Ty*Ty)); }
  float RecoSegment::pz()       const { return P/sqrt(Tx*Tx+Ty*Ty+1); }
  float RecoSegment::chi2()     const { return Chi2; }
  unsigned short int RecoSegment::ndf() const { return Nf; }
  float RecoSegment::phi()      const { return 180. + 57.29577957 * atan2(Ty,Tx); }
  float RecoSegment::theta()    const { return acos(evec()[2]); }
  float RecoSegment::eta()      const { return -log(tan(theta()/2.)); }
  int   RecoSegment::charge()   const { return Charge; }
  float RecoSegment::energy(double mass) const { return sqrt( mass*mass + P*P ); }
  float RecoSegment::xf(double mass) const { return (pz() - energy(mass)*0.99898)/0.93827; }
  float RecoSegment::rap(double mass) const { return 0.5*log((energy(mass)+pz())/(energy(mass)-pz())); }

  SVector<double,3> RecoSegment::xvec() const { return SVector<double,3>(X,Y,Z); }

  SVector<double,3> RecoSegment::tvec() const { return SVector<double,3>(Tx,Ty,1); }

  SVector<double,3> RecoSegment::evec() const { 
    return SVector<double,3>(1./mag(tvec()) * tvec());
  }

  SVector<double,3> RecoSegment::pvec() const {
    return SVector<double,3>(P/mag(tvec()) * tvec());
  }

  const SMatrix<double,5>& RecoSegment::COV() const { return COV_; }

  const SMatrix<double,5>& RecoSegment::CINV() const { return CINV_; }

  //==============================================================================
  // Access covariance matrix
  //==============================================================================
  float RecoSegment::cov_x(double dz=0.) const { 
    return Cov[0] + 2*dz*cov_x_tx() + dz*dz * cov_tx();
  }

  float RecoSegment::cov_y(double dz=0.) const {
    return Cov[2] + 2.*dz*cov_y_ty() + dz*dz * cov_ty();
  }

  float RecoSegment::cov_tx() const { return Cov[5]; }
  float RecoSegment::cov_ty() const { return Cov[9]; }
  float RecoSegment::cov_p()  const { return Cov[14]; }


  //==============================================================================
  // operator =
  //==============================================================================
  const RecoSegment& RecoSegment::operator=( const RecoSegment& rhs ) {
    if( this == &rhs ) return *this;

    pmatch = rhs.pmatch;
    X      = rhs.X;
    Y      = rhs.Y;
    Z      = rhs.Z;
    Zorig  = rhs.Zorig;
    Tx     = rhs.Tx;
    Ty     = rhs.Ty;
    P      = rhs.P;
    Rtra   = rhs.Rtra;
    Chi2   = rhs.Chi2;
    Charge = rhs.Charge;
    Cmp    = rhs.Cmp;
    Nf     = rhs.Nf;
    valid_ = rhs.valid_;
    COV_   = rhs.COV_;
    CINV_  = rhs.CINV_;

    for(unsigned int i=0; i<15; ++i) Cov[i] = rhs.Cov[i];
    
    return *this;
  }


  //==============================================================================
  // Print
  //==============================================================================
  std::ostream& RecoSegment::print( std::ostream& os ) const {

    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    os << " tx: "    << setw(10) << setprecision(3) << tx() 
       << " ty: "    << setw(10) << setprecision(3) << ty()
       << " x: "     << setw(10) << setprecision(2) << x()
       << " y: "     << setw(10) << setprecision(2) << y()
       << " z: "     << setw(10) << setprecision(2) << z()
       << " p: "     << setw(10) << setprecision(2) << p()
       << " chi2: "  << setw(10) << setprecision(2) << chi2();

    os.setf(ios::fixed,ios::floatfield);
    os << " charge: " << setw(2) << charge();
    if(Cmp != 0)
      os << " Cmp: " << Cmp;

    return os;
  }


  //==============================================================================
  // print_cov
  //==============================================================================
  std::ostream& RecoSegment::print_cov( std::ostream& os ) const {
    os << " Cov. Matrix: " << endl;
    os.setf(ios::left,ios::adjustfield);
    os.setf(ios::scientific,ios::floatfield);
    os << setw(12) << setprecision(3) << cov_x() << endl
       << setw(12) << setprecision(3) << cov_x_y()
       << setw(12) << setprecision(3) << cov_y() << endl
       << setw(12) << setprecision(3) << cov_x_tx()
       << setw(12) << setprecision(3) << cov_y_tx()
       << setw(12) << setprecision(3) << cov_tx() << endl
       << setw(12) << setprecision(3) << cov_x_ty()
       << setw(12) << setprecision(3) << cov_y_ty()
       << setw(12) << setprecision(3) << cov_tx_ty()
       << setw(12) << setprecision(3) << cov_ty() << endl
       << setw(12) << setprecision(3) << cov_x_p()
       << setw(12) << setprecision(3) << cov_y_p()
       << setw(12) << setprecision(3) << cov_tx_p()
       << setw(12) << setprecision(3) << cov_ty_p()
       << setw(12) << setprecision(3) << cov_p();

    return  os;
  }

  //==============================================================================
  // prob
  //==============================================================================
  float RecoSegment::trprob() const {
    return Prob(Chi2, Nf);
  }
  

  //==============================================================================
  // valid
  //==============================================================================
  void RecoSegment::valid() { valid_ = true; }


  //==============================================================================
  // invalid
  //==============================================================================
  void RecoSegment::invalid() { valid_ = false; }
 

  //==============================================================================
  // isValid
  //==============================================================================
  bool RecoSegment::isValid() const { return valid_; }

  //==============================================================================
  // collect
  //==============================================================================
  void RecoSegment::collect(vector<TrackIf*>& c) const {
    c.push_back(const_cast<RecoSegment*>(this));
  }

  //==============================================================================
  // PrepareForVertexing
  //==============================================================================
  bool RecoSegment::PrepareForVertexing() {

    // needed transformations before using it in SVertex
    //    sPinvToP(); // we will now write already sigma(p)
    propagate(0);
    FillCINV();
    return true;
  }

  //==============================================================================
  // propagate
  //==============================================================================
  bool RecoSegment::propagate(const double zz) {

    const double dz = zz - Z;
    if(dz == 0.) return true;

    Zorig = Z;
    X += dz * Tx;
    Y += dz * Ty;
    Z  = zz;

    Cov[0]  += 2*dz*Cov[3] + dz*dz * Cov[5];
    Cov[1]  += dz * (Cov[4] + Cov[6] + dz*Cov[8]);
    Cov[2]  += 2.*dz*Cov[7] + dz*dz * Cov[9];
    Cov[3]  += dz * Cov[5];
    Cov[4]  += dz * Cov[8];
    Cov[6]  += dz * Cov[8];
    Cov[7]  += dz * Cov[9];
    Cov[10] += dz * Cov[12];
    Cov[11] += dz * Cov[13];

    return true;
  }

  //==============================================================================
  // sPinvToP, Formula: dsigma(Q/p)/dp = dsigma(p)/dp * -Q/p*p
  //==============================================================================
  void RecoSegment::sPinvToP() {
    const double p2 = square(P);
    Cov[10] *= -Charge * p2;
    Cov[11] *= -Charge * p2;
    Cov[12] *= -Charge * p2;
    Cov[13] *= -Charge * p2;
    Cov[14] *= square(p2);
    return;
  }

  //==============================================================================
  // sPToPinv
  //==============================================================================
  void RecoSegment::sPToPinv() {
    const double p2 = square(P);
    Cov[10] /= -Charge * p2;
    Cov[11] /= -Charge * p2;
    Cov[12] /= -Charge * p2;
    Cov[13] /= -Charge * p2;
    Cov[14] /= square(p2);
    return;
  }

#ifdef USE_ROOT
  //==============================================================================
  // Print
  //==============================================================================
  void RecoSegment::Print(Option_t * ) {
    print( std::cout );
    std::cout << endl;
    return;
  }

  //==============================================================================
  // Streamer
  //==============================================================================
  void RecoSegment::Streamer(TBuffer &R__b) {
    
    // Stream an object of class RecoSegment.
    
    if (R__b.IsReading()) {
      Version_t R__v = R__b.ReadVersion();
      TObject::Streamer(R__b);
      R__b >> X;
      R__b >> Y;
      R__b >> Z;
      R__b >> Tx;
      R__b >> Ty;
      R__b >> P;
      R__b >> Chi2;
      R__b.ReadStaticArray(Cov);
      R__b >> Charge;
      Cmp.Streamer(R__b);
      if(R__v < 4) {
	static unsigned short int Rtra;
	R__b >> Rtra;
      }
      if(R__v < 3) {
	static unsigned int Ecal;
	R__b >> Ecal;
      }
      R__b >> Nf;
    } else {
      R__b.WriteVersion(RecoSegment::IsA());
      TObject::Streamer(R__b);
      R__b << X;
      R__b << Y;
      R__b << Z;
      R__b << Tx;
      R__b << Ty;
      R__b << P;
      R__b << Chi2;
      R__b.WriteArray(Cov, 15);
      R__b << Charge;
      Cmp.Streamer(R__b);
      R__b << Nf;
    }
  } // end of Streamer

  ClassImp(RecoSegment)
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif
