// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   05. Nov 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Reconstructed Track class
//
// changes:
// 05 Nov 2000 (TG) created
// 05 Nov 2000 took out StreamObject code from RtEvent.C
// 20 Nov 2000 (TG) added namespace CLUE
// 05 Dec 2000 (TG) disabled doStream() as it causes Seg. Faults in ARTE
// 14 Dec 2000 (TG) added StreamRelation(), call to StreamRelation() in
//                  StreamObject::Streamer
// 27 Dec 2000 (TG) end() optimization
// 02 Jan 2001 (TG) removed include files
// 22 Mar 2001 (TG) added call to event->StreamUser()
// 11 Jul 2001 (TG) EcalHit Streamer added
// 14 Aug 2001 (TG) added call to StreamRelation<MC_Vertex>(), modification in
//                  Streamer()
// 05 Sep 2001 (TG) RelStreamer: added StreamVersion as argument to get
//                  backward compatibility in reading relations
//
// *****************************************************************************

#ifdef USE_ROOT
#include <TROOT.h>
#include <TFile.h>
#include <TTree.h>

#include "clue/cluearte.hh"
#include "clue/RecoTrack.hh"
#include "clue/RecoVertex.hh"
#include "clue/MC_Vertex.hh"       // needed by MC_Track
#include "clue/MC_Track.hh"
#include "clue/EcalHit.hh"
#include "clue/TrackClones.hh"
#include "clue/StreamObject.hh"

void updateTOPP();  // defined in PdtEntry.C

#ifdef NAMESPACE
namespace CLUE {
#endif

  RtEvent* StreamObject::event = 0;

  //==============================================================================
  // StreamObject Constructors
  //==============================================================================
  StreamObject::StreamObject() {}


  StreamObject::StreamObject( RtEvent* e ) {
#ifdef ClueDEBUG
    cout << " Construct StreamObject " << e << endl;
#endif
    if( RtEvent::recoTrackLink!=0 ) {
      //cerr << "warning: _not_ changing track link" << endl;
    } else {
      event = e;
      RtEvent::recoTrackLink     = event->tracks();
      RtEvent::vdsTrackLink      = event->vds();
      RtEvent::otrSpacePointLink = event->otr();
      RtEvent::richTrackLink     = event->rich();
      RtEvent::EcalHitLink       = event->ecal();
      RtEvent::muonTrackLink     = event->muon();
      RtEvent::recoVertexLink    = event->vertices();
      RtEvent::trackClonesLink   = event->clones();
      RtEvent::mcVertexLink      = event->mcVertices();
      RtEvent::mcTrackLink       = event->mcTracks();
      RtEvent::wireLink          = event->wires();
    }
  }


  //==============================================================================
  // StreamRelation: Stream relations of ArteTable Objects
  //==============================================================================
  template <class T>
    void StreamRelation(TBuffer &R__b, const Version_t& StreamVersion) {
    const ArteTableIterator<T> end = ArteTable<T>::end();
    for(ArteTableIterator<T> it = ArteTable<T>::begin();
	it != end; ++it) {
      (*it)->RelStreamer(R__b, StreamVersion);
    }
    return;
  }

  //==============================================================================
  // StreamObject::Streamer 
  //==============================================================================
  void StreamObject::Streamer(TBuffer &R__b) {

    // first load TOPP table, if not already done
    assert( event->toppLink!=0 );
    if( event->toppLink->size()==0 ) {
      TIterator* i = gROOT->GetListOfFiles()->MakeIterator();
      TFile* f = 0;
      //    while( (f = reinterpret_cast<TFile*>( i->Next() ) )!=0 ) {
      while( (f = static_cast<TFile*>( i->Next() ) )!=0 ) {
	TTree* constants = 0;
	if( f->Get("Clue")!=0
	    //	  && ( constants = reinterpret_cast<TTree*>(f->Get("constants")) )
	    && ( constants = static_cast<TTree*>(f->Get("constants")) )
	    !=0 ) {
	  constants->SetBranchAddress("toppTable", &event->toppLink );
	  constants->GetEntry( 0 );
	  break;
	}
      }
      delete i;
      updateTOPP();
    }

    // Stream an object of class StreamObject.
    if (R__b.IsReading()) {
      const Version_t R__v = R__b.ReadVersion();
      TObject::Streamer(R__b);
      if(event) event->unlock();

      // Stream Relations of ArteTable Objects
      StreamRelation<TrackClones>(R__b, R__v);
      StreamRelation<RecoTrack>(R__b, R__v);
      StreamRelation<RecoVertex>(R__b, R__v);
      StreamRelation<MC_Track>(R__b, R__v);
      StreamRelation<EcalHit>(R__b, R__v);
      if (R__v > 1) {
	StreamRelation<MC_Vertex>(R__b, R__v);
      }
    } else {
      const Version_t R__v = Class_Version();
      R__b.WriteVersion(StreamObject::IsA());
      TObject::Streamer(R__b);

      // Stream Relations of ArteTable Objects
      StreamRelation<TrackClones>(R__b, R__v);
      StreamRelation<RecoTrack>(R__b, R__v);
      StreamRelation<RecoVertex>(R__b, R__v);
      StreamRelation<MC_Track>(R__b, R__v);
      StreamRelation<EcalHit>(R__b, R__v);
      StreamRelation<MC_Vertex>(R__b, R__v);
    }

    // Stream the user field
    event->StreamUser(R__b);
  }
  
  ClassImp(StreamObject)

#endif // USE_ROOT
#ifdef NAMESPACE
} // namespace CLUE
#endif
