#ifndef ECAL_HIT_HH
#define ECAL_HIT_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   08. Jun 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Photon candidate representation class
//
// changes:
// 08 Jun 2000 (TG) creation from C. Borgmeiers EcalHit class
// 17 Nov 2000 (TG) modified header files
// 20 Nov 2000 (TG) added namespace CLUE
// 11 Jan 2001 (TG) removed dependency from Point3D
// 02 Apr 2001 (TG) Tx, Ty, Wid added
// 04 Apr 2001 (TG) xvec(), tvec(), evec(), pvec() added
// 05 Apr 2001 (TG) removed Tx, Ty
// 23 Mai 2001 (TG) moved function definitions to EcalHit.icc
// 11 Jul 2001 (TG) Et(), Ez() added, renamed pvec to Evec, added TrackIf
//                  inheritance, added valid_ member
// 11 Jul 2001 (TG) added RecoTrack relation, COV_, CINV_, Cov members, added
//                  CTOR, print_cov, FillCINV()
// 12 Jul 2001 (TG) added cov_x_tx() cov_y_ty(), Zorig, calc_track, calc_CINV
// 16 Jul 2001 (TG) replaced Cov[] by Eerr (removed error on x,y), removed print_cov()
// 27 Jul 2001 (TG) xc_vec added
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
// 08 Jan 2002 (TG) added PrepareForVertexing();
// 09 Jan 2002 (TG) return value of calc_track() changed to bool
// 09 Jan 2002 (TG) changed return type of propagate() to bool
// 16 Jan 2002 (TG) added calc_track(VertexIf)
// 22 Jan 2002 (TG) added Xerr and Yerr, incremented class version, removed x0vec
// 16 Jul 2002 (WG) added En_C_T [energy of central tower], and geometryflag,
//                  incremented class version
//
// ********************************************************************
#include <iosfwd>
#include <cmath>
#include "arte/TObject.hh"
#include "interfaces/TrackIf.hh"
#include "grelation/ClueRelation.hh"

#ifndef __CINT__
#include "smatrix/SVector.hh"
#endif

class VertexIf;

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** ECAL Hit - Photon candidates

      The errors of the cluster position at the ECAL are considered to be small
      and are not stored in this class. It is assumed for the moment that the
      major error comes from the unknown photon origin.

      @memo Photon candidates found by ECAL

      @author T. Glebe
  */
  //==============================================================================
  // EcalHit
  //==============================================================================
  class EcalHit : public TObject, public TrackIf {
  public:
    /** @name --- Constructors --- */
    ///
    EcalHit();
#ifndef NO_ARTE
    /// only available with ARTE
    EcalHit( const ArtePointer<RCCL>& rccl );
#endif
    ///
    EcalHit( const EcalHit& rhs );
    ///
    ~EcalHit();
    ///
    EcalHit& operator=(const EcalHit& rhs);

    /** @name --- Constants and Enums --- */
    ///
    typedef enum { ECAL_none = 0,     // invalid section
		   ECAL_inner, 
		   ECAL_middle, 
		   ECAL_outer } ECAL_geom;


    /** @name --- TrackIf methods ---\\
     A photon track is defined by the ECAL cluster coordinates $x_c$, $y_c$,
     $z_c$ and an origin $x_0$, $y_0$, $z_0$. Use the \texttt{calc\_track()}
     function to define an origin.*/
    /// $x$ position of track
    float x()    const;
    /// $y$ position of track
    float y()    const;
    /// $z$ position of track
    float z()    const;
    /// Track $x$ position at $z$
    float x(float z)   const;
    /// Track $y$ position at $z$
    float y(float z)   const;
    /// $t_x$ of cluster calculated from point $x_0$
    float tx()   const;
    /// $t_y$ of cluster calculated from point $x_0$
    float ty()   const;
    /// Photon track momentum $|\vec{p}| = E_{cluster}$
    float p()    const;
    /// transverse Photon track momentum:\\ $p_t = p\cdot\sqrt{(t_x^2+t_y^2)/(t_x^2+t_y^2+1)}$
    float pt()   const;
    /// Photon track momentum along $z$: $p_z = p/\sqrt{t_x^2+t_y^2+1}$
    float pz()  const;

    /// azimuthal angle $\phi$ [deg]
    float phi() const;
    /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
    float theta() const;
    /// pseudorapidity:\\ $\eta = -\log\tan(\theta/2.)$
    float eta() const;
    /// Photon charge: 0
    int charge() const;
    
    /// Ecal cluster energy $E_{cluster}$ (\textrm{rm} argument has no effect)
    float energy(double rm = 0.) const;
    /// Feynman Variable:\\ $x_F = \frac{E+p_z}{(E+p_z)_{max}}$
    float xf(double rm = 0.) const;
    /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
    float rap(double rm = 0.) const;
    
#ifndef __CINT__
    /// $\vec{x} = (x,y,z)$
    SVector<double,3> xvec() const;
    /// $\vec{t} = (t_x,t_y,1)$ slope vector
    SVector<double,3> tvec() const;
    /// $\vec{e} = (e_x,e_y,e_z)$ unit vector along $\vec{p}$
    SVector<double,3> evec() const;
    /// $\vec{p} = (p_x,p_y,p_z)$ momentum vector
    SVector<double,3> pvec() const;
#endif

    /// is ECAL cluster valid?
    bool isValid() const;
    
#ifndef __CINT__
    /// dummy function: return 0 matrix
    const SMatrix<double,5>& COV() const;
    /// dummy function: return 0 matrix
    const SMatrix<double,5>& CINV() const;
#endif
    /// get $\sigma_x^2|_{z+dz}$
    float cov_x(double dz=0.)        const;
    /// get $\sigma_y^2|_{z+dz}$
    float cov_y(double dz=0.)        const;
    /// get $\sigma_{t_x}^2 = \sigma_x^2 / (z-z_0)^2$
    float cov_tx()                   const;
    /// get $\sigma_{t_y}^2 = \sigma_y^2 / (z-z_0)^2$
    float cov_ty()                   const;
    /// get $\sigma_{E}^2$
    float cov_p()                    const;

    /// propagate track + errors
    bool propagate(const double dz);
    
    /// mark cluster as valid
    void valid();
    /// mark cluster as invalid
    void invalid();
    /// collect TrackIf pointers
    void collect(vector<TrackIf*>& c) const;
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;

    /** @name --- dummy functions --- */
    /// dummy function: always return 0
    float chi2() const;
    /// dummy function: always return 0
    unsigned short int ndf()  const;

    /** @name --- EcalHit access functions --- */
    /// return $z$ before propagation
    float zorig() const;
    /// get $\sigma_x \times \sigma_{t_x}|_{z+dz}$
    float cov_x_tx(double dz=0.) const;
    /// get $\sigma_y \times \sigma_{t_y}|_{z+dz}$
    float cov_y_ty(double dz=0.) const;

    /** @name --- Operators --- */
    /// comparison between Track objects
    bool operator==(const TrackIf& rhs) const;
    
    /** @name --- EcalHit Access methods --- */
    /// width of cluster
    unsigned int width() const;
    /// $x_c$: cluster $x$
    float x_c()  const;
    /// $y_c$: cluster $y$
    float y_c()  const;
    /// $z_c$: cluster $z$
    float z_c()  const;
    /// $t_x \cdot p_z$ 
    float px()   const;
    /// $t_y \cdot p_z$ 
    float py()   const;
    /// $E_z = E / \sqrt{t_x^2+t_y^2+1}$
    float Ez()   const;
    /// $E_t = E \cdot\sqrt{(t_x^2+t_y^2)/(t_x^2+t_y^2+1)}$ 
    float Et()   const;

    /// $E_{Centr}$:  Energy of cluster's central tower
    float E_Centr() const;
    /// asym: Asymmetry of cluster
    float asym() const;

    /// section:  which part of ECAL the cluster was found in
    ECAL_geom section() const;

#ifndef __CINT__
    /// $\vec{x} = (x_c,y_c,z_c)$
    SVector<double,3> xc_vec() const;
    /// $\vec{E} = (E_x,E_y,E_z)$
    SVector<double,3> Evec() const;
    /**	calculate track.

	Calculate track parameters and track covariance matrix using the photon
	origin and the errors of the origin given by the vector $x_0$.
	\begin{displaymath}
	\vec{x}_0 = \left(\begin{array}{l}
	x_0\\
	y_0\\
	z_0\\
	\sigma_{x_0}\\
	\sigma_{y_0}\\
	\sigma_{z_0}
	\end{array}\right)
	\end{displaymath}
    */
    bool calc_track(const SVector<double,6>& x0);
    /** calculate track from Ecal cluster position and Vertex position, which
	is given as an argument. Use the Vertex covariance matrix to calculate
	the covariance matrix of the photon candidate. */
    bool calc_track(const VertexIf& rhs);
    /// calculate inverse cov. matrix
    bool calc_CINV();
#endif

    /** Prepare reconstructed photon for vertexing with SVertex package:
        propagate track parameters and errors to $z=0$, fill inverse covariance matrix
    */
    bool PrepareForVertexing();

    /** @name --- CLUE relations --- */
    /// EcalHit $\leftrightarrow$ RecoTrack
    ClueRelationList<RecoTrack> rtrack;  //! not persistent
    ///
    const ClueRelationList<RecoTrack>& rTrack() const;
    ///
    ClueRelationList<RecoTrack>& rTrack();
    ///
    ClueRelationList<RecoTrack>& container(const ArtePointer<RecoTrack>& p);
    
#ifdef USE_ROOT
    /// only availabe in ROOT: call Streamer of relation container
    void RelStreamer(TBuffer& R__b, const Version_t& StreamVersion);
#endif

  private:
    Float_t X;       // cluster X
    Float_t Y;       // cluster Y
    Float_t Z;       // cluster Z
    Float_t Esum;    // energy content
    Float_t Xerr;    // cov(X,X): squared error on cluster x coordinate
    Float_t Yerr;    // cov(Y,Y): squared error on cluster y coordinate
    Float_t Eerr;    // cov(E,E): Energy error measured by ECAL squared
    UChar_t Wid;     // cluster width
    Float_t En_C_T;  //  Energy of cluster's central tower 
		Float_t Asym_;   //  Asymmetry of cluster
		Int_t   Flag_;   //  cluster flag
		ECAL_geom Section; // section of ECAL:  one of ECAL_inner, ECAL_middle, ECAL_outer
								
		Bool_t  valid_;  //! validity of photon track
    Float_t Zorig;   //! z-value before propagation

    Float_t X_;      //! Photon track origin
    Float_t Y_;      //! Photon track origin
    Float_t Z_;      //! Photon track origin
    Float_t Tx_;     //! Photon track tx
    Float_t Ty_;     //! Photon track ty

#ifndef __CINT__
    SMatrix<double,5> COV_;         //! cov. matrix (not persistent)
    SMatrix<double,5> CINV_;        //! inverse cov. matrix (not persistent)
#endif

#ifdef USE_ROOT
  public:   
    /** @name --- ROOT only functions --- */
    ///
    void Print(Option_t * = "" );
    ///
    void Paint(Option_t *option);

    ClassDef(EcalHit,4)
#endif
  }; // end of class EcalHit


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<< ( std::ostream& os, const EcalHit& t );

#ifdef NAMESPACE
} // end of namespace CLUE
#endif

#ifndef __CINT__
#include "EcalHit.icc"
#endif
#endif
