#ifndef MC_TRACK_HH
#define MC_TRACK_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   28. Jan 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for Monte-Carlo tracks.
//
// changes:
// 28 Jan 1997 creation and 1st implementation by C. Borgmeier
// 03 Nov 2000 (TG) changed ArtePointers to references, #include <vector> added
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dec 2000 (TG) replaced ARTE relations by CLUE relations
// 07 Dec 2000 (TG) added copy constructor, operator= added
// 14 Dec 2000 (TG) RelStreamer method added
// 22 Dec 2000 (TG) changed type of iTopp, Mother, B_Vertex to UShort_t
// 12 Jan 2001 (TG) broke dependency to RecoTrack by changing rtrack to pointer
// 12 Jan 2001 (TG) removed GetLorentzVector(), added energy()
// 12 Jan 2001 (TG) removed beginVertex(), endVertex(), flight()
// 05 Apr 2001 (TG) pvec() added
// 25 Jul 2001 (TG) changed rtrack from pointer to object, removed const from
//                  RelStreamer()
// 13 Aug 2001 (TG) changed class version to 2, removed Mass, mass(rhs)
//                  members, added TrackIf inheritance
// 14 Aug 2001 (TG) added relation to end vertices, moved Mother to MC_Vertex,
//                  removed set_topp(), changed ARTE CTOR
// 15 Aug 2001 (TG) operator==(GTopp) added, added print_decay()
// 17 Aug 2001 (TG) added outsize(), changed topp() return type to ToppIf
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added, incremented class
//                  version, added relation to outgoing MC_Tracks, removed
//                  length(), added Flag member, incremented class version
// 13 Sep 2001 (TG) added Mother, mother(), set_mother(): MC mother track
// 09 Jan 2002 (TG) changed return type of propagate() to bool
//
// *****************************************************************************

#include <iosfwd>
#include "arte/TObject.hh"
#include "interfaces/TrackIf.hh"
#include "arte/artefwd.hh"
#include "clue/cluefwd.hh"
#include "grelation/ClueRelation.hh"

class ToppIf;
class MC_Vertex;
class MC_Track;

#ifndef __CINT__
#include "smatrix/SVector.hh"
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** Monte Carlo Track.
      In case of prompt decays (e.g. $\phi\to\textrm{K}^+ + \textrm{K}^-$) no
      begin- or end-vertices
      exist in ARTE (to reduce event size). Therefore we need in addition to
      begin- and end-vertex relations relations to mother- and outgoing tracks.
      In case of decays which have a travel distance this introduces some
      redundancy: the mother \& outgoing tracks can be accessed either directly
      or via the begin- and end-vertices.

      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // MC_Track
  //==============================================================================
  class MC_Track: public TObject, public TrackIf {  
  public:
    
    /** @name --- Constructors --- */
    ///
    MC_Track();
    ///
    MC_Track(const MC_Track& rhs);
#ifndef NO_ARTE
    /// constructor reading from ARTE table
    MC_Track( const ArtePointer<MTRA>& t );
#endif
    ///
    ~MC_Track();
    ///
    MC_Track& operator=( const MC_Track& );

    /** @name --- TrackIf Access methods --- */
    /// Track $x$ position
    float x()   const;
    /// Track $y$ position
    float y()   const;
    /// Track $x$ position at $z$
    float x(float z)   const;
    /// Track $y$ position at $z$
    float y(float z)   const;
    /// Track $z$ position
    float z()   const;
    /// Track slope $t_x$
    float tx()  const;
    /// Track slope $t_y$
    float ty()  const;
    /// Track momentum $p$
    float p()   const;
    /// transverse Track momentum $p_t$ ($\perp$ to $z$)
    float pt()  const;
    /// Track momentum along $z$
    float pz()  const;
    /// dummy function: returns 0
    float chi2() const;
    /// dummy function: returns 0
    unsigned short int ndf()  const;
    /// azimuthal angle $\phi$ [deg]
    float phi() const;
    /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
    float theta() const;
    /// pseudorapidity:\\ $\eta = -\log\tan(\theta/2.)$
    float eta() const;
    /// track charge
    int charge() const;
    
    /// track energy $E=\sqrt{m^2 + p^2}$, rm is ignored
    float energy(double rm = 0.) const;
    /// Feynman Variable:\\ $x_F = \frac{E+p_z}{(E+p_z)_{max}}$, rm is ignored
    float xf(double rm = 0.) const;
    /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$, rm is ignored
    float rap(double rm = 0.) const;
    
#ifndef __CINT__
    /// $\vec{x} = (x,y,z)$
    SVector<double,3> xvec() const;
    /// $\vec{t} = (t_x,t_y,1)$ slope vector
    SVector<double,3> tvec() const;
    /// $\vec{e} = (e_x,e_y,e_z)$ unit vector along $\vec{p}$
    SVector<double,3> evec() const;
    /// $\vec{p} = (p_x,p_y,p_z)$ momentum vector
    SVector<double,3> pvec() const;
    /// dummy function: returns zero matrix
    const SMatrix<double,5>& COV() const;
    /// dummy function: returns zero matrix
    const SMatrix<double,5>& CINV() const;
#endif
    
    /// dummy function: return 0
    float cov_x(double dz=0.)        const;
    /// dummy function: return 0
    float cov_y(double dz=0.)        const;
    /// dummy function: return 0
    float cov_tx()                   const;
    /// dummy function: return 0
    float cov_ty()                   const;
    /// dummy function: return 0
    float cov_p()                    const;

    /// dummy function: always return true
    bool isValid() const;
    
    /** @name --- Operators --- */
    /// comparison between Track objects
    bool operator==(const TrackIf& rhs) const;
    /// compare with GTOPP
    bool operator==(const GTopp& rhs) const;
    
    /** @name --- Expert methods --- */
    /// dummy function: do nothing
    void valid();
    /// dummy function: do nothing
    void invalid();
    /// collect TrackIf pointers
    void collect(vector<TrackIf*>& c) const;
    /// dummy function: do nothing
    bool propagate(const double dz);
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;
    
    /** @name --- Access methods --- */
    /// return topp mass $m$
    float mass()   const;
    /// return topp name
    const char* const name() const;
    /** MC flag: 
	0: just created; not yet processed\\
	1: expected to decay according to QMTRA\_LENG\\
	2: decayed resonance or electromagnetic decay (production vertex = decay vertex)\\
	3: decay with non-zero track length\\
	4: track ends in a electromagnetic or hadronic interaction\\
	5: track leaves the detector volume\\
	6: track stops inside detector volume\\
	7: track curls up inside detector volume\\
	8: track stopped by user
    */
    unsigned int flag() const;

    /// link to Topp table
    const ToppIf& topp() const;
    /// pointer to begin vertex
    const ArtePointer<MC_Vertex> bVertex() const;
    /// set pointer to begin vertex
    void set_bVertex(const ArtePointer<MC_Vertex>& p);
    /// pointer to begin vertex
    const ArtePointer<MC_Track> mother() const;
    /// set pointer to begin vertex
    void set_mother(const ArtePointer<MC_Track>& t);

    /** print MC decay chain. Options for additional information:
	f: print track flag
	m: print ArtePointer<MC\_Track>
	r: print ArtePointer<RecoTrack> (matches to reconstructed tracks)
    */
    void print_decay(unsigned int level, char opt = ' ',
		     const ArtePointer<MC_Track> mtra = 0, unsigned int lvl=0) const;
    /// no of outgoing tracks
    unsigned int outsize() const;
    
    /** @name --- CLUE relations --- */
    /// MC\_Track $\leftrightarrow$ RecoTrack relation
    ClueRelationList<RecoTrack> rtrack;      //! not persistent
    ///
    const ClueRelationList<RecoTrack>& rTrack() const;
    ///
    ClueRelationList<RecoTrack>& rTrack();
    ///
    ClueRelationList<RecoTrack>& 
    container(const ArtePointer<RecoTrack>& p);

    /// MC\_Track $\leftrightarrow$ MC\_Vertex relation (end vertices)
    ClueRelationList<MC_Vertex> endVtx;      //! not persistent
    ///
    const ClueRelationList<MC_Vertex>& EndVtx() const;
    ///
    ClueRelationList<MC_Vertex>& EndVtx();
    ///
    ClueRelationList<MC_Vertex>& 
    container(const ArtePointer<MC_Vertex>& p);
    
    /// MC\_Track $\leftrightarrow$ MC\_Track relation (outgoing tracks)
    ClueRelationList<MC_Track> endTrack;      //! not persistent
    ///
    const ClueRelationList<MC_Track>& EndTrack() const;
    ///
    ClueRelationList<MC_Track>& EndTrack();
    ///
    ClueRelationList<MC_Track>& 
    container(const ArtePointer<MC_Track>& p);
    
  private:
    /// helper function for print_decay
    void print_opt(char opt, const ArtePointer<MC_Track> mtra = 0) const;

    Float_t  Px;
    Float_t  Py;
    Float_t  Pz;
    UShort_t iTopp;      // TOPP pointer
    UShort_t B_Vertex;   // MC_Vertex pointer
    UShort_t Mother;     // MC mother track
    UShort_t Flag;       // MC flag

#ifdef USE_ROOT
  public:
    /** @name --- ROOT only functions --- */
    /// call Streamer of relation container
    void RelStreamer(TBuffer& R__b, const Version_t& StreamVersion);
    /// 
    void Print(Option_t *option="");
    
    /** Do not call this function directly, use Draw() instead. Before
	you can draw, create a TView. */
    virtual void Paint(Option_t *option="");

    ClassDef(MC_Track,5)
#endif
};  // end of class MC_Track


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<< ( std::ostream& os, const MC_Track& t );

#ifdef NAMESPACE
} // namespace CLUE
#endif

#ifndef __CINT__
#include "MC_Track.icc"
#endif
#endif
