#ifndef __PRIMARYVERTEX_HH
#define __PRIMARYVERTEX_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   31. Aug 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A class to fit primary vertices by using the SVertex package
//
// changes:
// 31 Aug 2001 (TG) creation
// 10 Okt 2001 (TG) added typedefs
// 12 Nov 2001 (TG) comments improved
//
// ********************************************************************

#include <iosfwd>
#include <vector>
#include "interfaces/VertexIf.hh"
#include "smatrix/SVertex.hh"

class TrackIf;
class VertexIf;

/** Primary Vertex class.

    This class is developed for the optimized fit of primary vertices. For
    primary vertices with $N_\textrm{tracks}\le 20$ the high-performance
    vertexing class \texttt{SVertex} is used (see the \texttt{SMatrix}
    documentation). For vertices with more tracks the slow but dynamic-sized
    \texttt{vt++} vertexing class is used (see \texttt{Vt++} documentation).

    @author T. Glebe
*/
//==============================================================================
// PrimaryVertex
//==============================================================================
class PrimaryVertex : public VertexIf {
public:
  /** @name --- Constructors --- */
  ///
  PrimaryVertex();
  ///
  virtual ~PrimaryVertex();

  /** @name --- Typedefs --- */
  ///
  typedef vector<TrackIf*>      t_vec;
  ///
  typedef t_vec::iterator       iterator;
  ///
  typedef t_vec::const_iterator const_iterator;

  /** @name --- VertexIf Access methods --- */
  /// $v_x =$ vertex $x$ position
  float vx()   const;
  /// $v_y =$ vertex $y$ position
  float vy()   const;
  /// $v_z =$ vertex $z$ position
  float vz()   const;
  /// $\chi^2$ of vertex fit
  float chi2() const;
  /// upper tail $\chi^2$ probability
  float prob() const;
  /// degrees of freedom of vertex fit
  unsigned short int ndf()  const;
  /// no of tracks in vertex
  unsigned short int ntracks() const;
  /// return a track pointer
  const TrackIf* track(unsigned int i) const;
  /// read/write track access
  const TrackIf*& track(unsigned int i);

#ifndef __CINT__
  /// vertex position $\vec{v} = (v_x,v_y,v_z)$
  SVector<double,3> vpos() const;
  /// vertex covariance matrix
  const SMatrix<double,3>& VCOV() const;
#endif

  /// $\sigma_x^2$ of vertex
  float vtx_cov_x()   const;
  /// $\sigma_y^2$ of vertex
  float vtx_cov_y()   const;
  /// $\sigma_z^2$ of vertex
  float vtx_cov_z()   const;

  /** @name --- Expert methods --- */
  /// used by operator<<()
  std::ostream& print( std::ostream& ) const;

  /** @name --- Expert functions --- */
  /// do Kalman-Filter fit
  bool findVertexVt();

  /// container for TrackIf pointers (public, for simplicity)
  t_vec tracks;

private:
  SVertex<20>      vtx;
};

//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<< (std::ostream& os, const PrimaryVertex& vt) {
  return vt.print( os );
}
#endif
