#ifndef RECO_VERTEX_HH
#define RECO_VERTEX_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   02. Apr 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for a reconstructed vertex
//
// changes:
// 02 Apr 1997 creation by I. Siccama, C. Borgmeier
// 30 Okt 2000 (TG) removed FORTRAN-VT
// 03 Nov 2000 (TG) added #include <vector> in front of defineRelation()
// 06 Nov 2000 (TG) removed member function addVertex()
// 20 Nov 2000 (TG) added namespace CLUE
// 06 Dez 2000 (TG) replaced ARTE relations by CLUE relations
// 14 Dez 2000 (TG) RelStreamer method added
// 12 Jan 2001 (TG) remove fit(), mass(), point(), resultTrack()
// 12 Jan 2001 (TG) removed some constructors
// 12 Jan 2001 (TG) decoupled class from Point3D
// 11 Feb 2001 (TG) added members Chi2, Ndf, Mass; removed inheritance
//                  from RecoTrack
// 27 Feb 2001 (TG) inherited RecoVertex from VertexIf
// 16 Mar 2001 (TG) Ntracks, ntracks(), xerr(), yerr(), zerr() added
// 16 Mar 2001 (TG) print_cov() added
// 18 Apr 2001 (TG) moved inline functions to RecoVertex.icc
// 24 Apr 2001 (TG) replaced isfit by valid_, added valid() and invalid()
// 25 Apr 2001 (TG) removed lock(), unlock(), locked(), prob() modified
// 08 Mai 2001 (TG) valid() added
// 13 Jun 2001 (TG) VCOV() added
// 09 Jul 2001 (TG) removed ntrack(), changed return type of ntracks()
// 25 Jul 2001 (TG) made out non persistent
// 26 Jul 2001 (TG) added track()
// 27 Jul 2001 (TG) increased class version to 2
// 05 Sep 2001 (TG) RelStreamer: StreamVersion added
// 22 Jan 2002 (TG) moved vpos(), VCOV() code to .C file
//
// ********************************************************************

#include <iosfwd>
#include <algorithm>
#include "arte/TObject.hh"
#include "interfaces/VertexIf.hh"
#include "grelation/ClueRelation.hh"

#ifndef __CINT__
#include "smatrix/SVector.hh"
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /**
     Reconstructed vertex
     @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // RecoVertex
  //==============================================================================
  class RecoVertex: public TObject, public VertexIf {
  public:
    /** @name --- Constructors --- */
    /// default constructor
    RecoVertex();

    /// copy constructor
    RecoVertex( const RecoVertex& rhs );

    /// destructor
    ~RecoVertex();

#ifndef NO_ARTE
    /// from ArtePointer
    RecoVertex( const ArtePointer<RVER>& rver );
#endif
    ///
    const RecoVertex& operator=( const RecoVertex& rhs );

    /** @name --- Operators --- */
    /// $z_1 < z_2$
    bool operator<( const RecoVertex& rhs ) const;

    /** @name --- VertexIf access methods --- */
    /// $v_x =$ vertex $x$ position
    float vx()   const;
    /// $v_y =$ vertex $y$ position
    float vy()   const;
    /// $v_z =$ vertex $z$ position
    float vz()   const;
    /// $\chi^2$ of vertex fit
    float chi2() const;
    /// degrees of freedom of vertex fit
    unsigned short int ndf() const;
    /// probability of the vertex fit
    float prob() const;
    /// no of tracks in vertex
    unsigned short int ntracks() const;
    /// returns pointer to track in vertex (slow access via relation!)
    const TrackIf* track(unsigned int i) const;
    /// returns null pointer + prints warning
    const TrackIf*& track(unsigned int i);
 
    /// $\sigma_x^2$ of vertex
    float vtx_cov_x()   const;
    /// $\sigma_y^2$ of vertex
    float vtx_cov_y()   const;
    /// $\sigma_z^2$ of vertex
    float vtx_cov_z()   const;

    /** @name --- Access methods --- */
    /// is vertex valid?
    bool valid() const;
    /// mark vertex as invalid
    void invalid();
    /// mark vertex as valid
    void set_valid();
    /// error in $x$
    float xerr() const;
    /// error in $y$
    float yerr() const;
    /// error in $z$
    float zerr() const;
    /// invariant mass in vertex
    float mass() const;

#ifndef __CINT__
    /// vertex position $\vec{v} = (v_x,v_y,v_z)$
    SVector<double,3> vpos() const;
    /// vertex covariance matrix
    const SMatrix<double,3>& VCOV() const;
#endif

    /// used by operator<<()
    std::ostream& print( std::ostream& os ) const;
    /// print covariance matrix
    std::ostream& print_cov( std::ostream& os ) const;

    /// set vertex $\chi^2$
    void set_chi2(float chi);
    /// set degrees of freedom
    void set_ndf(int nd);
    /// set invariant mass
    void set_mass(float ms);

    /// $\sigma_x \times \sigma_y$ of vertex
    float vtx_cov_x_y() const;
    /// $\sigma_x \times \sigma_z$ of vertex
    float vtx_cov_x_z() const;
    /// $\sigma_y \times \sigma_z$ of vertex
    float vtx_cov_y_z() const;

    /// set $\sigma_x^2$ of vertex
    void set_vtx_cov_x  (const float c);
    /// set $\sigma_x \times \sigma_y$ of vertex
    void set_vtx_cov_x_y(const float c);
    /// set $\sigma_y^2$ of vertex
    void set_vtx_cov_y  (const float c);
    /// set $\sigma_x \times \sigma_z$ of vertex
    void set_vtx_cov_x_z(const float c);
    /// set $\sigma_y \times \sigma_z$ of vertex
    void set_vtx_cov_y_z(const float c);
    /// set $\sigma_z^2$ of vertex
    void set_vtx_cov_z  (const float c);

    /// set vertex position
    void set_xyz(const float x, const float y, const float z);

    /** @name --- CLUE relations --- */
    /// RecoVertex $\leftrightarrow$ RecoTrack (relations to outgoing tracks)
    ClueRelationList<RecoTrack> out;    //! non persistent
    ///
    const ClueRelationList<RecoTrack>& Out() const;
    ///
    ClueRelationList<RecoTrack>& Out();
    ///
    ClueRelationList<RecoTrack>& 
    container(const ArtePointer<RecoTrack>& p);

  private:
    Float_t  Vx;                     // vertex x coordinate
    Float_t  Vy;                     // vertex y coordinate
    Float_t  Vz;                     // vertex z coordinate
    Float_t  Chi2;                   // chi2 of vertex fit
    UShort_t Ndf;                    // degrees of freedom of vertex fit
    Float_t  Mass;                   // invariant mass in vertex
    Float_t  covMatrix[6];           // lower triangle covariance matrix
    UShort_t Ntracks;                // no of tracks in vertex
    UShort_t flag;                   // status flag (=0 if ok)
    
    int   fitFlag;                   //!
    int   momFlag;                   //! mode of fit
    bool  valid_;                    //! is vertex valid?

#ifndef __CINT__
    mutable SMatrix<double,3> VCOV_; //! cov. matrix (not persistent)
#endif

#ifdef USE_ROOT
  public:
    /// only availabe in ROOT: call Streamer of relation container
    void RelStreamer(TBuffer& R__b, const Version_t& StreamVersion);
    /// only available in ROOT
    void Clear(Option_t *option="");
    /// only available in ROOT
    void Print(Option_t *option="");
    /// only available in ROOT
    void Paint( Option_t *option="" );

    ClassDef(RecoVertex,2)
#endif
  }; // end of class RecoVertex

  
  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<< (std::ostream& os, const RecoVertex& vt) {
    return vt.print( os );
  }

#ifndef __CINT__
#include "RecoVertex.icc"
#endif

#ifdef NAMESPACE
} // namespace CLUE
#endif

#endif
