#ifndef CLUETOPP_HH
#define CLUETOPP_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   20. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Particle property description class
//
// changes:
// 20 Jul 1997 creation by C. Borgmeier
// 14 Nov 2000 (TG) code cleanup, doc++ comments added
// 20 Nov 2000 (TG) added namespace CLUE
// 11 Jan 2001 (TG) changed type of Anti to UShort_t
// 29 Jan 2001 (TG) added anti_chrg(), anti_name()
// 28 Feb 2001 (TG) operator==() added
// 05 Mar 2001 (TG) moved anti() to anti_ptr(), added new anti()
// 15 Aug 2001 (TG) moved ToppIf member definitions to .C file
// 05 Sep 2001 (TG) lund() added to ToppIf
//
// ********************************************************************

#include <iosfwd>
#include "arte/TObject.hh"
#include "interfaces/ToppIf.hh"

template <class T> class ArtePointer;
class TOPP;

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** TOPP table entry.

      Class to represent particle properties.

      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // TOPP
  //==============================================================================
  class Topp : public TObject, public ToppIf {
  public:
    /** @name --- Constructors --- */
    ///
    Topp();
#ifndef NO_ARTE
    /// only available with ARTE
    Topp( const ArtePointer<TOPP>& topp );
    ///
    const Topp& operator=( const ArtePointer<TOPP>& topp );
#endif
  
    /** @name --- ToppIf Access methods --- */
    /// particle name
    const char* const name() const;
    /// mass
    float mass() const;
    /// width
    float wid()  const;
    /// truncate Breit-Wigner at mass - trbw
    float trbw() const;
    /// life time
    float life() const;
    /// 3 * charge
    int chrg()   const;
    /// Do we represent a valid Topp entry?
    bool valid() const;
    /// get anti particle
    const ToppIf& anti() const;
    /// LUND code
    int lund()   const;

    /** @name --- Access Methods --- */
    /** colour information:\\
	0 = colour singlet;\\
	2 = gluon; 1 = quark or anti-diquark;\\
	-1 = anti-quark or diquark
    */
    int colo()   const;
    /// LUND decay flag (0 = no decay)
    int deca()   const;
    /// GEANT code
    int gean()   const;
    
    /// anti-particle row number
    const ArtePointer<Topp> anti_ptr() const;

    /// set anti-particle row number
    void set_anti(const ArtePointer<Topp>& topp);

    /** @name --- Operators --- */
    ///
    bool operator==(const ToppIf& rhs) const;

    /** @name --- Expert functions --- */
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;
    
private:
    char     Name[13];
    Float_t  Mass;
    Float_t  Wid;
    Float_t  Trbw;
    Float_t  Life;
    Int_t    Chrg;
    Int_t    Colo;
    Int_t    Deca;
    Int_t    Lund;
    Int_t    Gean;
    UShort_t Anti;
  
#ifdef USE_ROOT
  public:
    /** @name --- ROOT only functions --- */
    ///
    void Print(Option_t *option="");
    
    ClassDef(Topp,1)
#endif
 }; // end of class ClTOPP


  //==============================================================================
  // operator <<
  //==============================================================================
  std::ostream& operator<<( std::ostream& os, const Topp& t );

#ifdef NAMESPACE
} // namespace CLUE
#endif
#ifndef __CINT__
#include "Topp.icc"
#endif
#endif /* CLUETOPP_HH */
