#ifndef CL_WIRE_HH
#define CL_WIRE_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   11. Mai 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for a HERA-B wire target.
//
// changes:
// 11 Mai 2000 (TG) creation
// 03 Nov 2000 (TG) removed #ifdef __CINT__, changed include files
// 20 Nov 2000 (TG) added namespace CLUE
// 12 Jan 2001 (TG) removed dependency from Point3D
// 09 Feb 2001 (TG) added RTAR constructor
// 04 Apr 2001 (TG) xvec(), dxvec(), dvec() added
// 05 Dec 2002 (WG) added wire-number wnr and accessor wnr()
//
// **********************************************************************

#include <iosfwd>
#include "arte/TObject.hh"
#include "arte/artefwd.hh"

#ifndef __CINT__
#include "smatrix/SVector.hh"
#endif

#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to  fake doc++

  /**
     Target wire class.
     @author T. Glebe
  */
  //==============================================================================
  //  Wire
  //==============================================================================
  class Wire : public TObject {
  public:
    /** @name --- Constructors --- */
    /// 
    Wire();
#ifndef NO_ARTE
    /// only available in ARTE
    Wire( const ArtePointer<GTAR>& wire );
    /// only available in ARTE
    Wire( const ArtePointer<RTAR>& wire );
#endif
    ///
    Wire( const Wire& rhs );
    ///
    ~Wire();

    /** @name --- Access methods --- */
    /// x of wire center
    float  x() const;
    /// y of wire center
    float  y() const;
    /// z of wire center
    float  z() const;
    
    /// size in x-direction
    float dx() const;
    /// size in y-direction
    float dy() const;
    /// size in z-direction
    float dz() const;

	/// wire number
	int   wnr() const;
    
#ifndef __CINT__
    /// position vector $\vec{x} = (x,y,z)$
    SVector<double,3>  xvec() const;
    /// size vector $\vec{dx} = (dx,dy,dz)$
    SVector<double,3> dxvec() const;
    /// direction vector $\vec{d} = (1,0,0)$ or $\vec{d} = (0,1,0)$
    SVector<double,3> dvec() const;
#endif

    /** @name --- Expert functions --- */
    /// used by operator<<()
    std::ostream& print( std::ostream& ) const;
    
  private:
    Float_t X;
    Float_t Y;
    Float_t Z; 

    Float_t dX;
    Float_t dY;
    Float_t dZ;
    
    Int_t   Wnr;
	
#ifdef USE_ROOT
  public:
    /** @name --- ROOT only functions --- */
    ///
    void Print( Option_t *options="" );
    
    ///
    void Paint( Option_t *option="" );
    
    ClassDef(Wire,2)
#endif
      
  }; // end of Wire
  

  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<< ( std::ostream& os, const Wire& w ) {
    return w.print( os );
  }
  
#ifdef NAMESPACE
} // end of namespace CLUE
#endif

#ifndef __CINT__
#include "Wire.icc"
#endif
#endif
