// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   14. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Combines particles
//
// changes:
// 14 Jul 1997 creation and implementation by C. Borgmeier
// 03 Nov 2000 (TG) moved #include "TrackBase.hh" from header to source file
// 20 Nov 2000 (TG) added namespace CLUE
// 27 Dec 2000 (TG) end() optimization
//
// *****************************************************************************

#include <algorithm>
#include <iostream>

#include "gate/ObjArray.hh"  // needed for Combiner.hh
#include "clue/Combiner.hh"
#include "clue/TrackBase.hh"
#include "geometry/ClLorentzVector.hh"
#include "gate/ClonesArray.hh"


#ifdef NAMESPACE
namespace CLUE {
#endif

  //==============================================================================
  // Constructors
  //==============================================================================
  Combiner::Combiner() {
    vec_v* listP = new vec_v;
    listData = listP;
  }
  

  Combiner::Combiner( const Combiner& rhs ) { operator=( rhs ); }


  Combiner::~Combiner() {
#if 1 || defined(__GNUC__) && __GNUC__==2 && __GNUC_MINOR__ < 8
    delete listData;
#endif
  }


  Combiner::Combiner( TObjArray *l1,
		      TObjArray *l2,
		      TObjArray *l3,
		      TObjArray *l4, 
		      TObjArray *l5,
		      TObjArray *l6)
    : allowCollisionPattern( 0 ) {
    TObjArray_v ll;
    ll.push_back(l1);
    if (l2) ll.push_back(l2);
    if (l3) ll.push_back(l3);
    if (l4) ll.push_back(l4);
    if (l5) ll.push_back(l5);
    if (l6) ll.push_back(l6);
    
    init(ll);
  }


  Combiner::Combiner( TObjArray &l1, int mask )
    : allowCollisionPattern( mask ) {
    TObjArray_v ll;
    ll.push_back(&l1);
    
    init(ll);
  }
  
  
  Combiner::Combiner( TObjArray& l1, TObjArray& l2, int mask )
    : allowCollisionPattern( mask ) {
    
    TObjArray_v ll;
    ll.push_back(&l1);
    ll.push_back(&l2);
    
    init(ll);
  }


  Combiner::Combiner( TObjArray& l1, TObjArray& l2, TObjArray& l3, int mask )
    : allowCollisionPattern( mask ) {
    
    TObjArray_v ll;
    ll.push_back(&l1);
    ll.push_back(&l2);
    ll.push_back(&l3);
    
    init(ll);
  }


  Combiner::Combiner( TObjArray& l1,
		      TObjArray& l2,
		      TObjArray& l3,
		      TObjArray& l4,
		      int mask )
    : allowCollisionPattern( mask ) {
    
    TObjArray_v ll;
    ll.push_back(&l1);
    ll.push_back(&l2);
    ll.push_back(&l3);
    ll.push_back(&l4);
    
    init(ll);
  }
  

  Combiner::Combiner( TObjArray& l1,
		      TObjArray& l2,
		      TObjArray& l3,
		      TObjArray& l4,
		      TObjArray& l5,
		      int mask )
    : allowCollisionPattern( mask ) {
    
    TObjArray_v ll;
    ll.push_back(&l1);
    ll.push_back(&l2);
    ll.push_back(&l3);
    ll.push_back(&l4);
    ll.push_back(&l5);
  
    init(ll);
  }


  Combiner::Combiner( std::vector<GObjArray<PatternMatch*>*> &ll, int mask )
    : allowCollisionPattern( mask ) {
    
    TObjArray_v& casted = *reinterpret_cast<TObjArray_v*>( &ll );
    init( casted );
  }


  //==============================================================================
  // makeCl
  //==============================================================================
  GObjArray<TrackBase*>* makeCl( TObjArray* obj ) {
    return static_cast<GObjArray<TrackBase*>*>( obj );
  }


  //==============================================================================
  // init
  //==============================================================================
  void Combiner::init(TObjArray_v &ll) {
    
    vec_v* listP = new vec_v;
    listData = listP;
    vec_v& lists = *listP;
  
    //copy( ll.begin(), ll.end(), inserter( lists, lists.end() ) );
    std::transform( ll.begin(), ll.end(), back_inserter( lists ),
		    makeCl );

    const vec_vit end = lists.end();
    for( vec_vit i=lists.begin(); i != end; ++i) {
      vecplusit& v = *i;

      // look for first not locked track
      while( v.iter!=v.list->end()
	     && (*v.iter)->locked( allowCollisionPattern ) ) {
	++v.iter;
      }
      if( v.iter==v.list->end() ) {
	// all tracks are locked
	done = 1;
	return;
      }
      (*v.iter)->lock(); // lock track
      v.startList = end;
      vec_vit j = i;
      while( j != lists.begin() ) {
	--j;
	if( j->list==i->list ) {
	  v.startList = j;
	  break;
	}
      }
      currList.push_back( *v.iter );
    }

    done = 0;
  }


  //==============================================================================
  // operator =
  //==============================================================================
  Combiner& Combiner::operator=( const Combiner& rhs ) {
    if( this == &rhs ) return *this;

    vec_v& lists          = *listData;
    const vec_v& rhslists = *rhs.listData;
    lists                 = rhslists;
    done                  = rhs.done;
    allowCollisionPattern = rhs.allowCollisionPattern;
    currList = rhs.currList;
    std::vector<int> startIdx( rhslists.size() );
    int i = rhslists.size();
    while( i>0 ) {
      --i;
      if( rhslists[i].startList==rhslists.end() ) {
	lists[i].startList = lists.end();
      } else {
	for( int j = i-1; j>=0; --j ) {
	  if( rhslists[i].startList==&rhslists[j] ) {
	    lists[i].startList = &lists[j];
	    break;
	  }
	}
      }
    }
    
    return *this;
  }


  //==============================================================================
  // operator ++
  //==============================================================================
  Combiner& Combiner::operator++() {
    if (!done) { nextComb(); }
    return *this;
  }


  //==============================================================================
  // operator ++
  //==============================================================================
  Combiner Combiner::operator++(int) {
    
    Combiner tmp( *this );
    ++(*this);
    return tmp;
  }


  //==============================================================================
  // orderMatters
  //==============================================================================
  void Combiner::orderMatters() {
    vec_v& lists = *listData;
    const vec_vit end = lists.end();
    for( vec_vit i = lists.begin(); i != end; ++i ) {
      (*i).startList = end;
    }
  }


  //==============================================================================
  // mass
  //==============================================================================
  float Combiner::mass() const {

    // #ifdef USE_ROOT
    LorentzVector l( 0,0,0,0 );

    const GClonesArray<TrackBase*>::const_iterator end = currList.end();
    for( GClonesArray<TrackBase*>::const_iterator i = currList.begin();
	 i != end; ++i ) {
      l += (*i)->GetLorentzVector();
    }

    return l.mag();
    // #endif
  }


  //==============================================================================
  // print
  //==============================================================================
  void Combiner::print( std::ostream& os ) const {

    if( listData!=0 ) {
      std::copy( listData->begin(), listData->end(),
		 ostream_iterator<vecplusit>( os, "/" ) );
    }
    //#ifdef USE_ROOT
    os << ":" << (done? "done": "") 
       << ": " << currList;
    //#endif
  }


#ifdef USE_ROOT
  //==============================================================================
  // Print
  //==============================================================================
  void Combiner::Print( Option_t* ) { print( std::cout ); }


  void Combiner::Streamer(TBuffer &R__b) { TObject::Streamer( R__b ); }
#endif


  //==============================================================================
  // nextComb
  //==============================================================================
  // find next combination
  // return false if done
  void Combiner::nextComb() {
    vec_v& lists = *listData;
    done = !nextComb( lists.begin(), lists.end() );
  }


  //==============================================================================
  // nextComb
  //==============================================================================
  // find next combination in for lists begin upto end
  // return false if done
  bool Combiner::nextComb( vecplusit* begin, vecplusit* end ) {
    vec_v& lists = *listData;

    if( !done ) {
      vec_vit i = end;
      if( i!=begin ) {
	--i;                                  // begin on the rightmost list
	(*(*i).iter)->unlock();               // free rightmost element
	if( incrIndex( *i ) ) {
	  
#ifdef USE_ROOT
	  currList.set( i-begin, *(*i).iter ); // ok. no need
#else
	  currList[i-begin] = *(*i).iter;      //     to touch other lists
#endif
	  (*(*i).iter)->lock();
	  return true;
	} else {
	  if( nextComb( begin, i ) ) {     // find next comb in remaining lists
	    
	    if( (*i).startList!=lists.end() ) { // same list already present?
	      (*i).iter = (*i).startList->iter;
	      if( !incrIndex( *i) ) {           // is this wrong?
		return false;
	      }
	    }
	    else {
	      (*i).iter = (*i).list->begin();
	    }
#ifdef USE_ROOT
	    currList.set( i-begin, *(*i).iter );
#else
	    currList[i-begin] = *(*i).iter;
#endif
	    return true;
	  }
	}
      }
    }
    return false;
  }
 
 
  //==============================================================================
  // incrIndex
  //==============================================================================
  bool Combiner::incrIndex( vecplusit& v ) {

    while( v.iter!=v.list->end() ) {
      ++v.iter;
      if( v.iter!=v.list->end() ) {
	if( !(*v.iter)->locked( allowCollisionPattern ) ) {
	  return true;
	}
      }
    }
    return false;
  }

#ifdef USE_ROOT
  ClassImp(Combiner)
#endif
#ifdef NAMESPACE
} // end of namespace CLUE
#endif
