#ifndef PATTERN_MATCH_HH
#define PATTERN_MATCH_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   22. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: describe a track pattern
//
// changes:
// 22 Jul 1997 creation and 1st implementation by C. Borgmeier
// 03 Nov 2000 (TG) code polishing
// 20 Nov 2000 (TG) added namespace CLUE
//
// *****************************************************************************

#include <iosfwd>
#include "clue/TrackBase.hh"
#include "clue/cluefwd.hh"


#ifdef NAMESPACE
namespace CLUE {
#endif
  class abc; // to fake doc++

  /** Class for describing a track pattern.
      @memo Pattern Match class
      @author C. Borgmeier
  */
  //==============================================================================
  // PatternMatch
  //==============================================================================
  class PatternMatch: public TrackBase {
  public:
    /** @name --- Constructors --- */
    ///
    PatternMatch();
    
    /** Create a particle connecting match and pat. If owns==true,
	delete track in the destructor */
    PatternMatch( const TrackBase& match, const Pattern& pat, bool owns = true );

    ///
    PatternMatch( const PatternMatch& rhs );

    /// Virtual destructor
    virtual ~PatternMatch();

    /// Assignment operator
    const PatternMatch& operator=( const PatternMatch& rhs );

    /** @name --- Access methods --- */
    ///
    virtual const RecoTrack* getRecoTrack() const;
    ///
    virtual bool isCC() const;
    ///
    virtual float mass() const;

    /// Return the Pattern this particle is matched to.
    inline const Pattern* GetPattern() const { return pattern; }

//      /// $(p_x,p_y,p_z,\sqrt{p^2 + m^2})$
//      LorentzVector GetLorentzVector() const;

    /** @name --- Expert functions --- */
    /// Moves contents to target and clears self.
    void transfer( PatternMatch& target );
    /// used by operator<<()
    std::ostream& print( std::ostream& os ) const;
    ///
    virtual void lock();
    ///
    virtual void unlock();
    ///
    virtual int locked( int bit = 0 ) const;
    
    
    TrackBase* track;

    GObjArray<PatternMatch*> children;
    
    Pattern* pattern;
    
    Bool_t ccFlag;

    Bool_t ownsTrack;

#ifdef USE_ROOT
    /// only available in ROOT
    void Print( Option_t* = "" );

    ClassDef(PatternMatch, 1)
#endif
  }; // end of class PatternMatch


  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<<( std::ostream& os, const PatternMatch& m ) {
    return m.print( os );
  }

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif
