// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   Mit Jul 11 13:45:50 MEST 2001
//
// author:    glebe
//
// Description: Code which reconstructs a pi0 -> gamma gamma in the ecal.
//
// changes:
// 14 Sep 2001 (TG) renamed sdecaysTo() to decaysTo() and sswith() to with()
//
// ********************************************************************

#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TPaveText.h"
#include "BEE.hh"
#include "include/Neutana.hh"

static const float zECAL = 1348.3; 

//==============================================================================
// Constructor
//==============================================================================
Neutana::Neutana(RtEvent* evt) :
  event(evt),
  Photon(false),
  pizero(false),
  kzero(false)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool Neutana::EventSelect() {
  return true;
}

//==============================================================================
// DefinePattern
//==============================================================================
bool Neutana::DefinePattern() {

  Photon = photon.with(pt>0.3);

  pizero = pi_zero.decaysTo(Photon & Photon);
  //  pizero->massrange(1.);
  pizero->massrange(0.05);

  kzero = K_zero.decaysTo(pizero & pizero);
  kzero->massrange(0.1);

  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool Neutana::DefineHistograms() {
  nhits     = new TH1F("nhits","no of ECAL hits", 100, 0., 100.);
  PiZero    = new TNtuple("PiZero","pi0","m:pt1:pt2:nh:d:p1:p2");
  Kzero     = new TNtuple("Kzero","K0","m:pt1:pt2:nh:prob:p1:p2");
  return true;
}

//==============================================================================
// Event
//==============================================================================
bool Neutana::Event() {

  Photon->clear();

  const ArtePointer<Wire> w(1);
  SVector<double,6> origin;
  origin.place_at(w->xvec(),0);
  origin.place_at(w->dxvec(),3);

  nhits->Fill(ArteTable<EcalHit>::size());

  //----------------------------------------------------------------
  // Photon
  //----------------------------------------------------------------
  const ArteTableIterator<EcalHit> end = ArteTable<EcalHit>::end();
  for(ArteTableIterator<EcalHit> it = ArteTable<EcalHit>::begin();
      it != end; ++it) {
    const ArtePointer<EcalHit> ehit = *it;
    //    cout << "ehit size: " << ehit->rtrack.size() << endl;
    ehit->calc_track(origin);
    // select photons which are not associated to charged tracks
    if(ehit->rtrack.size() == 0 && Photon->evalTrackProp(*ehit) == true) {
      //      cout << "add photon" << endl;
      //      ehit->print_cov(cout);
      //      cout << endl;
      // ========== Set Photon origin to wire =======
      ehit->PrepareForVertexing();
#ifdef XXX
      ehit->propagate(0);
      if(ehit->calc_CINV() == false) {
	cout << "Matrix inversion failed!!" << endl;
      }
      cout << "EcalHit propagated to 0: " << *ehit << endl;
      cout << "COV: " << endl << ehit->COV() << endl;
      cout << "CINV: " << endl << ehit->CINV() << endl;
#endif
      
      Photon->push_back(*ehit);

#ifdef XXX
      const TrackIf& eh = Photon->matches().back();
      cout << "EcalHit propagated to 0: " << eh << endl;
      cout << "COV: " << endl << eh.COV() << endl;
      cout << "CINV: " << endl << eh.CINV() << endl;
      const SMatrix<double,5> ID = eh.COV() * eh.CINV();
      cout << "COV*CINV: " << endl << ID << endl;
      return true;
#endif
    }
  }

  //----------------------------------------------------------------
  // Pi0
  //----------------------------------------------------------------
  pizero->combine();

  const TwoProngCandidateIterator PM_end = pizero->matches().end();
  for(TwoProngCandidateIterator PM_it = pizero->matches().begin();
      PM_it != PM_end; ++PM_it) {
    TwoProngCandidate pi0cand = *PM_it;

#ifdef XXX
    cout << "DecayPatternMatch: " << endl << pi0cand << endl
	 << "-----------------------------------------------------------" <<
      endl;
#endif

    const SVertex<2>& vtx = *pi0cand->vertex();
    const TrackIf& t1 = *vtx.track(0);
    const TrackIf& t2 = *vtx.track(1);
    
    float nt[7];
    nt[0] = pi0cand->mass();
    nt[1] = t1.pt();
    nt[2] = t2.pt();
    nt[3] = ArteTable<EcalHit>::size();
    // cluster dist at ECAL
    nt[4] = sqrt(square(t1.x(zECAL)-t2.x(zECAL)) + square(t1.y(zECAL)-t2.y(zECAL)));
    nt[5] = t1.p();
    nt[6] = t2.p();

    if(nt[4]>50) {
      pi0cand->invalid(); // mark this pi0 as invalid
      continue;
    }

    PiZero->Fill(nt);
    //    cout << "mass: " << nt[0] << endl;
  } // for(PM_it...

  pizero->calc_mother();

  //----------------------------------------------------------------
  // K0
  //----------------------------------------------------------------
  kzero->combine(true);

  const TwoProngCandidateConstIterator KPM_end = kzero->matches().end();
  for(TwoProngCandidateConstIterator PM_it = kzero->matches().begin();
      PM_it != KPM_end; ++PM_it) {
    const TwoProngCandidate k0cand = *PM_it;

    //    cout << "Kzero: " << k0cand << endl;

    const SVertex<2>& vtx = *k0cand->vertex();
    const TrackIf& t1 = *vtx.track(0);
    const TrackIf& t2 = *vtx.track(1);
    
    float nt[7];
    nt[0] = k0cand->mass();
    nt[1] = t1.pt();
    nt[2] = t2.pt();
    nt[3] = ArteTable<EcalHit>::size();
    nt[4] = k0cand->prob();
    nt[5] = t1.p();
    nt[6] = t2.p();

#ifdef XXX
    cout << "mass: " << k0cand->mass() << " chi2: " << k0cand->chi2() 
	 << " ndf: " << k0cand->ndf() << " prob: " << nt[4]
	 << endl;
#endif

    Kzero->Fill(nt);
  } // for(PM_it...

#ifdef XXX
  cout << "EcalHit size: " << setw(3) << ArteTable<EcalHit>::size()
       << " Photon size: " << setw(3) << Photon->matches().size() 
       << " Pi0 size: "    << setw(3) << pizero->matches().size()
       << " K0 size: "     << setw(3) << kzero->matches().size()
       << endl;
#endif


  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool Neutana::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool Neutana::End() {
  return true;
}

