#ifndef G_CLONES_ARRAY_HH
#define G_CLONES_ARRAY_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Feb 1999
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: GClonesArray is a template wrapper 
//              for the ROOT TClonesArray class.
//
// changes:
// 27 Feb 1999 creation by C. Borgmeier
// 13 Nov 2000 (TG) added doc++ comments
// 14 Nov 2000 (TG) declared some member functions inline
// 20 Nov 2000 (TG) added namespace CLUE
// 14 Nov 2001 (TG) improved comments, moved function bodies to .icc file,
//                  added clear()
//
// ********************************************************************

#ifdef USE_ROOT

#include <iterator.h>
#include <TClonesArray.h>

#ifdef NAMESPACE
#include "gate/RtypesT.h"

namespace CLUE {
#endif

  template <class T> class GClonesArray;
  template <class T> class GObjArrayIterator;
  template <class T> class ConstGObjArrayIterator;
  template <class T> class GObjArray;

  
  /** GClonesArray is a STL compatible template wrapper for the ROOT
      TClonesArray class.
      This class is only available in ROOT. Outside ROOT, GClonesArray is
      typedef'd to a STL vector.

      \textbf{Note:} GClonesArray assumes that the container type is a pointer type!

      @memo Array of similar objects.
      @author C. Borgmeier
  */
  //==============================================================================
  // GClonesArray
  //==============================================================================
  template<class T>
    class GClonesArray: public TClonesArray {
    public:
      /** @name --- Typedefs --- */
      ///
      typedef T                         value_type;
      ///
      typedef value_type*               pointer;
      ///
      typedef size_t                    size_type;
      ///
      typedef value_type&               reference;
      ///
      typedef const value_type&         const_reference;
      ///
      typedef GObjArrayIterator<T>      iterator;
      ///
      typedef ConstGObjArrayIterator<T> const_iterator;
  
      /** @name --- Constructors --- */
      ///
      GClonesArray();
      ///
      GClonesArray( unsigned int newsize );
      ///
      GClonesArray( const GClonesArray<T>& rhs );
      ///
      ~GClonesArray();
      ///
      GClonesArray<T>& operator=( const GClonesArray<T>& rhs );
      
      /** @name --- Access methods --- */
      /// begin iterator
      iterator       begin();
      ///
      const_iterator begin() const;
      /// end iterator
      iterator       end();
      ///
      const_iterator end() const;      
      /// container size
      size_type size() const;
      /// is container empty?
      bool empty() const;
      
      /// add an object to container
      value_type push_back( T obj );

      /// delete last element
      void pop_back();
      /// erase element at iterator pos
      void erase( iterator position );
      /// erase elements in iterator range
      void erase( iterator p1, iterator p2 );
      /// clear container
      void clear();

      /// access via index
      value_type GetAt( int idx );      
      /// access via index
      TObject* operator[]( int idx ) const;
      
      /// set value via index
      void set( int idx, T v );
      
      /** @name --- Expert methods --- */
      /// used by operator<<()
      std::ostream& print( std::ostream& os ) const;
      
      /// put elements of this container into newArray container
      void saveObjects( GClonesArray<T>& newArray ) const;
      /// delete container and fill it with oldArray elements
      void copyObjectsFrom( const GObjArray<T>& oldArray );
      
      ///
      void Print( Option_t* = "" );
      
      ClassDefT(GClonesArray<T>,1)
 }; // end of class GClonesArray


  //==============================================================================
  // operator <<
  //==============================================================================
  template<class T>
    inline std::ostream& operator<<( std::ostream& os, const GClonesArray<T>& ar ) {
    return ar.print( os );
  }

#ifdef NAMESPACE
} // namespace CLUE

MyClassDefT2(CLUE,GClonesArray,T)
  ; // for doc++
#else
ClassDefT2(GClonesArray,T)
  ; // for doc++
#endif

#else

#include <vector>
#define GClonesArray std::vector

#endif  /* USE_ROOT */

//#include "ClonesArray.icc"
#endif  /* G_CLONES_ARRAY_HH */
