#ifndef __GCOMBINER_HH
#define __GCOMBINER_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   04. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: template for a container combiner
//
// changes:
// 04 Jan 2001 (TG) creation
// 05 Jan 2001 (TG) added begin(), end(), size(), getcur to GCombinerIterator
// 13 Jan 2001 (TG) changed return type of print() functions
// 13 Jan 2001 (TG) GCombinerIterator: declared curr and getcur
//                  mutable, changed type of several member functions to const
// 13 Jan 2001 (TG) optimized collision() function
// 22 Jan 2001 (TG) IteratorPair: added incr_reset() member
// 23 Jan 2001 (TG) bug fixes in documentation
// 30 Jan 2001 (TG) GCombiner: made begin(), end() const
// 07 Feb 2001 (TG) added deleteall(), ncomb()
// 08 Feb 2001 (TG) added GCombiner::valid()
// 08 Feb 2001 (TG) added cbegin(), cend(), citerator, const_citerator to loop
//                  over the rellist container container
// 01 Mai 2001 (TG) GCombiner CTOR added
// 09 Jul 2001 (TG) GCominerIterator::iterator, ::const_iterator added
// 09 Jul 2001 (TG) non-const begin(), end() added
//
// *****************************************************************************

#include <iosfwd>
#include <list>
#include <vector>
#include "GC_trait.hh"

template <class T> class GCombiner;
template <class T> class GCombinerIterator;
template <class T> class GUniqueCombinerIterator;

/** GCombiner helper class.
    This class performs the bookkeeping of the begin, end and current iterator
    for each container saved by the Combiner class. The running interator
    can be resetted to the begin of the container in case it reached the end of
    the container. This class avoids that one has to declare the
    \texttt{GCombinerIterator} as a friend of the \texttt{GCombiner} class and
    thus reduces the coupling between those classes.

    \textbf{Note:} This class is not for public use.
    @memo Combiner helper class
    @author T. Glebe
 */
//============================================================================
// Class IteratorPair - class for iterator range
//============================================================================
template <class Iterator>
class IteratorPair {
  ///
  typedef IteratorPair<Iterator> self;

public:
  /** @name --- Constructors --- */
  ///
  IteratorPair() {}
  ///
  IteratorPair(const IteratorPair& rhs) : 
    begin(rhs.begin), end(rhs.end), iter(rhs.iter) {}
  ///
  IteratorPair(const Iterator& a) : 
    begin(a), end(a), iter(a) {}
  ///
  IteratorPair(const Iterator& lower, const Iterator& upper) :
    begin(lower), end(upper), iter(lower) {}

  /** @name --- Access methods --- */
  /// has iter reached end?
  inline bool iteratable() const { return iter != end; }
  /// do an iteration step
  inline self& iterate() { ++iter; return *this; }

  /** @name --- Operators --- */
  ///
  inline bool operator==(const IteratorPair<Iterator>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const IteratorPair<Iterator>& rhs) const {
    return iter != rhs.iter;
  }
  ///
  inline bool operator<(const IteratorPair<Iterator>& rhs) const {
    return iter < rhs.iter;
  }

  /** @name --- Expert functions --- */
  ///
  inline void reset() { iter = begin; }
  ///
  inline void incr_reset() { if(distance(begin,end)>1) ++begin; iter = begin; }
  ///
  inline void print(std::ostream& os) const { os << *iter; }
  ///
  inline Iterator getIter() const { return iter; }

private:
  Iterator begin;  // lower bound
  Iterator end;    // upper bound
  Iterator iter;
};


//==============================================================================
// operator<<
//==============================================================================
template<class Iterator>
inline std::ostream& operator<<(std::ostream& s, const IteratorPair<Iterator>& a) {
  a.print(s);
  return s;
}

/** Container class.
    The Combiner class stores pointers to containers whose elements should
    be combined. The number of containers is unlimited.
    @memo Combiner
    @author T. Glebe
*/
//============================================================================
// Class GCombiner - container for container references
//============================================================================
template <class T>
class GCombiner {
#ifndef __CINT__
  typedef typename GC_trait<T>::Gvalue_type             Gvalue_type;
  typedef typename GC_trait<T>::Gpointer                Gpointer;
  typedef typename GC_trait<T>::GCombiner_v             GCombiner_v;
  typedef typename GC_trait<T>::GCombiner_v::value_type value_type;
  typedef typename GC_trait<T>::GCombiner_cit           GCombiner_cit;
  typedef typename GC_trait<T>::GCombiner_cit           citerator;
#else
  typedef GC_trait<T>::Gvalue_type             Gvalue_type;
  typedef GC_trait<T>::Gpointer                Gpointer;
  typedef GC_trait<T>::GCombiner_v             GCombiner_v;
  typedef GC_trait<T>::GCombiner_v::value_type value_type;
  typedef GC_trait<T>::GCombiner_cit           GCombiner_cit;
  typedef GC_trait<T>::GCombiner_cit           citerator;
#endif

public:
  /** @name --- Typedef's --- */
  ///
  typedef GCombinerIterator<T>             iterator;
  ///
  typedef GCombinerIterator<T>             const_iterator;

  /** @name --- Constructors --- */
  ///
  GCombiner();
  ///
  GCombiner(const GCombiner<T>& rhs);
  ///
  GCombiner(const Gpointer[], const unsigned int N);
  ///
  virtual ~GCombiner();

  /** @name --- Access functions --- */
  ///
  inline void push_back(const Gvalue_type& rel) { rellist.push_back(const_cast<T*>(&rel)); }
  ///
  inline Gpointer back() const { return rellist.back(); }
  ///
  inline int  size() const { return rellist.size(); }
  ///
  inline bool empty() const { return rellist.empty(); }
  /// check if combiner is valid: true if each container size is > 0
  bool valid() const;

  /** @name --- Iterator methods --- */
#ifndef __CINT__
  /// to iterate container container, not available in CINT
  inline    citerator cbegin() const { return rellist.begin(); }
  /// to iterate container container, not available in CINT
  inline    citerator cend()   const { return rellist.end(); }
#endif
  /// to iterate combinations
  inline     iterator begin()  const { return iterator(rellist, false); }
  /// to iterate combinations
  inline     iterator end()    const { return iterator(rellist, true); }

  /// direct access
  Gpointer operator[](unsigned int i) const;

  /** @name --- Operators --- */
  ///
  GCombiner<T>& operator=(const GCombiner<T>& rhs);
  ///
  inline bool operator==(const GCombiner<T>& rhs) const {
    return rellist == rhs.rellist;
  }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const;
  /// empty combiner
  inline void clear(void) { rellist.clear(); }
  /// empty combiner, call delete for pointers
  void deleteit(void);
  /// compute number of combinations
  unsigned int ncomb(void) const;
  
protected:
  GCombiner_v  rellist;
}; // class GCombiner


//==============================================================================
// operator<<
//==============================================================================
template<class T>
inline std::ostream& operator<<(std::ostream& s, const GCombiner<T>& a) {
  return a.print(s);
}


/** Combination iterator.
    The \texttt{GCombinerIterator} iterates over all combinations which are
    possible by the given container objects. The \texttt{GCombinerIterator}
    represents a vector of iterators of the container objects. Therefore this
    iterator behaves very different from a STL iterator and is not
    STL conform! The \texttt{GCombinerIterator} itself has a \texttt{begin()}
    and \texttt{end()} operator to loop over the elements of the current
    combination. An \texttt{operator*()} for receiving the vector of container
    iterators and an \texttt{operator[]()} method for direct access to the
    container iterators is provided.
    @memo Iterates the combinations.
    @author T. Glebe
 */ 
//============================================================================
// Class GCombinerIterator: non conform to STL!!!
//============================================================================
template <class T>
class GCombinerIterator {
  
  typedef GCombinerIterator<T> self;
#ifndef __CINT__
  typedef typename GC_trait<T>::GCombiner_v        GCombiner_v;
  typedef typename GC_trait<T>::GCombiner_cit      GCombiner_cit;
  typedef typename GC_trait<T>::GIterator          GIterator;
  typedef typename GC_trait<T>::GIterator_v        GIterator_v;
  typedef typename GC_trait<T>::GIterator_it       GIterator_it;
  typedef typename GC_trait<T>::GIterator_cit      GIterator_cit;
  typedef typename GC_trait<T>::ItPair             ItPair;

  typedef typename GC_trait<T>::GCombination       GCombination;
  typedef typename GC_trait<T>::GCombination_it    GCombination_it;
  typedef typename GC_trait<T>::GCombination_cit   GCombination_cit;
#else
  typedef GC_trait<T>::GCombiner_v        GCombiner_v;
  typedef GC_trait<T>::GCombiner_cit      GCombiner_cit;
  typedef GC_trait<T>::GIterator          GIterator;
  typedef GC_trait<T>::GIterator_v        GIterator_v;
  typedef GC_trait<T>::GIterator_it       GIterator_it;
  typedef GC_trait<T>::GIterator_cit      GIterator_cit;
  typedef GC_trait<T>::ItPair             ItPair;

  typedef GC_trait<T>::GCombination       GCombination;
  typedef GC_trait<T>::GCombination_it    GCombination_it;
  typedef GC_trait<T>::GCombination_cit   GCombination_cit;
#endif

public:
  ///
  typedef GCombination_it  iterator;
  ///
  typedef GCombination_cit const_iterator;

  /** @name --- Constructors --- */
  ///
  GCombinerIterator() : getcur(false) {}
  ///
  GCombinerIterator(const GCombinerIterator<T>& rhs) :
    iter(rhs.iter), curr(rhs.curr), last(--iter.end()), getcur(false) {}
  ///
  GCombinerIterator(const GCombiner_v& v, bool isend = false) :
    getcur(false) {
    const GCombiner_cit cend  = v.end();
    const GCombiner_cit clast = --v.end();
    for(GCombiner_cit it = v.begin(); it != cend; ++it) {
      if(isend)
	if(it == clast)
	  iter.push_back(ItPair((*it)->end()));
	else
	  iter.push_back(ItPair((*it)->begin()));
      else
	iter.push_back(ItPair((*it)->begin(), (*it)->end()));
    } // for GCombiner
    last = --iter.end(); // remember last element
  }
  ///
  virtual ~GCombinerIterator() {}

  /** @name --- Access methods --- */
  ///
  inline GCombination_it begin() { getCombination(); return curr.begin(); }
  ///
  inline GCombination_it end()   { getCombination(); return curr.end(); }
  ///
  inline GCombination_cit begin() const { getCombination(); return curr.begin(); }
  ///
  inline GCombination_cit end()   const { getCombination(); return curr.end(); }
  ///
  inline int size() const { return iter.size(); }

  /** @name --- Operators --- */
  ///
  inline self&  operator++() { iterate(); return *this; }
  ///
  inline const GCombination&  operator*() const { getCombination(); return curr; }
  ///
  inline const GIterator& operator[](unsigned int i) const { 
    getCombination(); return curr[i]; }
  ///
  inline bool operator==(const GCombinerIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const GCombinerIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

  /** @name --- Expert functions --- */
  ///
  std::ostream& print(std::ostream& os) const;

protected:
  /// algorithm for an iteration step
  virtual void iterate();
  ///
  void getCombination() const;

  GIterator_v           iter;   // vector of container iterators
  mutable GCombination  curr;   // vector of current iterators
  GIterator_it          last;   // last iterator in iter
  mutable bool          getcur; // getCombination() status variable
}; // class GCombinerIterator


//==============================================================================
// operator<<
//==============================================================================
template<class T>
inline std::ostream& operator<<(std::ostream& s, const GCombinerIterator<T>& a) {
  return a.print(s);
}


/** Generic collision decection.
    In general, not every combination of objects from different containers is a
    useful one. In order to sort out those invalid combinations, one can define
    a collision function, which evaluates a current combination (represented by
    a \texttt{GCombinationIterator} if any two elements of the combination
    collides. In this case the function returns \texttt{true}, else
    \texttt{false}.
    This collision function detects a combination as invalid if any two objects
    are equal according to their \texttt{operator==()} method.

    The user can define his own collision functions according to the function
    below and introduce other collision criterias. E.g. in case of containers
    carrying track objects one could detect a collision in case any track of
    the combination is a clone of another track in the combination.

    In order to optimize this function, the containers must provide a
    \texttt{T::value\_type} typedef which gives the storage type of the
    container. Example: \texttt{vector<int>::value\_type} is \texttt{int}.
    @memo Collision function
    @author T. Glebe
*/
//============================================================================
// collision function: returns true if two iterators point to objects which
//                     are equal according to operator==()
//============================================================================
template <class T>
bool collision(GCombinerIterator<T>& rhs) {
  typedef GC_trait<T>::GCombination_cit GCombination_cit;
#ifndef __CINT__
  typedef typename T::value_type Val_t;
#else
  typedef T::value_type Val_t;
#endif

  const GCombination_cit vend = rhs.end();
  for(GCombination_cit cit = rhs.begin(); cit!= vend; ++cit) {

    const Val_t val = **cit; // optimization

    GCombination_cit cit2 = cit;
    for( ++cit2; cit2 != vend; ++cit2 ) {
      // attention: we compare values, not iterators!
      if(val == **cit2) {
	//	cout << " found collision! " << endl;
	return true;
      }
    }
  }
  //  cout << endl;
  return false;
}

//============================================================================
// PtrCollision function: returns true if two iterators point to objects which
//                     are equal according to their pointers
//============================================================================
template <class T>
bool PtrCollision(GCombinerIterator<T>& rhs) {
  typedef GC_trait<T>::GCombination_cit GCombination_cit;
#ifndef __CINT__
  typedef typename T::value_type Val_t;
#else
  typedef T::value_type Val_t;
#endif

  const GCombination_cit vend = rhs.end();
  for(GCombination_cit cit = rhs.begin(); cit!= vend; ++cit) {

    const Val_t* val = &**cit; // optimization

    GCombination_cit cit2 = cit;
    for( ++cit2; cit2 != vend; ++cit2 ) {
      // attention: we compare values, not iterators!
      if(val == &**cit2) {
	//	cout << " found collision! " << endl;
	return true;
      }
    }
  }
  //  cout << endl;
  return false;
}

//============================================================================
// load the implementation file
//============================================================================
#ifndef __CINT__
#include "GCombiner.icc"
#endif
#endif
