#ifndef __RPOINT3D_H__
#define __RPOINT3D_H__
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   15. Jun 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Point3D interface
//
// changes:
// 15 Jun 2000 (TG) creation
// 13 Nov 2000 (TG) added doc++ comments
//
// ********************************************************************

#include "geometry/ThreeVector.hh"

#ifdef USE_ROOT

// typedef ThreeVector Point3D;
#ifdef NAMESPACE
namespace CLUE {
#endif

  /**
     Point3D - ROOT based implementation for use with ROOT
     For member functions see ThreeVector documentation.
     @memo ROOT based implementation
     @author T. Glebe
  */
  //==============================================================================
  // Point3D
  //==============================================================================
  class Point3D: public ThreeVector {
  public:
    /** @name --- Constructors --- */
    ///
    Point3D() {}
    ///
    Point3D( const Point3D& rhs ): ThreeVector( rhs ) {}
    ///
    Point3D( const ThreeVector& rhs ): ThreeVector( rhs ) {}
    ///
    Point3D( const TVector3& rhs ): ThreeVector( rhs ) {}
    ///
    Point3D( const double x, const double y, const double z ): ThreeVector( x, y, z ) {}
    ///
    const Point3D& operator=( const Point3D& rhs ) {
      *static_cast<ThreeVector*>(this) = rhs;
      return *this;
    }
    
#ifdef USE_ROOT
    ClassDef(Point3D,1)
#endif
  }; // end of class Point3D
#ifdef NAMESPACE
} // namespace CLUE
#endif

//  //==============================================================================
//  // operator +
//  //==============================================================================
//  inline Point3D operator+( const Point3D& lhs, const ThreeVector& rhs ) {
//    Point3D tmp( lhs );
//    tmp += rhs;
//    return tmp;
//  }

//  //==============================================================================
//  // operator -
//  //==============================================================================
//  inline Point3D operator-( const Point3D& lhs, const ThreeVector& rhs ) {
//    Point3D tmp( lhs );
//    tmp -= rhs;
//    return tmp;
//  }

#else
#include <CLHEP/Geometry/Point3D.h>
#include "arte/TObject.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  /**
     Point3D - CLHEP based implementation for use outside ROOT
     For HepPoint3D member functions see CLHEP documentation.
     @memo CLHEP implementation
     @author C. Borgmeier
  */
  //==============================================================================
  // ThreeVector
  //==============================================================================
  class Point3D: public TObject, public HepPoint3D {
  public:
    /** @name --- Constructors --- */
    ///
    Point3D() {}
    ///
    Point3D( const Point3D& rhs ): HepPoint3D( rhs ) {}
    ///
    Point3D( const HepPoint3D& rhs ): HepPoint3D( rhs ) {}
    ///
    Point3D( const double x, const double y, const double z ): HepPoint3D( x, y, z ) {}
    ///
    const Point3D& operator=( const Point3D& rhs ) {
      *static_cast<HepPoint3D*>(this) = rhs;
      return *this;
    }
    
  /** @name --- Access methods --- */
    /// x() can be used, too
    inline double X() const { return x(); }
    /// y() can be used, too
    inline double Y() const { return y(); }
    /// z() can be used, too
    inline double Z() const { return z(); }
    
    /// set x-component in cartesian coordinates
    inline void set_x( const double x ) { setX(x); }
    /// set y-component in cartesian coordinates
    inline void set_y( const double y ) { setY(y); }
    /// set z-component in cartesian coordinates
    inline void set_z( const double z ) { setZ(z); }
  }; // end of class Point3D

#define Cross cross
#define Dot dot
#define Mag2 mag2
#define Mag mag
#define Unit unit
#define Phi phi
#define Theta theta

  //==============================================================================
  // operator +
  //==============================================================================
  inline Point3D operator+( const Point3D& lhs, const Hep3Vector& rhs ) {
    Point3D tmp( lhs );
    tmp += rhs;
    return tmp;
  }
  
  //==============================================================================
  // operator -
  //==============================================================================
  inline Point3D operator-( const Point3D& lhs, const Hep3Vector& rhs ) {
    Point3D tmp( lhs );
    tmp -= rhs;
    return tmp;
  }
  
  //==============================================================================
  // operator <<
  //==============================================================================
  inline ostream& operator<<( ostream& os, const HepPoint3D& rhs ) {
    return os << '(' << rhs.x() << ',' << rhs.y() << ',' << rhs.z() << ')';
  }

#ifdef NAMESPACE
} // namespace CLUE
#endif
#endif // USE_ROOT

#endif

/*
  /// point in 3d (adaption from CLHEP)
  class Point3D: public ThreeVector
  {
  public:
  Point3D() {}

  // Constructor
  Point3D(Double_t x, Double_t y=0, Double_t z=0) : ThreeVector(x, y, z) {}
  
  // Copy constructor
  Point3D(const ThreeVector& v) : ThreeVector(v) {}
  
  // Assignment
  Point3D& operator=(const ThreeVector &v) {
  setX(v.x()); setY(v.y()); setZ(v.z()); return *this;
  }
 
  void Print( Option_t *option="" );

  // Distance squared to the origin
  Double_t distance2() { return mag2(); }

  // Distance squared to the point
  Double_t distance2(Point3D &p) {
  Double_t delx = p.x()-x(), dely = p.y()-y(), delz = p.z()-z();
  return delx*delx + dely*dely + delz*delz;
  }

  // Distance to the origin
  Double_t distance() { return sqrt(distance2()); }

  // Distance to the point
  Double_t distance(Point3D &p) { return sqrt(distance2(p)); }

  // Transformation
  // Point3D& transform(const Transform3D &m);

  //void Streamer(TBuffer &R__b);
  ClassDef(Point3D,1)
  };
*/

