#ifndef __GRelationLIST_HH
#define __GRelationLIST_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Nov 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: relation container class
//
// changes:
// 27 Nov 2000 (TG) creation
// 30 Nov 2000 (TG) added #ifdef __CINT__ to make temlate classes work in CINT.
//                  CINT cannot handle traits, therefore the traits mechanism
//                  is disabled in case of CINT.
// 04 Dec 2000 (TG) made traits mechanism work for CINT
// 07 Dec 2000 (TG) GRelationList copy constructor added
// 27 Dec 2000 (TG) end() optimization
// 24 Apr 2001 (TG) GR_trait added
// 15 Aug 2001 (TG) all GRelationIterators: bug fix in operator--()
//
// *****************************************************************************

#include <iosfwd>
#include <list>

#ifdef USE_ROOT
#include "arte/TObject.hh"
#endif

template <class T> class GRelationList;
template <class T> class GRelationIterator;
template <class T> class ConstGRelationIterator;
template <class T> class ReverseGRelationIterator;
template <class T> class ConstReverseGRelationIterator;

/** \texttt{G\_trait} allows to configure the template classes.
    As a default, the Relations are based on a STL \texttt{list} container.
    This allows fast addition and removal of object relations with the cost of
    no sorting and no possiblity for direct access. Per default, pointers to
    objects are stored in the STL container. The behaviour of the
    GRelation container and iterator can be configured by specializing
    the G\_trait structure. For an example see HERA-B Note 00-188.

    @memo   Class configuration
    @author T. Glebe
*/
//============================================================================
// The G-Trait
//============================================================================
template <class T>
struct G_trait {
  typedef T                                            Gvalue_type;
  typedef T*                                           Gpointer;
  typedef T* const                                     Gconst_pointer;
  typedef T&                                           Greference;
  typedef const T&                                     Gconst_reference;
  typedef std::list<Gpointer>                          GRelation_v;
  typedef std::list<Gpointer>::iterator                GRelation_it;
  typedef std::list<Gpointer>::const_iterator          GRelation_cit;
  typedef std::list<Gpointer>::reverse_iterator        GRelation_rit;
  typedef std::list<Gpointer>::const_reverse_iterator  GRelation_rcit;
};


/** \texttt{GR\_trait} is used for type promotion. 
    This is needed in case you want to create a relation to a base class.
    Then just specialize \texttt{GR\_trait} to return the base type of the
    class.
    Example:
    \begin{verbatim}
class A : public BaseIf {
...
};

template <>
struct GR_trait<A> {
  typedef BaseIf val;
};

class B {
public:
  GRelationList<BaseIf> rel;

  GRelationList<BaseIf>& container(const BaseIf& rhs) { return rel; }
...
};

int main() {
  A a;
  B b;

  createSingleRelation(b,a);
}
    \end{verbatim}

    @memo Type promotion.
    @author T. Glebe
*/
//=============================================================================
// Type promotion trait
//=============================================================================
template <class T>
struct GR_trait {
  typedef T val;
};

/** Create bidirectional Relation between two Objects.

    @memo Many-to-Many
    @author T. Glebe
*/
//============================================================================
// create Relation
//============================================================================
template <class L, class R>
const bool createRelation(L& l, R& r);

/** Create bidirectional Relation between two Objects, where one
    relation containers is given explicitly as an argument.

    @memo Many-to-Many
    @author T. Glebe
*/
//============================================================================
// create Relation
//============================================================================
template <class L, class R>
const bool createRelation(L& l, GRelationList<R>& rcont, R& r);

/** Create bidirectional Relation between two Objects, where the
    relation containers are given explicitly as an argument.

    @memo Many-to-Many
    @author T. Glebe
*/
//============================================================================
// create Relation
//============================================================================
template <class L, class R>
const bool createRelation(L& l, GRelationList<R>& rcont,
			  R& r, GRelationList<L>& lcont);

/** Create unidirectional Relation between two Objects.

    @memo One-to-Many
    @author T. Glebe
*/
//============================================================================
// create Single Relation
//============================================================================
template <class L, class R>
const bool createSingleRelation(L& l, R& r);

/** Create unidirectional Relation between two Objects, where the
    relation container is given by an argument

    @memo One-to-Many
    @author T. Glebe
*/
//============================================================================
// create Single Relation
//============================================================================
template <class L, class R>
const bool createSingleRelation(L& l, GRelationList<R>& rcont, R& r);

/** Remove bidirectional Relation between two Objects.

    @memo  Many-to-Many
    @author T. Glebe
*/
//============================================================================
// remove Relation
//============================================================================
template <class L, class R>
const bool removeRelation(L& l, R& r);

/** Remove bidirectional Relation between two Objects, where one container
    is given by an argument.

    @memo  Many-to-Many
    @author T. Glebe
*/
//============================================================================
// remove Relation
//============================================================================
template <class L, class R>
const bool removeRelation(L& l, GRelationList<R>& rcont, R& r);

/** Remove bidirectional Relation between two Objects, where the containers
    are given by an argument.

    @memo  Many-to-Many
    @author T. Glebe
*/
//============================================================================
// remove Relation
//============================================================================
template <class L, class R>
const bool removeRelation(L& l, GRelationList<R>& rcont,
			  R& r, GRelationList<L>& lcont);

/** Remove unidirectional Relation between two Objects.

    @memo  One-to-Many
    @author T. Glebe
*/
//============================================================================
// remove Single Relation
//============================================================================
template <class L, class R>
const bool removeSingleRelation(L& l, R& r);

/** Remove unidirectional Relation between two Objects, where the container
    is given by an argument.

    @memo  One-to-Many
    @author T. Glebe
*/
//============================================================================
// remove Single Relation
//============================================================================
template <class L, class R>
const bool removeSingleRelation(L& l, GRelationList<R>& rcont,
				R& r);

/** Relation container class.
    As a default, the GRelationList container is mapped onto a STL
    \texttt{list} container. Therefore no direct access to relations is
    possible. The \texttt{operator[]()} emulates direct access. This is of
    course time intensive and should not be used if the size of the container is
    large. For other containers the \texttt{operator[]()} method can be
    overloaded.

    @memo Container.
    @author T. Glebe
*/
//============================================================================
// Class GRelationList - object relation container
//============================================================================
template <class T>
class GRelationList {
#ifndef __CINT__
  typedef typename G_trait<T>::Gvalue_type             Gvalue_type;
  typedef typename G_trait<T>::Gpointer                Gpointer;
  typedef typename G_trait<T>::GRelation_v             GRelation_v;
  typedef typename G_trait<T>::GRelation_v::value_type value_type;
  typedef typename G_trait<T>::GRelation_cit           GRelation_cit;
#else
  typedef G_trait<T>::Gvalue_type             Gvalue_type;
  typedef G_trait<T>::Gpointer                Gpointer;
  typedef G_trait<T>::GRelation_v             GRelation_v;
  typedef G_trait<T>::GRelation_v::value_type value_type;
  typedef G_trait<T>::GRelation_cit           GRelation_cit;
#endif

public:
  /** @name --- Typedef's --- */
  ///
  typedef GRelationIterator<T>             iterator;
  ///
  typedef ConstGRelationIterator<T>        const_iterator;
  ///
  typedef ReverseGRelationIterator<T>      reverse_iterator;
  ///
  typedef ConstReverseGRelationIterator<T> const_reverse_iterator;

  /** @name --- Constructors --- */
  ///
  GRelationList();
  ///
  GRelationList(const GRelationList<T>& rhs);
  ///
  virtual ~GRelationList();

  /** @name --- Access functions --- */
  ///
  inline void push_back(const Gvalue_type& rel) { rellist.push_back(const_cast<T*>(&rel)); }
  ///
  inline Gpointer back() const { return rellist.back(); }
  ///
  inline int  size() const { return rellist.size(); }
  ///
  inline bool empty() const { return rellist.empty(); }

  /** @name --- Iterator methods --- */
  ///
  inline               iterator begin()        { return iterator(rellist.begin()); }
  ///
  inline               iterator end()          { return iterator(rellist.end()); }
  ///
  inline         const_iterator begin()  const { return const_iterator(rellist.begin()); }
  ///
  inline         const_iterator end()    const { return const_iterator(rellist.end()); }
  ///
  inline       reverse_iterator rbegin()       { return reverse_iterator(rellist.rbegin()); }
  ///
  inline       reverse_iterator rend()         { return reverse_iterator(rellist.rend()); }
  ///
  inline const_reverse_iterator rbegin() const { return const_reverse_iterator(rellist.rbegin()); }
  ///
  inline const_reverse_iterator rend()   const { return const_reverse_iterator(rellist.rend()); }

  /// direct access
  Gpointer operator[](unsigned int i) const;

  /** @name --- Operators --- */
  ///
  const GRelationList<T>& operator=(const GRelationList<T>& rhs);
  ///
  inline bool operator==(const GRelationList<T>& rhs) const {
    return rellist == rhs.rellist;
  }

  /** @name --- Expert functions --- */
  ///
  void print(std::ostream& os) const;
  /// get rid of relation
  inline void unbook(const Gvalue_type& rel) { rellist.remove(const_cast<T*>(&rel)); }
  /// remove all relations without unbooking them
  inline void clear(void) { rellist.clear(); }
  ///
  template <class L>
  void unbook_all(const L& lhs) const {
    const GRelation_cit end = rellist.end();
    for(GRelation_cit it = rellist.begin(); it != end; ++it) {
      value_type rhs = *it;
      
      // container with L's in R:
      GRelationList<typename GR_trait<L>::val>& lhs_cont = rhs->container(lhs);
      lhs_cont.unbook(lhs);
    }
  }

protected:
  GRelation_v  rellist;
}; // class GRelationList

//==============================================================================
// operator<<
//==============================================================================
template<class T>
inline std::ostream& operator<<(std::ostream& s, const GRelationList<T>& a) {
  a.print(s);
  return s;
}


/** Forward Relation Iterator.
    This is a STL compatible iterator for the GRelationList container.
    The iterator works on the container which is specified in the
    \texttt{G\_trait} structure.

    @memo Iterator.
    @author T. Glebe
*/
//============================================================================
// Class GRelationIterator
//============================================================================
template <class T>
class GRelationIterator {
  
  typedef GRelationIterator<T> self;
#ifndef __CINT__
  typedef typename G_trait<T>::GRelation_it        GRelation_it;
  typedef typename G_trait<T>::Gpointer            Gpointer;
  typedef typename G_trait<T>::Greference          Greference;
  typedef typename GRelation_it::iterator_category iterator_category;
  typedef typename GRelation_it::value_type        value_type;
  typedef typename GRelation_it::pointer           pointer;
  typedef typename GRelation_it::reference         reference;
  typedef typename GRelation_it::size_type         size_type;
  typedef typename GRelation_it::difference_type   difference_type;
#else
  typedef G_trait<T>::GRelation_it        GRelation_it;
  typedef G_trait<T>::Gpointer            Gpointer;
  typedef G_trait<T>::Greference          Greference;
  typedef GRelation_it::value_type        value_type;
  typedef GRelation_it::pointer           pointer;
  typedef GRelation_it::reference         reference;
  typedef GRelation_it::size_type         size_type;
  typedef GRelation_it::difference_type   difference_type;
#endif

public:
  /** @name --- Constructors --- */
  ///
  GRelationIterator() {};
  ///
  GRelationIterator(const GRelation_it& it):iter(it) {}
  ///
  virtual ~GRelationIterator() {};

  /** @name --- Operators --- */
  ///
  inline self&  operator++() { ++iter; return *this; }
  ///
  inline self&  operator--() { --iter; return *this; }
  ///
  inline Gpointer  operator*()  const { return *iter; }
  ///
  inline Gpointer  operator->() const { return *iter; }
  ///
  inline bool operator==(const GRelationIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const GRelationIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

protected:
  GRelation_it iter;
}; // class GRelationIterator


/** Constant Forward Relation Iterator
    This is a STL compatible iterator for the GRelationList container.
    The values returned by this iterator are constant and cannot be modified
    (read-only).
    The iterator works on the container which is specified in the
    \texttt{G\_trait} structure.

    @memo Iterator
    @author T. Glebe
*/
//============================================================================
// Class ConstGRelationIterator
//============================================================================
template <class T>
class ConstGRelationIterator {

  typedef ConstGRelationIterator<T> self;
#ifndef __CINT__
  typedef typename G_trait<T>::GRelation_cit        GRelation_cit;
  typedef typename G_trait<T>::Gconst_pointer       Gconst_pointer;
  typedef typename G_trait<T>::Gconst_reference     Gconst_reference;
  typedef typename GRelation_cit::iterator_category iterator_category;
  typedef typename GRelation_cit::value_type        value_type;
  typedef typename GRelation_cit::pointer           pointer;
  typedef typename GRelation_cit::reference         reference;
  typedef typename GRelation_cit::size_type         size_type;
  typedef typename GRelation_cit::difference_type   difference_type;
#else
  typedef G_trait<T>::GRelation_cit        GRelation_cit;
  typedef G_trait<T>::Gconst_pointer       Gconst_pointer;
  typedef G_trait<T>::Gconst_reference     Gconst_reference;
  typedef GRelation_cit::value_type        value_type;
  typedef GRelation_cit::pointer           pointer;
  typedef GRelation_cit::reference         reference;
  typedef GRelation_cit::size_type         size_type;
  typedef GRelation_cit::difference_type   difference_type;
#endif

public:
  /** @name --- Constructors --- */
  ///
  ConstGRelationIterator() {};
  ///
  ConstGRelationIterator(const GRelation_cit& it):iter(it) {}
  ///
  virtual ~ConstGRelationIterator() {};

  /** @name --- Operators --- */
  ///
  inline self&          operator++() { ++iter; return *this; }
  ///
  inline self&          operator--() { --iter; return *this; }
  ///
  inline Gconst_pointer     operator*()  const { return *iter; }
  ///
  inline const Gconst_pointer operator->() const { return *iter; }
  ///
  inline bool operator==(const ConstGRelationIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const ConstGRelationIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

protected:
  GRelation_cit iter;
}; // class ConstGRelationIterator


/** Reverse Relation Iterator.
    This is a STL compatible iterator for the GRelationList container.
    The iterator traverse the GRelationList container in inverse order as
    the GRelationIterator.
    The iterator works on the container which is specified in the
    \texttt{G\_trait} structure.

    @memo Iterator
    @author T. Glebe
*/
//============================================================================
// Class ReverseGRelationIterator
//============================================================================
template <class T>
class ReverseGRelationIterator {
  
  typedef ReverseGRelationIterator<T> self;
#ifndef __CINT__
  typedef typename G_trait<T>::GRelation_rit        GRelation_rit;
  typedef typename G_trait<T>::Gpointer             Gpointer;
  typedef typename G_trait<T>::Greference           Greference;
  typedef typename GRelation_rit::iterator_category iterator_category;
  typedef typename GRelation_rit::value_type        value_type;
  typedef typename GRelation_rit::pointer           pointer;
  typedef typename GRelation_rit::reference         reference;
  typedef typename GRelation_rit::difference_type   difference_type;
#else
  typedef G_trait<T>::GRelation_rit        GRelation_rit;
  typedef G_trait<T>::Gpointer             Gpointer;
  typedef G_trait<T>::Greference           Greference;
  typedef GRelation_rit::value_type        value_type;
  typedef GRelation_rit::pointer           pointer;
  typedef GRelation_rit::reference         reference;
  typedef GRelation_rit::difference_type   difference_type;
#endif

public:
  /** @name --- Constructors --- */
  ///
  ReverseGRelationIterator() {};
  ///
  ReverseGRelationIterator(const GRelation_rit& it):iter(it) {}
  ///
  virtual ~ReverseGRelationIterator() {};
  
  /** @name --- Operators --- */
  ///
  inline self&  operator++() { ++iter; return *this; }
  ///
  inline self&  operator--() { --iter; return *this; }
  ///
  inline Gpointer operator*()  const { return *iter; }
  ///
  inline Gpointer operator->() const { return *iter; }
  ///
  inline bool operator==(const ReverseGRelationIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const ReverseGRelationIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

protected:
  GRelation_rit iter;
}; // class ReverseGRelationIterator


/** Constant Reverse Relation Iterator.
    This is a STL compatible iterator for the GRelationList container.
    The iterator traverse the GRelationList container in inverse order as
    the ConstGRelationIterator. The values returned by this iterator
    are constant and cannot be modified (read-only).
    The iterator works on the container which is specified in the
    \texttt{G\_trait} structure.

    @memo Iterator
    @author T. Glebe
*/
//============================================================================
// Class ConstReverseGRelationIterator
//============================================================================
template <class T>
class ConstReverseGRelationIterator {
  
  typedef ConstReverseGRelationIterator<T> self;
#ifndef __CINT__
  typedef typename G_trait<T>::GRelation_rcit        GRelation_rcit;
  typedef typename G_trait<T>::Gconst_pointer        Gconst_pointer;
  typedef typename G_trait<T>::Gconst_reference      Gconst_reference;
  typedef typename GRelation_rcit::iterator_category iterator_category;
  typedef typename GRelation_rcit::value_type        value_type;
  typedef typename GRelation_rcit::pointer           pointer;
  typedef typename GRelation_rcit::reference         reference;
  typedef typename GRelation_rcit::difference_type   difference_type;
#else
  typedef G_trait<T>::GRelation_rcit        GRelation_rcit;
  typedef G_trait<T>::Gconst_pointer        Gconst_pointer;
  typedef G_trait<T>::Gconst_reference      Gconst_reference;
  typedef GRelation_rcit::value_type        value_type;
  typedef GRelation_rcit::pointer           pointer;
  typedef GRelation_rcit::reference         reference;
  typedef GRelation_rcit::difference_type   difference_type;
#endif

public:
  /** @name --- Constructors --- */
  ///
  ConstReverseGRelationIterator() {};
  ///
  ConstReverseGRelationIterator(const GRelation_rcit& it):iter(it) {}
  ///
  virtual ~ConstReverseGRelationIterator() {};

  /** @name --- Operators --- */
  ///
  inline self&          operator++() { ++iter; return *this; }
  ///
  inline self&          operator--() { --iter; return *this; }
  ///
  inline Gconst_pointer operator*()  const { return *iter; }
  ///
  inline const Gconst_pointer  operator->() const { return *iter; }
  ///
  inline bool operator==(const ConstReverseGRelationIterator<T>& rhs) const {
    return iter == rhs.iter;
  }
  ///
  inline bool operator!=(const ConstReverseGRelationIterator<T>& rhs) const {
    return iter != rhs.iter;
  }

protected:
  GRelation_rcit iter;
}; // class ConstReverseGRelationIterator

//============================================================================
// load the implementation file
//============================================================================
#include "GRelationList.icc"
#endif
