#ifndef __TRACKIF_HH
#define __TRACKIF_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Track object interface
//
// changes:
// 27 Feb 2001 (TG) creation
// 06 Apr 2001 (TG) added x(z), y(z)
// 07 Apr 2001 (TG) added cov_x(), cov_y()
// 10 Apr 2001 (TG) size_trait, CINV_mom(), CINV_nomom() added
// 17 Apr 2001 (TG) CINV_nomom removed, renamed CINV_mom() to CINV()
// 27 Apr 2001 (TG) pt(), xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 11 Jun 2001 (TG) added COV(), cov_tx(), cov_ty(), cov_p()
// 28 Jun 2001 (TG) added charge(), pz(), energy(), xf(), rap(), removed xfabs()
// 05 Jul 2001 (TG) collect() added
// 09 Jul 2001 (TG) added isValid(), valid(), invalid()
// 10 Jul 2001 (TG) operator==() added
// 11 Jul 2001 (TG) removed xfabs comments
// 28 Aug 2001 (TG) added documentation of SPMember objects
// 12 Nov 2001 (TG) comments improved
// 09 Jan 2002 (TG) changed return type of propagate() to bool
//
// ********************************************************************
#include <iosfwd>
#include <vector>

template <class T, unsigned int D> class SVector;
#ifndef __CINT__
#include "smatrix/SMatrix.hh"
#endif

/** Abstract class which defines the minimal interface for a Track object.

    @memo Track Interface
    @author T. Glebe
*/
//==============================================================================
// TrackIf
//==============================================================================
class TrackIf {
public:
  /** @name --- Access methods --- */
  /// Track $x$ position
  virtual float x()   const = 0;
  /// Track $y$ position
  virtual float y()   const = 0;
  /// Track $x$ position at $z$
  virtual float x(float z)   const = 0;
  /// Track $y$ position at $z$
  virtual float y(float z)   const = 0;
  /// Track $z$ position
  virtual float z()   const = 0;
  /// Track slope $t_x$
  virtual float tx()  const = 0;
  /// Track slope $t_y$
  virtual float ty()  const = 0;
  /// Track momentum $p$
  virtual float p()   const = 0;
  /// transverse Track momentum $p_t$ ($\perp$ to $z$)
  virtual float pt()  const = 0;
  /// Track momentum along $z$
  virtual float pz()  const = 0;
  /// $\chi^2$ of track fit
  virtual float chi2() const = 0;
  /// degrees of freedom of track fit
  virtual unsigned short int ndf()  const = 0;
  /// azimuthal angle $\phi$ [deg]
  virtual float phi() const = 0;
  /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
  virtual float theta() const = 0;
  /// pseudorapidity\\ $\eta = -\log\tan(\theta/2.)$
  virtual float eta() const = 0;
  /// track charge
  virtual int charge() const = 0;

  /// track energy $E=\sqrt{m_0^2 + \vec{p}^2}$
  virtual float energy(double rm = 0.) const = 0;
  /// Feynman variable\\ $x_F = \frac{E+p_z}{(E+p_z)_{max}}$
  virtual float xf(double rm = 0.) const = 0;
  /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
  virtual float rap(double rm = 0.) const = 0;

#ifndef __CINT__
  /// $\vec{x} = (x,y,z)$
  virtual SVector<double,3> xvec() const = 0;
  /// slope vector $\vec{t} = (t_x,t_y,1)$
  virtual SVector<double,3> tvec() const = 0;
  /// $\vec{e} = (e_x,e_y,ez)$ unit vector along $\vec{p}$
  virtual SVector<double,3> evec() const = 0;
  /// momentum $\vec{p} = (p_x,p_y,p_z)$
  virtual SVector<double,3> pvec() const = 0;
  /** covariance matrix
      \begin{displaymath}
      C = \left(
      \begin{array}{ccccc}
      c_{x,x} & c_{x,y} & c_{x,t_x} &
      c_{x,t_y} & c_{x,p} \\
      c_{y,x} & c_{y,y} & c_{y,t_x} &
      c_{y,t_y} & c_{y,p} \\
      c_{t_x,x} & c_{t_x,y} & c_{t_x,t_x} &
      c_{t_x,t_y} & c_{t_x,p} \\
      c_{t_y,x} & c_{t_y,y} & c_{t_y,t_x} &
      c_{t_y,t_y} & c_{t_y,p} \\
      c_{p,x} & c_{p,y} & c_{p,t_x} &
      c_{p,t_y} & c_{p,p}
      \end{array}\right)
      \end{displaymath}
   */
  virtual const SMatrix<double,5>& COV() const = 0;
  /// inverse covariance matrix $C^{-1}$
  virtual const SMatrix<double,5>& CINV() const = 0;
#endif

  /// return $\sigma_x^2|_{z+dz}$
  virtual float cov_x(double dz=0.)        const = 0;
  /// return $\sigma_y^2|_{z+dz}$
  virtual float cov_y(double dz=0.)        const = 0;
  /// return $\sigma_{t_x}^2$
  virtual float cov_tx()                   const = 0;
  /// return $\sigma_{t_y}^2$
  virtual float cov_ty()                   const = 0;
  /// return $\sigma_p^2$
  virtual float cov_p()                    const = 0;

  /// is track valid?
  virtual bool isValid() const = 0;

  /** @name --- Operators --- */
  /// comparison between Track objects
  virtual bool operator==(const TrackIf& rhs) const = 0;

  /** @name --- Expert methods --- */
  /// mark track as valid
  virtual void valid() = 0;
  /// mark track as invalid
  virtual void invalid() = 0;
  /// collect TrackIf pointers
  virtual void collect(vector<TrackIf*>& c) const = 0;
  /// propagate track parameters \& covariance matrix to $z$
  virtual bool propagate(const double z) = 0;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const = 0;
}; // end of class TrackIf


//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<< (std::ostream& os, const TrackIf& t) {
  return t.print(os);
}


//==============================================================================
// TrackIf Pattern elements
//==============================================================================
template <class C, class T> class SPMember;
template <class C, class T> class SPConstant;
template <class C, class T, class P> class SPMemberOnePar;

/**@name TrackIf Property objects.

   These are global objects which can be used in cut expressions for
   PatternMatch and DecayPatternMatch objects.
 */
//@{
/// access to TrackIf::x()
extern const SPMember<TrackIf,float> x;
/// access to TrackIf::y()
extern const SPMember<TrackIf,float> y;
/// access to TrackIf::z()
extern const SPMember<TrackIf,float> z;

/// access to TrackIf::tx()
extern const SPMember<TrackIf,float> tx;
/// access to TrackIf::ty()
extern const SPMember<TrackIf,float> ty;
/// access to TrackIf::p()
extern const SPMember<TrackIf,float> p;
/// access to TrackIf::pt()
extern const SPMember<TrackIf,float> pt;
/// access to TrackIf::pz()
extern const SPMember<TrackIf,float> pz;
/// access to TrackIf::chi2()
extern const SPMember<TrackIf,float> chi2;
/// access to TrackIf::ndf()
extern const SPMember<TrackIf,unsigned short int> ndf;

/// access to TrackIf::phi()
extern const SPMember<TrackIf,float> azim;
/// access to TrackIf::theta()
extern const SPMember<TrackIf,float> theta;
/// access to TrackIf::eta()
extern const SPMember<TrackIf,float> psrap;

/// access to TrackIf::charge()
extern const SPMember<TrackIf,int>   charge;

/// access to TrackIf::energy()
extern const SPMemberOnePar<TrackIf,float,double> energy;
/// access to TrackIf::xf()
extern const SPMemberOnePar<TrackIf,float,double> xf;
/// access to TrackIf::rap()
extern const SPMemberOnePar<TrackIf,float,double> rap;

/// always \texttt{true}
extern const SPConstant<TrackIf,bool> all;
/// access to TrackIf::isValid()
extern const SPMember<TrackIf,bool>   isValid;
//@}
#endif
