#ifndef __VERTEXIF_HH
#define __VERTEXIF_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Feb 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Vertex object interface
//
// changes:
// 27 Feb 2001 (TG) creation
// 24 Apr 2001 (TG) valid(), invalid() added
// 13 Jun 2001 (TG) added VCOV(), vtx_cov_x(), vtx_cov_y(), vtx_cov_z()
// 09 Jul 2001 (TG) valid(), invalid() moved to TrackIf
// 26 Jul 2001 (TG) added track()
// 28 Aug 2001 (TG) added documentation of SPMember objects
// 12 Nov 2001 (TG) comments improved
//
// ********************************************************************

template <class T, unsigned int D> class SVector;
class TrackIf;

#ifndef __CINT__
#include "smatrix/SMatrix.hh"
#endif

/** Abstract class which defines the minimal interface for a Vertex object.

    @memo Vertex Interface
    @author T. Glebe
*/
//==============================================================================
// VertexIf
//==============================================================================
class VertexIf {
public:
  /** @name --- Access methods --- */
  /// $v_x =$ vertex $x$ position
  virtual float vx()   const = 0;
  /// $v_y =$ vertex $y$ position
  virtual float vy()   const = 0;
  /// $v_z =$ vertex $z$ position
  virtual float vz()   const = 0;
  /// $\chi^2$ of vertex fit
  virtual float chi2() const = 0;
  /// upper tail $\chi^2$ probability
  virtual float prob() const = 0;
  /// degrees of freedom of vertex fit
  virtual unsigned short int ndf()  const = 0;
  /// no. of tracks in vertex
  virtual unsigned short int ntracks() const = 0;
  /// return a daughter track pointer
  virtual const TrackIf* track(unsigned int i) const = 0;
  /// read/write track access
  virtual const TrackIf*& track(unsigned int i) = 0;

#ifndef __CINT__
  /// vertex position $\vec{v} = (v_x,v_y,v_z)$
  virtual SVector<double,3> vpos() const = 0;
  /** vertex covariance matrix
      \begin{displaymath}
      V = \left(
      \begin{array}{ccc}
      v_{x,x} & v_{x,y} & v_{x,z}\\
      v_{y,x} & v_{y,y} & v_{y,z}\\
      v_{z,x} & v_{z,y} & v_{z,z}\\
      \end{array}\right)
      \end{displaymath}
   */
  virtual const SMatrix<double,3>& VCOV() const = 0;
#endif

  /// $\sigma_x^2$ of vertex
  virtual float vtx_cov_x()   const = 0;
  /// $\sigma_y^2$ of vertex
  virtual float vtx_cov_y()   const = 0;
  /// $\sigma_z^2$ of vertex
  virtual float vtx_cov_z()   const = 0;

  /** @name --- Expert methods --- */
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const = 0;
}; // end of class VertexIf


//==============================================================================
// operator <<
//==============================================================================
inline std::ostream& operator<< (std::ostream& os, const VertexIf& t) {
  return t.print(os);
}


//==============================================================================
// VertexIf Pattern elements
//==============================================================================
template <class C, class T> class SPMember;
template <class C, class T> class SPConstant;
template <class C, class T, class P> class SPMemberOnePar;

/**@name VertexIf Property objects.

   These are global objects which can be used in cut expressions for
   DecayPatternMatch objects.
 */
//@{
/// access to VertexIf::vx()
extern const SPMember<VertexIf,float> vx;
/// access to VertexIf::vy()
extern const SPMember<VertexIf,float> vy;
/// access to VertexIf::vz()
extern const SPMember<VertexIf,float> vz;

/// access to VertexIf::chi2()
extern const SPMember<VertexIf,float> vchi2;
/// access to VertexIf::prob()
extern const SPMember<VertexIf,float> prob;

/// access to VertexIf::ndf()
extern const SPMember<VertexIf,unsigned short int> vndf;
//@}
#endif
