#!/bin/bash
##########################################################
#
# type:      shell script
#
# created:   29. Mar 2001
#
# author:    Thorsten Glebe
#            HERA-B Collaboration
#            Max-Planck-Institut fuer Kernphysik
#            Saupfercheckweg 1
#            69117 Heidelberg
#            Germany
#            E-mail: T.Glebe@mpi-hd.mpg.de
#
# Description: script to generate the skeleton of an analysis class
#
##########################################################

PROG=`basename $0`
BEEPATH=`dirname $0`
HELP="usage: $PROG <class name>"

if [ $# -ne 1 ]; then
    echo $HELP
    exit 1;
else
    PROJ=$1
fi

HHFILE=${PROJ}.hh
CCFILE=${PROJ}.C
DATE=`date`

if [ ! -d include ]; then
    echo "create include directory"
    mkdir include
    if [ $? -ne 0 ]; then 
      echo "cannot create include dir!"
      exit 1
    else
      HHFILE=include/${HHFILE}
    fi
else
  HHFILE=include/${HHFILE}
fi

echo "Create ${HHFILE}..."
#
# create header file
# 
(
echo "\
#ifndef __${PROJ}_HH
#define __${PROJ}_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   $DATE
//
// author:    $USER
//
// Description: 
//
// changes:
//
// ********************************************************************

#include \"interfaces/AnalysisIf.hh\"

class TH1F;
class TH2F;
class TNtuple;
class RtEvent;

//==============================================================================
// $PROJ
//==============================================================================
class $PROJ : public AnalysisIf {
public:
  /** @name --- Constructors --- */
  ///
  ${PROJ}(RtEvent* evt);
  ///
  virtual ~${PROJ}() {}

  /** @name --- Access methods --- */
  ///
  bool EventSelect();
  ///
  bool DefineHistograms();
  ///
  bool DefinePattern();
  ///
  bool Event();
  ///
  bool EventCleanup();
  ///
  bool End();
  ///
  inline const char *name() const { return \"$PROJ\"; }

private:
  RtEvent* event;
}; // end of $PROJ
#endif
"
) > $HHFILE

#
# create .C file
#
echo "Create ${CCFILE}..."
(
echo "\
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   $DATE
//
// author:    $USER
//
// Description: 
//
// changes:
//
// ********************************************************************
#include <iostream>
#include \"TH1.h\"
#include \"TH2.h\"
#include \"TNtuple.h\"
#include \"TPaveText.h\"
#include \"BEE.hh\"
#include \"$HHFILE\"

//==============================================================================
// Constructor
//==============================================================================
$PROJ::$PROJ(RtEvent* evt) :
 event(evt)
{}

//==============================================================================
// EventSelect
//==============================================================================
bool $PROJ::EventSelect() {
  return true;
}

//==============================================================================
// DefinePattern
//==============================================================================
bool $PROJ::DefinePattern() {
  return true;
}

//==============================================================================
// DefineHistograms
//==============================================================================
bool $PROJ::DefineHistograms() {
  return true;
}

//==============================================================================
// Event
//==============================================================================
bool $PROJ::Event() {
  return true;
}

//==============================================================================
// EventCleanup
//==============================================================================
bool $PROJ::EventCleanup() {
  return true;
}

//==============================================================================
// End processing
//==============================================================================
bool $PROJ::End() {
  return true;
}
"
) > $CCFILE

#
# create main.C file
#
if [ ! -f main.C ]; then
  echo "create main.C..."
(
echo "\
// *******************************************************************
//
// source:
//
// type:      source code
//
// created:   01. Mar 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Main routine of a root based BEE job.
//              The AnalysisChain is used to queue user defined 
//              analysis objects.
//
// changes:
// 01 Mar 2001 (TG) creation
// 30 Mar 2001 (TG) command line processing added
// 30 Mar 2001 (TG) load_files(), basename() added
// 30 Apr 2001 (TG) added offset
// 12 Jun 2001 (TG) added update
// 10 Jul 2001 (TG) load_files: check if file exists
// 16 Jul 2001 (TG) added event->Clear() (bug fix for ROOT version < 3.00)
//
// ********************************************************************
#include <iostream>
#include <iomanip>
#include <string>
#include <getopt.h>
#include <signal.h>

#include \"TROOT.h\"
#include \"TChain.h\"
#include \"TFile.h\"

#include \"BEE.hh\"
#include \"${PROJ}.hh\"

bool load_files(TChain* f, int argc, char* const* argv);   // see below
string basename(const char* str);                          // see below
TFile* outputfile(const string& file, const string& mode); // see below
void   main_end();   // see below

RtEvent* event = 0;
static int myoptind = 0;

const string MSGstring(\"======> \");
static TFile* histos = 0;
static AnalysisChain Analysis;   // analysis chain object

void sig_int(int);   // CTRL-c signal handler

//==============================================================================
// main
//==============================================================================
int main(int argc, char* const* argv) {

  // install signal handler
  if( signal(SIGINT, sig_int) == SIG_ERR ) {
    cerr << MSGstring << \"Warning: Could not install signal handler\" << endl;
  }

  string pname = basename(argv[0]);
  string outfile(\"first_analysis.root\");
  string fileopt(\"recreate\");
  unsigned int offset = 0;
  int Eventno = -1;

  // ================ Command Line parsing  =====================
  unsigned int nentr = 0;
  int c;
  while((c = getopt(argc, argv, \"?hn:s:o:e:u\")) != -1) {
    switch(c) {
    case 'n':
      nentr = atoi(optarg);
      break;
    case 's':
      offset = atoi(optarg);
      break;
    case 'o':
      outfile = optarg;
      break;
    case 'e':
      Eventno = atoi(optarg);
      break;
    case 'u':
      fileopt = \"update\";
      break;
    case 'h': // FALL_THROUGH
    case '?':
      cerr << endl
	   << \"Usage: \" << pname << \" [-n <events>] [-s <offset>] [-e event] [-o output file] [-u] files...\" << endl
	   << \" <events>   : number of events to process (default: all)\" << endl
	   << \" <offset>   : event number to start with (default: 0)\" << endl
	   << \" files      : BEE files to process (wildcards are allowed)\" << endl
	   << \" output file: ROOT histogram output file (default: \" 
	   << outfile << \")\" << endl
	   << \"-u          : open histogram output file in UPDATE mode (default: \" << fileopt << \")\" << endl
	   << \" event      : experiment event number of an event under investigation\" << endl
	   << \"Example:\" << endl
	   << \"   \" << pname << \" -n 5000 /scratch/BEE/*.root -o out.root\" << endl
	   << endl << \"Typing CTRL-c: quit event processing\"
           << endl;
      exit(0);
      break;
    default:
      cerr << \" option \" << static_cast<char>(c) << \" not known! \" << endl;
      break;
    }
  }

  // optind will be resetted by ROOT initialization
  myoptind = optind; 

  // ========================== ROOT Setup =======================
  TROOT minexam(\"BEE\",\"A BEE offline job\");

  // Define Histogram Output file
  histos = outputfile(outfile, fileopt);
  //  TFile* histos = new TFile(outfile.c_str(), fileopt.c_str());

  event = new RtEvent();
  
  TChain* f = new TChain(\"Clue\");
  cout << MSGstring << \"Set up event branch.\" << endl;
  f->SetBranchAddress(\"event\",&event);

  // load file via command line
  if(load_files(f,argc,argv) == false) {
    cerr << MSGstring << \"Error: No data file given!\" << endl;
    exit(0);
  }

  f->GetEntry(offset);
  const Stat_t NEVENTS = f->GetEntries();

  // ======================= Define Analysis objects =============

  // instantiate analysis
  ${PROJ}  AnaJob(event);
  Analysis.push_back(AnaJob);
  
  Analysis.DefinePattern();
  Analysis.DefineHistograms();

  // ========================== Event loop ========================
  const unsigned int nentries = static_cast<unsigned int>((nentr==0) ? NEVENTS - offset : nentr);

  if(Eventno > 0) {
    // SEARCH FOR A SINGLE EVENT
    unsigned int i = offset;
    cout << \"Try to find Event \" << Eventno << \" starting at \" << i;
    cout.flush();
    do {
      event->Clear(); // bug fix for ROOT versions < 3.00
      f->GetEntry(i);
      if(i%1000 == 0) { cout << \".\"; cout.flush(); }
      ++i;
    } while(event->evt() != Eventno && i < NEVENTS);
    cout << endl;
    if(i == NEVENTS) {
      cout << \"Couldn't find event \" << Eventno << endl;
    } else {
      cout << \" Event \" << Eventno << \" found! (TTree entry no \" << i << \")\" << endl;
      Analysis.Event();
      Analysis.EventCleanup();
    }
  } else {
    // STANDARD EVENT PROCESSING
    for(unsigned int i = offset; i<nentries; ++i) {
      event->Clear(); // bug fix for ROOT versions < 3.00
      f->GetEntry(i);
      if(i%100 == 0)
	cout << \" Event \" << i 
	     << \", events to go: \" << nentries - i << endl;
      
      Analysis.Event();
      Analysis.EventCleanup();
    } // for i
  }

  // =========================== The End ==========================
  main_end();

  return 0;
}

//==============================================================================
// main_end: end of Analysis job
//==============================================================================
void main_end() {
  Analysis.End();

  histos->Write();
  histos->Close();
}

#include <sys/stat.h>
//==============================================================================
// load_files: return true if a file was loaded, else false
//==============================================================================
bool load_files(TChain* f, int argc, char* const* argv) {

  struct stat buf;

  unsigned int count = 0;
  if (myoptind < argc) {
    while (myoptind < argc) {
      const char* file = argv[myoptind];

      // file exists
      if(stat(file,&buf) == 0) {
	cout << MSGstring << \"Add file \" << file << endl;
	f->Add(file);
	++count;
      } else {
	cerr << MSGstring << \"Warning: Could not access file \" << file << endl;
      }
      ++myoptind;
    }
  }
  return (count>0) ? true : false;
}

//==============================================================================
// basename
//==============================================================================
string basename(const char* str) {

  const string prg(str);
  string::size_type pos = prg.rfind('/');

  if(pos != string::npos) {
    return prg.substr(pos+1);
  } else {
    return prg;
  }
}

//==============================================================================
// open output file
//==============================================================================
TFile* outputfile(const string& filename, const string& mode) {
  const char* file = filename.c_str();
  const string oldstr = filename + string(\"_old\");

  // check if file exists
  struct stat buf;
  
  // file exists
  if(stat(file,&buf) == 0) {
    cerr << MSGstring << \"File \" << file << \" exists, rename it to \" << oldstr << endl;
    if(rename(file, oldstr.c_str()) != 0) {
      cerr << MSGstring << \"Error in renaming \" << file << \" to \" << oldstr << endl;
    }
  }

  cout << MSGstring << \"The output file is: \" << file << endl;

  return new TFile(file, mode.c_str());
}

//==============================================================================
// sig_int: signal hander for user exit
//==============================================================================
void sig_int(int signo) {
  cout << MSGstring << \"End of event processing due to user interrupt\" << endl;
  main_end();
  exit(0);
}
"
) > main.C
fi
exit 0;