#ifndef __SDecayPatternMatch_HH
#define __SDecayPatternMatch_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   23. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class for handling matched vertices.
//
// changes:
// 23 Apr 2001 (TG) creation
// 24 Apr 2001 (TG) valid(), invalid(), TrackIf members added
// 25 Apr 2001 (TG) added prob()
// 27 Apr 2001 (TG) pt(), xfabs() added
// 02 Mai 2001 (TG) phi(), theta(), eta() added
// 02 Mai 2001 (TG) operator==() added
// 07 Mai 2001 (TG) vertex() added
// 18 Mai 2001 (TG) set_user(), user(), user_size(), user_, N_ added
// 06 Jun 2001 (TG) added inheritance from DecayPatternMatchIf, added massProb()
// 11 Jun 2001 (TG) added massError(), cov_tx(), cov_ty(), cov_p()
// 13 Jun 2001 (TG) added vtx_cov_x(), vtx_cov_y(), vtx_cov_z()
// 28 Jun 2001 (TG) removed template parameter C and T
// 29 Jun 2001 (TG) added charge(), pz(), energy(), xf(), rap(), removed xfabs()
// 05 Jul 2001 (TG) collect() added
// 06 Jul 2001 (TG) added scale arg to massProb()
// 09 Jul 2001 (TG) bug fix: moved SPatternMatch-Relation to DecayPatternMatchIf
// 09 Jul 2001 (TG) valid(), invalid() removed (now part of TrackIf), added ntracks
// 25 Jul 2001 (TG) added track()
// 17 Aug 2001 (TG) CTOR added
// 20 Aug 2001 (TG) moved user field + access functions to SPatternMatch
// 21 Aug 2001 (TG) inheritance from new DecayPatternMatchIf class (located in
//                  interfaces package), added getTrackIf, added PatternMatch
//                  relations (out), removed maxImpact()
// 22 Aug 2001 (TG) changed stored pattern to SPattern
// 24 Aug 2001 (TG) added mcmother_
// 27 Aug 2001 (TG) CTOR added
// 28 Aug 2001 (TG) ismc_ added, CTOR added, set_mother(), set_MC(),
//                  PatternMatch(i) added, getRecoTrack(), getMC_Track() added
// 29 Aug 2001 (TG) added operator==(GTopp)
// 05 Sep 2001 (TG) added user_, N_, set_user(), user_size(), user() members
// 06 Sep 2001 (TG) removed getTrackIf() (obsolete)
// 19 Dec 2001 (TG) added Vtx() relation
// 02 Jan 2002 (TG) changed type of out() relation to PatternMatchIf
// 07 Jan 2002 (TG) added get() member
// 09 Jan 2002 (TG) changed return type of propagate() to bool
// 11 Jan 2002 (TG) copy constructor added, operator=() added
// 17 Jan 2002 (TG) added vertex() const
//
// ********************************************************************
#include <iosfwd>
#include "interfaces/DecayPatternMatchIf.hh"
#include "grelation/GRelationList.hh"

template <unsigned int NTR> class SDecayPattern;
template <unsigned int NTR> class SVertex;

/** Pattern Match class for Decays.
    SDecayPatternMatch represents a SVertex vertex object together with the
    DecayPattern which was used to select the vertex object.

    @memo Matched decays.
    @author T. Glebe
*/
//==============================================================================
// SDecayPatternMatch
//==============================================================================
template <unsigned int NTR>
class SDecayPatternMatch : public DecayPatternMatchIf {
public:
  static const unsigned int size = NTR;

  /** @name --- Constructors --- */
  ///
  SDecayPatternMatch();
  /// copy CTOR
  SDecayPatternMatch(const SDecayPatternMatch<NTR>& rhs);
  /// create from MC
  SDecayPatternMatch(const MC_Track& t,
		     const SmartPointer<SPattern>& vpat);
  /// create from reconstructed track
  SDecayPatternMatch(const TrackIf& t,
		     const SmartPointer<SPattern>& vpat);
  /// create from Vertex
  SDecayPatternMatch(SmartPointer<SVertex<NTR> >& v,
		     const SmartPointer<SPattern>& vpat);
  /// take over ownership of vertex object
  SDecayPatternMatch(SVertex<NTR>* v,
		     const SmartPointer<SPattern>& vpat);
  ///
  virtual ~SDecayPatternMatch();

  ///
  SDecayPatternMatch<NTR>& operator=(const SDecayPatternMatch<NTR>& rhs);

  /** @name --- TrackIf access methods ---
      \textbf{Note:} these methods apply to the mother track. Take care that
      the mother track was computed from the vertex information before calling
      one of these functions!
   */
  /// mother track $x$ position
  float x()   const;
  /// mother track $y$ position
  float y()   const;
  /// mother track $x$ position at $z$
  float x(float z)   const;
  /// mother track $y$ position at $z$
  float y(float z)   const;
  /// mother track $z$ position
  float z()   const;
  /// mother track slope $t_x$
  float tx()  const;
  /// mother track slope $t_y$
  float ty()  const;
  /// mother track momentum $p$
  float p()   const;
  /// transverse momentum $p_t$
  float pt()  const;
  /// momentum along $z$
  float pz()  const;
  /// azimuthal angle $\phi$ [deg]
  float phi() const;
  /// polar angle $\theta = \cos^{-1}(e_z)$ [deg]
  float theta() const;
  /// rapidity $\eta = -\log\tan(\theta/2.)$
  float eta() const;
  /// track charge
  int   charge() const;

  /// track energy $E=\sqrt{m^2 + p^2}$
  float energy(double rm = 0.) const;
  /// Feynman-$x$: $x_F = \frac{E+p_z}{(E+p_z)_{max}}$
  float xf(double rm = 0.) const;
  /// Rapidity $y = \frac{1}{2}\ln(\frac{E+p_z}{E-p_z})$
  float rap(double rm = 0.) const;
  
  /// is mother track valid?
  virtual bool isValid() const;
  /// mark mother track as valid
  virtual void valid();
  /// mark mother track as invalid
  virtual void invalid();

#ifndef __CINT__
  /// $\vec{x} = (x,y,z)$
  SVector<double,3> xvec() const;
  /// slope vector $\vec{t} = (t_x,t_y,1)$
  SVector<double,3> tvec() const;
  /// $\vec{e} = (e_x,e_y,e_z)$ unit vector along $\vec{p}$
  SVector<double,3> evec() const;
  /// momentum vector $\vec{p} = (p_x,p_y,p_z)$
  SVector<double,3> pvec() const;
  /// covariance matrix
  const SMatrix<double,5>& COV() const;
  /// inverse covariance matrix
  const SMatrix<double,5>& CINV() const;
#endif

  /// get $\sigma_x^2|_{z+dz}$
  float cov_x(double dz=0.)        const;
  /// get $\sigma_y^2|_{z+dz}$
  float cov_y(double dz=0.)        const;
  /// get $\sigma_{t_x}^2$
  float cov_tx()            const;
  /// get $\sigma_{t_y}^2$
  float cov_ty()            const;
  /// get $\sigma_p^2$
  float cov_p()             const;

  /// collect pointers
  void collect(vector<TrackIf*>& c) const;
  /// propagate track parameters \& covariance matrix to $z$
  bool propagate(const double z);

  /** @name --- PatternMatchIf access methods --- */
  /// get TrackIf part of reconstructed particle
  const TrackIf& getTrackIf() const;
  /// get the mother RecoTrack object (if available)
  const RecoTrack& getRecoTrack() const;
  /// get the mother MC\_Track object (if available)
  const MC_Track& getMC_Track() const;
  /// get the mother track as a RecoSegment object
  const RecoSegment& getRecoSegment() const;
  /// read only pattern access
  const SPattern& getPattern() const;
  /// invariant mass of vertex
  double mass() const;
  /// error on invariant mass
  double massError() const;
  /// calculate a probabilty including the mass
  double massProb(double scale) const;
  /// does this match contain MC tracks?
  bool isMCmatch() const;
  /// calculate mother track + cov. matrix, if possible
  bool calc_mother();
  /// calculate mother track
  bool calc_mother_tr();
  /// calculate mother covariance matrix
  bool calc_mother_cov();

  /** @name --- VertexIf Access methods --- */
  /// $v_x=$ vertex $x$ position
  float vx()   const;
  /// $v_y=$ vertex $y$ position
  float vy()   const;
  /// $v_z=$ vertex $z$ position
  float vz()   const;
  /// $\chi^2$ of vertex fit
  float chi2() const;
  /// degrees of freedom of vertex fit
  unsigned short int ndf()  const;
  /// upper tail $\chi^2$ probability
  float prob() const;
  /// no of tracks in vertex
  unsigned short int ntracks() const;
  /// return i-th track pointer
  const TrackIf* track(unsigned int i) const;
  /// read/write track access
  const TrackIf*& track(unsigned int i);
  
  /// vertex position $\vec{v} = (v_x,v_y,v_z)$
  SVector<double,3> vpos() const;
  /// vertex covariance matrix
  const SMatrix<double,3>& VCOV() const;

  /// $\sigma_x^2$ of vertex
  float vtx_cov_x()   const;
  /// $\sigma_y^2$ of vertex
  float vtx_cov_y()   const;
  /// $\sigma_z^2$ of vertex
  float vtx_cov_z()   const;

  /** @name --- SVertex specific members --- */
  /// vertex position $\vec{v} = (v_x,v_y,v_z)$ (fast access)
  const SVector<double,3>& vposR() const;
  /// inverse vertex cov. matrix
  const SMatrix<double,3>& VCINV() const;
  /// mother track parameters $\vec{m}=(x,y,z,t_x,t_y,p)$
  const SVector<double,6>& mother() const;

  /** @name --- Operators --- */
  /// compare Track member with argument pointers
  bool operator==(const TrackIf& rhs) const;
  /// compare Topp pointer with decay pattern
  bool operator==(const ToppIf& rhs) const;
  /// compare with GTopp
  bool operator==(const GTopp& rhs) const;

  /** @name --- DecayPatternMatchIf access functions --- */
  /// return i-th PatternMatchIf pointer
  const PatternMatchIf* PatternMatch(unsigned int i) const;

  /** @name --- Expert functions --- */
  /// get PatternMatchIf pointer via GTopp object
  const PatternMatchIf* get(const GTopp& rhs) const;
  /// readonly access to SVertex object
  const SmartPointer<SVertex<NTR> > vertex() const;
  /// r/w access to SVertex object
  SmartPointer<SVertex<NTR> > vertex();
  /** Factory function: create mother track as RecoSegment object on
      heap. \textbf{Attention}: the user has to delete the pointer himself! 
      In case this track is again input to a \texttt{TrackPattern} container,
      the SPatternMatch objects can take the ownership of the object and take
      care of the destruction. See the \texttt{Pattern} manual for an example.
  */
  RecoSegment* createMotherSegment() const;
  /// set mother track manually
  void set_mother(const TrackIf& rhs);
  /// set MC flag manually
  void set_MC(bool ismc);
  /// set a user field of given size, return pointer to it
  float* set_user(unsigned int size);
  /// get pointer to user field
  float* user() const;
  /// get size of user field
  unsigned int user_size() const;
  /// used by operator<<()
  std::ostream& print(std::ostream& os) const;

  /** @name --- Relations --- */
  /// DecayPatternMatch $\rightarrow$ PatternMatch (relation to outgoing tracks)
  GRelationList<PatternMatchIf> out; //! not persistent
  ///
  const GRelationList<PatternMatchIf>& Out() const;
  ///
  GRelationList<PatternMatchIf>& Out();
  ///
  GRelationList<PatternMatchIf>& container(const PatternMatchIf& p);

  /** PatternMatch $\rightarrow$ DecayPatternMatch: relation of outgoing
      particles to decay (=mother) objects */
  GRelationList<DecayPatternMatchIf> vtx; //! not persistent
  ///
  const GRelationList<DecayPatternMatchIf>& Vtx() const;
  ///
  GRelationList<DecayPatternMatchIf>& Vtx();
  ///
  GRelationList<DecayPatternMatchIf>& container(const DecayPatternMatchIf& p);
  

private:
  SmartPointer<SVertex<NTR> >     vtx_;      // vertex object
  mutable SmartPointer<SPattern>  vtxpat;    // decay pattern
  const TrackIf*                  mother_;   // a manually putted mother track (if available)
  bool                            ismc_;     // are MC tracks involved?
  float*                          user_;     // pointer to a user array
  unsigned int                    N_;        // size of user array
};


//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<(std::ostream& os,
                                const SDecayPatternMatch<NTR>& rhs) {
  return rhs.print(os);
}

#include "pattern/SDecayPatternMatch.icc"
#endif
