#ifndef __SDECAYPATTERNMATCHCONTAINER_HH
#define __SDECAYPATTERNMATCHCONTAINER_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   30. Apr 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A DecayPatternMatch container class
//
// changes:
// 30 Apr 2001 (TG) creation
// 01 Mai 2001 (TG) select(), do_combine() added
// 03 Mai 2001 (TG) massrange() added
// 18 Mai 2001 (TG) operator&() added, CTOR, pcList(), pmList(), print(),
//                  operator<<() added
// 18 Mai 2001 (TG) added member Comb, prob_collision()
// 01 Jun 2001 (TG) check_coll, do_collision_free_combine(), check_collision() added
// 06 Jun 2001 (TG) removed check_coll_do_collision_free_combine(),
//                  check_collision() because they are obsolete due to a
//                  superior Combiner class (GSCombiner)
// 12 Jun 2001 (TG) small bug fix in operator<<(SmartPointer...)
// 13 Jun 2001 (TG) added massProb_collision(), massProb_bg_collision
// 25 Jun 2001 (TG) moved typedefs to P_trait and DP_trait
// 28 Jun 2001 (TG) removed template parameter C and T
// 02 Jul 2001 (TG) changed ptrlist added, type of pmlist to SPMP_l, pmatches() added
// 03 Jul 2001 (TG) added calc_mother(), calc_mother_tr(), calc_mother_cov()
// 06 Jul 2001 (TG) check_combination() added, added bool arg to combine(), do_combine()
//                  added scale arg to massProb_collision
// 25 Jul 2001 (TG) updated comments
// 22 Aug 2001 (TG) removed pattern, changed return type of dpattern()
// 23 Aug 2001 (TG) changed dpattern_ type to SDecayPattern<NTR>*, added push_back()
// 27 Aug 2001 (TG) added selectRECOfromMCtruth(), added push_back(SVertex)
// 28 Aug 2001 (TG) added selectNORECOfromMCtruth(), push_back(MC_Track),
//                  push_back(SDecayPatternMatch)
// 15 Jan 2002 (TG) removed pcList(), dpattern(), select(), getComplement()
//
// ********************************************************************

#include <iosfwd>
#include "pattern/P_trait.hh"

class ToppIf;
class TrackIf;
class VertexIf;
class CutIf;
class MC_Track;
template <unsigned int NTR> class SDecayPattern;
template <unsigned int NTR> class SDecayPatternMatch;
template <unsigned int NTR> class SPatternExpression;
template <class T> class SmartPointer;
template <unsigned int NTR> class SVertex;

/** Class which represents a container for DecayPattern matches.

    @memo Class for reconstruction of decays.
    @author T. Glebe
*/
//==============================================================================
// SDecayPatternMatchContainer
//==============================================================================
template <unsigned int NTR>
class SDecayPatternMatchContainer {

  typedef typename DP_trait<NTR>::SDPM_l         SDPM_l;
  typedef typename DP_trait<NTR>::SDPM_it        SDPM_it;
  typedef typename DP_trait<NTR>::SDPM_cit       SDPM_cit;
  typedef typename DP_trait<NTR>::SDPMP_l        SDPMP_l;
  typedef typename DP_trait<NTR>::SDPMP_cit      SDPMP_cit;
  typedef SDecayPattern<NTR>                     SDP;

public:
  /** @name --- Constructors --- */
  ///
  SDecayPatternMatchContainer();
  ///
  SDecayPatternMatchContainer(const ToppIf& topp,
			      const SPatternExpression<NTR>& rhs);
  ///
  virtual ~SDecayPatternMatchContainer();

  /** @name --- Access methods --- */
  /// get pattern pointer
  const SP_sp& pattern() const;
  /// set a Track Property
  void with(const SProperty<TrackIf,bool>& rhs);
  /// set a Vertex Property
  void with(const SProperty<VertexIf,bool>& rhs);
  /// read only access to decay pattern matches
  const typename DP_trait<NTR>::SDPM_l& matches() const;
  /// read/write access to decay pattern matches
  typename DP_trait<NTR>::SDPM_l& matches();
  /// read only access to decay pattern matches (for internal use only)
  const SPMP_l& pmatches() const;
  /// read/write access to decay pattern matches (for internal use only)
  SPMP_l& pmatches();
  /// set mass range cut around expected invariant mass
  void massrange(double mr);
  /// return mass range cut
  double massrange() const;
  /// for internal use only
  const SPMP_l* const * pmList() const;
  /// add a MC decay pattern match
  SmartPointer<SDecayPatternMatch<NTR> > push_back(const MC_Track& mtra);
  /// add a reconstructed decay pattern match
  SmartPointer<SDecayPatternMatch<NTR> > push_back(const TrackIf& track);
  /// add a SVertex decay pattern match
  SmartPointer<SDecayPatternMatch<NTR> > push_back(SVertex<NTR>* vtx);
  /// add a decay pattern match
  SmartPointer<SDecayPatternMatch<NTR> > push_back(const SDecayPatternMatch<NTR>& dpm);
  /// add a decay pattern match
  SmartPointer<SDecayPatternMatch<NTR> > push_back(const SmartPointer<SDecayPatternMatch<NTR> >& dpm);

  /** @name --- Expert methods --- */
  /** Do combinatorics and fill decay pattern match container. Compute vertex
      via Kalman filter and check vertex properties. Accept only combinations whose
      inv. mass lies within mass window (to be set via \texttt{massrange()}). If
      argument is \texttt{true}, then it is explicitly checked whether in the
      decay chain a track appears more than once. If yes, the combination is skipped.
      In many cases this check is not needed because of the capabilities of the
      combiner code. See the \texttt{Pattern} manual.
  */
  bool combine(bool check = false);
  /** Do combinatorics and fill decay pattern match container. Do not compute a
      vertex. In case a track or vertex property was defined, compute a vertex using
      a topological method (no Kalman filter!), in this case via the
      \texttt{SVertex::calcVertex2D()} method. Just accept combinations whose
      inv. mass lies within mass window (to be set via \texttt{massrange()}). If
      argument is \texttt{true}, then it is explicitly checked whether in the
      decay chain a track appears more than once. If yes, the combination is skipped.
      In many cases this check is not needed because of the capabilities of the
      combiner code. See the \texttt{Pattern} manual.
  */
  bool combine_no_vertex(bool check = false);
  /** Select decay pattern from MC truth (MC\_Track table) and fill tracks
      into container. */
  bool selectMCtruth();
  /** Select decay pattern from MC truth (MC\_Track table) which have been
      reconstructed and fill (reconstructed) tracks into container. */
  bool selectRECOfromMCtruth();
  /** Select decay pattern from MC truth (MC\_Track table) which could not be
      fully reconstructed and fill tracks into container. In case of
      reconstructed tracks, fill RecoTrack, in case of not reconstructed
      tracks, fill MC\_Track. */
  bool selectNORECOfromMCtruth();
  /// set a cut object
  void set_cut(const CutIf& c);
  /// if DecayPatternMatches share same track, flag DPM with smaller prob as invalid
  bool prob_collision();
  /// if DecayPatternMatches share same track, flag DPM with smaller massProb as invalid
  bool massProb_collision(double scale = 0.1);
  /// massProb\_collision function for background studies
  bool massProb_bg_collision();
  /// calculate mother tracks with covariance matrix, ignore invalid pattern matches
  bool calc_mother();
  /// calculate mother tracks, ignore invalid pattern matches
  bool calc_mother_tr();
  /// calculate cov. matrix of mother tracks, ignore invalid pattern matches
  bool calc_mother_cov();

  /** @name --- Expert methods --- */
  /// used in operator<<()
  std::ostream& print(std::ostream& os) const;

protected:
  /** For internal use only. Checks whether a track appears more than once in a
      combination. If yes, the track combination is skipped.
  */
  bool check_combination(const GSC_it& it) const;
  /// routine which does the work
  virtual bool do_combine(bool check);
  /// routine which does the work
  virtual bool do_combine_no_vertex(bool check);

  SDPM_l        dpmlist;     // output: decay pattern matches
  SPMP_l        ptrlist;     // list of pattern match pointers (=mother particles)
  const SPMP_l* pmlist[NTR]; // input: pattern match pointer lists
  SDP*          dpattern_;   // the decay pattern
  SP_sp         pattern_;    // decay pattern type converted to SPattern
  GSC           Comb;        // Combiner object
};

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<( std::ostream& os,
				 const SDecayPatternMatchContainer<NTR>& a ) {
  return a.print( os );
}

//==============================================================================
// operator <<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <unsigned int NTR>
inline std::ostream& operator<<( std::ostream& os, const
				 SmartPointer<SDecayPatternMatchContainer<NTR> >& a ) {
  return a.null() ? os << "(nil)" : a->print( os );
}

#include "pattern/SDecayPatternMatchContainer.icc"
#endif
