#ifndef SPROPERTY_PATTERN_HH
#define SPROPERTY_PATTERN_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   27. Jun 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A persistent pattern expression engine for TrackIf member functions.
//
// changes:
// 27 Jun 2001 (TG) creation from ClProperty classes
// 29 Jun 2001 (TG) SPMemberOnePar, SPMemberTwoPar: added operator() for
//                  parameter modifications
// 04 Jul 2001 (TG) added template parameter C
//
// **********************************************************************

#include <iosfwd>
#include <iostream>
#include <string>

class TrackIf;
class VertexIf;
template <class C, class T, class X> class SPExpression;
template <class C, class T> class SPConstant;

/** Base class of the SProperty expression engine. All operators which are
    allowed in expressions are defined here.
    @memo for internal use only
    @author T. Glebe
*/
//==============================================================================
// SProperty
//==============================================================================
template <class C, class T>
class SProperty {
public:
  /** @name --- Constructors --- */
  ///
  SProperty() {}
  ///
  virtual ~SProperty() {}
  
  /** @name --- Operators --- */
  /// evaluate pattern
  virtual T evaluate( const C& ) const;

  /// less than - operator< has problems in CINT
  SPExpression<C,T,bool> lt( const SProperty<C,T>& rhs ) const;
  /// less than - operator< has problems in CINT
  SPExpression<C,T,bool> lt( const T& rhs ) const;

  ///
  SPExpression<C,T,bool> operator<( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,bool> operator<( const T& rhs ) const;
  
  ///
  SPExpression<C,T,bool> operator>( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,bool> operator>( const T& rhs ) const;
  
  ///
  SPExpression<C,T,bool> operator==( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,bool> operator==( const T& rhs ) const;
  
  ///
  SPExpression<C,T,bool> operator!=( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,bool> operator!=( const T& rhs ) const;
  
  ///
  SPExpression<C,T,bool> operator&( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,bool> operator|( const SProperty<C,T>& rhs ) const;

  ///
  SPExpression<C,T,T> operator+( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,T> operator+( const T& rhs ) const;
  
  ///
  SPExpression<C,T,T> operator-( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,T> operator-( const T& rhs ) const;
  
  ///
  SPExpression<C,T,T> operator*( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,T> operator*( const T& rhs ) const;
  
  ///
  SPExpression<C,T,T> operator/( const SProperty<C,T>& rhs ) const;
  ///
  SPExpression<C,T,T> operator/( const T& rhs ) const;
  
  /** @name --- Expert functions --- */
  /// copy object to heap, return pointer to it
  virtual SProperty<C,T>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& os ) const;
  ///
  virtual string getString() const;

protected:
  ///
  SProperty( const SProperty<C,T>& rhs ) {}
  ///
  const SProperty<C,T>& operator=( const SProperty<C,T>& rhs );
}; // end of class SProperty


//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SProperty<TrackIf,T>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SProperty<VertexIf,T>& p )
{ return p.print(os); }


/** Binary expression for the SProperty expression engine.
    @memo for internal use only
    @author T. Glebe
*/
//==============================================================================
// SPExpression
//==============================================================================
template <class C, class T, class X>
class SPExpression: public SProperty<C,X> {
public:
  /** @name --- Constructors --- */
  ///
  SPExpression();
  ///
  SPExpression( const SProperty<C,T>& lhs,
		const SProperty<C,T>& rhs,
		X (*func)( const T&, const T& ),
		const char* nam);
  ///
  SPExpression( const SPExpression<C,T,X>& rhs );
  ///
  ~SPExpression();
  ///
  const SPExpression<C,T,X>& operator=( const SPExpression<C,T,X>& );
  
  /** @name --- Expert functions --- */
  /// evaluate pattern
  X evaluate( const C& obj ) const;
  /// copy object to heap, return pointer to it
  virtual SProperty<C,X>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& os ) const;
  ///
  virtual string getString() const;
  
private:
  SProperty<C,T>* lhs;
  SProperty<C,T>* rhs;
  X (*func)( const T&, const T& );

  const char *name; // name of expression func
}; // SPExpression


//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<TrackIf,T,int>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<TrackIf,T,float>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<TrackIf,T,double>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<VertexIf,T,int>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<VertexIf,T,float>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPExpression<VertexIf,T,double>& p )
{ return p.print(os); }


/** Leaf class for the SProperty expression engine.
    @memo for member functions with no argument
    @author T. Glebe
*/
//==============================================================================
// SPMember
//==============================================================================
template <class C, class T>
class SPMember: public SProperty<C,T> {
public:
  /** @name --- Constructors --- */
  ///
  SPMember() {}
  ///
  SPMember( T (C::*f)() const, const char* nam );
  ///
  ~SPMember();
      
  /** @name --- Expert functions --- */
  /// evaluate pattern
  T evaluate( const C& part ) const;
  /// copy object to heap, return pointer to it
  virtual SProperty<C,T>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const;
  ///
  virtual string getString() const;

private:
  T (C::*func)() const;
  const char* name;       // name of member function func
}; // SPMember

//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMember<TrackIf,T>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMember<VertexIf,T>& p )
{ return p.print(os); }


/** Leaf class for the SProperty expression engine.
    @memo for member functions with one argument
    @author T. Glebe
*/
//==============================================================================
// SPMemberOnePar
//==============================================================================
template <class C, class T, class P>
class SPMemberOnePar: public SProperty<C,T> {
public:
  /** @name --- Constructors ---  */
  ///
  SPMemberOnePar() {}
  ///
  SPMemberOnePar( T (C::*f)( P ) const, P par, const char* nam );
  ///
  ~SPMemberOnePar();
      
  /** @name --- Expert functions --- */
  /// evaluate pattern
  virtual T evaluate( const C& part ) const;
  /// set a new parameter
  const SPMemberOnePar<C,T,P>& operator()(const P& val) const;
  /// copy object to heap, return pointer to it
  virtual SProperty<C,T>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const;
  /// 
  virtual string getString() const;
  
private:
  T (C::*func)( P par ) const;
  mutable P par;
  const char* name;       // name of member function func
}; // SPMemberOnePar


//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<TrackIf,T,int>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<TrackIf,T,float>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<TrackIf,T,double>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<VertexIf,T,int>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<VertexIf,T,float>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberOnePar<VertexIf,T,double>& p )
{ return p.print(os); }


/** Leaf class for the SProperty expression engine.
    @memo for member functions with two arguments
    @author T. Glebe
*/
//==============================================================================
// SPMemberTwoPar
//==============================================================================
template <class C, class T, class P1, class P2>
class SPMemberTwoPar: public SProperty<C,T> {
public:
  /** @name --- Constructors --- */
  ///
  SPMemberTwoPar() {}
  ///
  SPMemberTwoPar( T (C::*f)( P1, P2 ) const, P1 par1, P2 par2,
		  const char* nam );
  ///
  ~SPMemberTwoPar();
      
  /** @name --- Expert functions --- */
  /// evaluate pattern
  T evaluate( const C& part ) const;
  /// set new parameters
  const SPMemberTwoPar<C,T,P1,P2>& operator()(const P1& val1, const P2& val2) const;
  /// copy object to heap, return pointer to it
  virtual SProperty<C,T>* propertyClone() const;
  
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const;
  ///
  virtual string getString() const;
  
private:
  T (C::*func)( P1 par1, P2 par2 ) const;
  mutable P1 par1;
  mutable P2 par2;
  const char* name;       // name of member function func
}; // SPMemberTwoPar


//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberTwoPar<TrackIf,T,int,int>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPMemberTwoPar<VertexIf,T,int,int>& p )
{ return p.print(os); }


/** Leaf class for the SProperty expression engine.
    @memo for constant values
    @author T. Glebe
*/
//==============================================================================
// SPConstant
//==============================================================================
template <class C, class T>
class SPConstant: public SProperty<C,T> {
public:
  /** @name --- Constructors --- */
  ///
  SPConstant() {}
  ///
  SPConstant( T constant );
  ///
  ~SPConstant();
  
  /** @name --- Expert functions --- */
  /// evaluate pattern
  T evaluate( const C& part ) const;
  /// copy object to heap, return pointer to it
  virtual SProperty<C,T>* propertyClone() const;
  /// used by operator<<()
  virtual std::ostream& print( std::ostream& ) const;
  ///
  virtual string getString() const;
      
private:
  T constant;
}; // class SPConstant
  

//==============================================================================
// operator<<: the second template argument is a problem, therefore hardcoded
//==============================================================================
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPConstant<TrackIf,T>& p )
{ return p.print(os); }
template <class T>
inline std::ostream& operator<<(std::ostream& os, const SPConstant<VertexIf,T>& p )
{ return p.print(os); }

// Load the implementation file
#include "SProperty.icc"
#endif // SPROPERTY_PATTERN_HH
