#ifndef SMART_POINTER_HH
#define SMART_POINTER_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   22. Jul 1997
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: A Pointer which does reference counting.
//
// changes:
// 22 Jul 1997 created by C. Borgmeier
// 31 Jan 2001 (TG) renamed Wrapper to SmartData
// 31 Jan 2001 (TG) made SmartData private struct of SmartPointer
// 01 Feb 2001 (TG) operator-> added
// 07 Feb 2001 (TG) added operator==, operator!=, null()
// 07 Feb 2001 (TG) declared all operators inline
// 27 Aug 2001 (TG) CTOR added
// 08 Jan 2002 (TG) moved all inline functions to .icc file
//
// ********************************************************************
#include <iosfwd>

template <class T> class SmartPointer;
  
/** Reference counting pointer.
    Note: The object the Pointer is pointing to must be on the heap!
    In case of a stack object, give the object as a CTOR argument,
    then SmartPointer will generate a copy of it on the heap.
    
    Example:
    \begin{enumerate}
    \item heap object:\\
        T* obj = new T;\\
        SmartPointer<T> ptr(obj); // ok
    
    \item stack object:\\
        T obj;\\
        SmartPointer<T> ptr(obj); // ok, uses copy of obj. on heap
    \item default object on heap:\\
        SmartPointer<T> ptr(true); // object generated by default CTOR
    \item empty SmartPointer:\\
        SmartPointer<T> ptr(); // ptr points to NULL
    \end{enumerate}
    @author C. Borgmeier, T. Glebe
*/
//==============================================================================
// SmartPointer used in Pattern
//==============================================================================
template <class T>
class SmartPointer {
public:
  /** @name --- Constructors --- */
  /// usedef == true => create default T object on heap
  SmartPointer( bool usedef = false );
  /// takes object ownership
  SmartPointer( T* obj );
  /// takes object ownership
  SmartPointer( const T* obj );
  /// copies object
  SmartPointer( const T& obj );
  /// copies pointer
  SmartPointer( const SmartPointer<T>& arg );
  /// object is deleted together with the last pointer to it
  ~SmartPointer();
  
  /** @name --- Operators --- */
  /// assignment operator for reference counting
  SmartPointer<T>& operator=( const SmartPointer<T>& rhs );
  /// r/w dereference
  T& operator*();
  /// readonly dereference
  const T& operator*() const;
  /// r/w access
  T* operator->();
  /// readonly access
  const T* operator->() const;
  
  /// compare the objects, not the pointers!
  bool operator<( const SmartPointer<T>& arg ) const;
  /// compare Smart Pointers
  bool operator==(const SmartPointer<T>& rhs) const;
  /// compare Smart Pointers
  bool operator!=(const SmartPointer<T>& rhs) const;

  /** @name --- Expert methods --- */
  /// used by operator<<()
  std::ostream& print( std::ostream& os ) const;

  /// check if pointer points to an object
  bool null() const;

private:
  // bookkeeping structure
  struct SmartData {
    //
    SmartData();
    //
    SmartData( const T& obj);
    //
    SmartData( T* obj);
    //
    SmartData( const T* obj);
    //
    ~SmartData();
    
    T* object;
    unsigned int count;
  }; // end of struct SmartData
  
  SmartData<T>* pointer; // pointer to object!
}; // end of class SmartPointer


//==============================================================================
// operator <<
//==============================================================================
template <class T>
inline std::ostream& operator<<( std::ostream& os, const SmartPointer<T>& p ) {
  return p.print(os);
}

#ifndef __CINT__  
#include "SmartPointer.icc"
#endif
#endif
