#ifndef __GDECAYPATTERN_HH
#define __GDECAYPATTERN_HH
// ********************************************************************
//
// source:
//
// type:      source code
//
// created:   31. Jan 2001
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Class describing a decay pattern.
//              This class is based on the GDecayPatternBase class
//              and is an adoption to a real Vertexing class.
//
// changes:
// 05 Feb 2001 (TG) creation
// 22 Feb 2001 (TG) added collptr and set_collision()
// 23 Feb 2001 (TG) dm and massrange() members added. dm is the range
//                  around the Topp table mass in which vertexing should
//                  be done (see select function)
// 05 Mar 2001 (TG) getString() added
// 08 Mar 2001 (TG) moved dm and massrange() to DecayPatternBase
// 08 Mar 2001 (TG) push_back(), me(), checkMass(), VertexCheck() added
//
// ********************************************************************
#include <iosfwd>
#include "pattern/GDecayPatternBase.hh"

template <class C, class T> class ClProperty;
namespace VERTEX { class Vertex; }
typedef ClProperty<VERTEX::Vertex, int>   VtProperty;
template <class C> class CollisionIf;

/** A class for defining object selection pattern for particle decays.
    GDecayPattern handles the cuts for the fitted vertex together with the
    Vertex object. GDecayPattern is specialized to use Vt++ vertices. In order
    to integrate vertex objects from other vertex codes, one has to define a
    new GDecayPattern object.

    @memo Generic class
    @author T. Glebe
 */
//==============================================================================
// GDecayPattern
//==============================================================================
template <class C, class T>
class GDecayPattern : public GDecayPatternBase<C,T> {
  typedef SmartPointer<GDecayPattern<C,T> >  GDPatPtr;
  typedef list<GDPatPtr>                   GDPat_l;
  typedef list<GDPatPtr>::iterator         GDPat_it;
  typedef list<GDPatPtr>::const_iterator   GDPat_cit;

public:
  /** @name --- Constructors --- */
  ///
  GDecayPattern();
  ///
  GDecayPattern( const GDecayPattern<C,T>& rhs);
  ///
  GDecayPattern( const ToppIf& tp, const GDecayPattern<C,T>& rhs);
  ///
  GDecayPattern( const ToppIf& tp, 
		 const GPattern<C,T>& lhs, const GPattern<C,T>& rhs);
  ///
  virtual ~GDecayPattern();
  ///
  GDecayPattern<C,T>& operator=( const GDecayPattern<C,T>& rhs );

  /** @name --- Operators --- */
  ///
//    GDecayPattern<C,T>& operator&( const GPattern<C,T>& rhs);
//    ///  put everything into a single vertex (to be improved)
//    GDecayPattern<C,T>& operator&( const GDecayPattern<C,T>& rhs);
  ///
//    GDecayPattern<C,T>& operator|( const GDecayPattern<C,T>& rhs) {
//      // rhs must be kept separate.
//      return *this;
//    }

  /** @name --- User functons --- */
  /** InputContainer: Container which holds pointers to track objects.
      OutputContainer: Container which holds pointers to pattern matches.
      OutVal: exact pattern match type. Cannot be determined via
              OutputContainer::value\_type as this returns the pointer type.
      Select all tracks from InputContainer which passes property 
      criteria and put them as PatternMatches in OutputContainer.
      In case OR\_CC is set for charge conjugation handling, also the charge
      conjugated DecayPattern is processed.
  */
  template <class InputContainer, class OutputContainer, class OutVal>
  bool select(const InputContainer&  src, OutputContainer& trg,
	      const OutVal& ov);
  /// Make selection ignoring the charge conjugation flag
  template <class InputContainer, class OutputContainer, class OutVal>
  bool DoSelection(const InputContainer&  src, OutputContainer& trg,
	      const OutVal& ov) const;
  /// assign a primary vertex to each PatternMatch
  template <class InputContainer, class OutputContainer>
  bool attachPrimary(const InputContainer& src, OutputContainer& trg) const;

  /// ClProperty for Vertex cuts
  void with(const VtProperty& rhs);
  /// set a collision class
  void set_collision(const CollisionIf<C>& rhs);
  /// get the vertex cut pattern string
  string getVertexString() const;
  /// get the pattern string
  string getString() const;

  /** @name --- Expert functions --- */
  ///
  virtual std::ostream& print(std::ostream& os) const;
  ///
  void massCheck(const VERTEX::Vertex& rhs,
		 bool& accept, bool& CCaccept) const;
  ///
  bool VertexCheck(const VERTEX::Vertex& vtx) const;

  /// return SmartPointer which points to me
  const GDPatPtr& me() const;
  /// check if pattern is compatible with rhs pattern
  bool compatible( const GDecayPattern<C,T>& rhs ) const;
  /// Any: add DecayPattern to vertex with
  void push_back( const GDecayPattern<C,T>& rhs);
  /// merge two pattern
  void merge(const GDecayPattern<C,T>& rhs);

protected:
  VtProperty*       vtxprop; // cuts for Vertices
  CollisionIf<C>*   collptr; // object collision class
#ifndef __CINT__
  mutable GDPatPtr  me_;
#endif
  GDPat_l           decays;  // list of possible decays
}; // end of GDecayPattern

#ifndef __CINT__
#include "pattern/GDecayPattern.icc"
#endif
#endif
