#ifndef PROPERTY_PATTERN_HH
#define PROPERTY_PATTERN_HH
// **********************************************************************
//
// source:
//
// type:      source code
//
// created:   16. Sep 1998
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: The pattern expression engine.
//
// changes:
// 16 Sep 1998 creation by C. Borgmeier
// 03 Nov 2000 (TG) code polishing
// 14 Nov 2000 (TG) inlined some members of ClProperty
// 20 Nov 2000 (TG) added namespace CLUE
// 24 Nov 2000 (TG) added lt() member function as operator< has problems in CINT
// 24 Jan 2001 (TG) made operators inline, removed select() functions
// 02 Mar 2001 (TG) getString() added
//
// **********************************************************************

#include <iosfwd>
#include <iostream>
#include <string>
#include <strstream>
#include "arte/TObject.hh"
#include "gate/RtypesT.h"
#include "gate/gatefwd.hh"
#include "clue/cluefwd.hh"

#ifdef NAMESPACE
namespace CLUE {
#endif

  template <class T> int lessfunc    (const T& lhs, const T& rhs);
  template <class T> int morefunc    (const T& lhs, const T& rhs);
  template <class T> int equalsfunc  (const T& lhs, const T& rhs);
  template <class T> int notequalfunc(const T& lhs, const T& rhs);
  template <class T> T   plusfunc    (const T& lhs, const T& rhs);
  template <class T> T   minusfunc   (const T& lhs, const T& rhs);
  template <class T> T   timesfunc   (const T& lhs, const T& rhs);
  template <class T> T   dividefunc  (const T& lhs, const T& rhs);
  template <class T> int andfunc     (const T& lhs, const T& rhs);
  template <class T> int orfunc      (const T& lhs, const T& rhs);


  /** Expression template base class.
      @memo class ClProperty
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClProperty
  //==============================================================================
  template <class C, class T>
    class ClProperty: public TObject {
    public:
      /** @name --- Constructors --- */
      ///
      ClProperty() {}
      ///
      virtual ~ClProperty() {}

      /** @name --- Operators --- */
      // This operator tests the property
      // virtual T operator()( const C& ) const;
      /// evaluate pattern
      virtual T evaluate( const C& ) const;

      // This operator tests the property
      //      virtual T operator()( const TrackBase& ) const;
      
      /// less than - operator< has problems in CINT
      inline ClExpression<C,T,int> lt( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &lessfunc, " < " ); }
      /// less than - operator< has problems in CINT
      inline ClExpression<C,T,int> lt( const T& rhs ) const
	{ return operator<( ClConstant<C,T>( rhs ) ); }

      ///
      inline ClExpression<C,T,int> operator<( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &lessfunc, " < " ); }
      ///
      inline ClExpression<C,T,int> operator<( const T& rhs ) const
	{ return operator<( ClConstant<C,T>( rhs ) ); }

      ///
      inline ClExpression<C,T,int> operator>( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &morefunc, " > " ); }
      ///
      inline ClExpression<C,T,int> operator>( const T& rhs ) const
	{ return operator>( ClConstant<C,T>( rhs ) ); }

      ///
      inline ClExpression<C,T,int> operator==( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &equalsfunc, " == " ); }
      ///
      inline ClExpression<C,T,int> operator==( const T& rhs ) const
	{ return *this == ClConstant<C,T>( rhs ); }
      
      ///
      inline ClExpression<C,T,int> operator!=( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &notequalfunc, " != " ); }
      ///
      inline ClExpression<C,T,int> operator!=( const T& rhs ) const
	{ return operator!=( ClConstant<C,T>( rhs ) ); }

      ///
      inline ClExpression<C,T,T> operator+( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,T>( *this, rhs, &plusfunc, " + " ); }
      ///
      inline ClExpression<C,T,T> operator+( const T& rhs ) const
	{ return *this + ClConstant<C,T>( rhs ); }
      
      ///
      inline ClExpression<C,T,T> operator-( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,T>( *this, rhs, &minusfunc, " - " ); }
      ///
      inline ClExpression<C,T,T> operator-( const T& rhs ) const
	{ return *this - ClConstant<C,T>( rhs ); }
      
      ///
      inline ClExpression<C,T,T> operator*( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,T>( *this, rhs, &timesfunc, " * " ); }
      ///
      inline ClExpression<C,T,T> operator*( const T& rhs ) const
	{ return *this * ClConstant<C,T>( rhs ); }
      
      ///
      inline ClExpression<C,T,T> operator/( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,T>( *this, rhs, &dividefunc, " / " ); }
      ///
      inline ClExpression<C,T,T> operator/( const T& rhs ) const 
	{ return *this / ClConstant<C,T>( rhs ); }

      ///
      inline ClExpression<C,T,int> operator&( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &andfunc, " & " ); }
      ///
      inline ClExpression<C,T,int> operator|( const ClProperty<C,T>& rhs ) const
	{ return ClExpression<C,T,int>( *this, rhs, &orfunc, " | " ); }
      
      
      /** @name --- Expert functions --- */
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      //      virtual ClProperty<C,T>* propertyClone() const;
      /// used by operator<<()
      virtual std::ostream& print( std::ostream& os ) const;
      ///
      virtual string getString() const {
	return string("no cuts");
      }

    protected:
      ///
      ClProperty( const ClProperty<C,T>& rhs ) {}
      ///
      const ClProperty<C,T>& operator=( const ClProperty<C,T>& );
      
#ifdef USE_ROOT
      ClMyDefT2(ClProperty<C,T>,1);
#endif
    }; // end of class ClProperty


  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClProperty<T,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClProperty<T,float>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClProperty<T,double>& p )
    { return p.print(os); }


  /** Expression template.
      @memo class ClExpression
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClExpression
  //==============================================================================
  template <class C, class T, class X>
    class ClExpression: public ClProperty<C,X> {
    public:
      /** @name --- Constructors --- */
      ///
      ClExpression();
      ///
      ClExpression( const ClProperty<C,T>& lhs,
		    const ClProperty<C,T>& rhs,
		    X (*func)( const T&, const T& ),
		    const char* nam);
      ///
      ClExpression( const ClExpression<C,T,X>& rhs );
      ///
      ~ClExpression();
      ///
      const ClExpression<C,T,X>& operator=( const ClExpression<C,T,X>& );

      //
      // X operator()( const C& obj ) const;
      // This operator tests the property
      //      X operator()( const TrackBase& ) const;

      /** @name --- Expert functions --- */
      /// evaluate pattern
      X evaluate( const C& obj ) const;
      ///
      virtual ClProperty<C,X>* propertyClone() const;
      //      virtual ClProperty<C,X>* propertyClone() const;
      /// used by operator<<()
      virtual std::ostream& print( std::ostream& os ) const;
      ///
      virtual string getString() const {
	return string(lhs->getString()) + string(name) + string(rhs->getString());
      }
      
    private:
      typedef ClProperty<C,T> P_type; // to help CINT parsing

      P_type* lhs;
      P_type* rhs;
      X (*func)( const T&, const T& );

      const char *name; // name of expression func

#ifdef USE_ROOT
      ClMyDefT3(ClExpression<C,T,X>, 1);
#endif
  }; // ClExpression


  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClExpression<T,int,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClExpression<T,float,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClExpression<T,float,float>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClExpression<T,double,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClExpression<T,double,double>& p )
    { return p.print(os); }


  /** class ClMember
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClMember
  //==============================================================================
  template <class C, class T>
    class ClMember: public ClProperty<C,T> {
    public:
      /** @name --- Constructors --- */
      ///
      ClMember() {}
#ifndef __CINT__
      ///
      ClMember( T (C::*f)() const, const char* nam );
#endif
      ///
      ~ClMember();
      
      /** @name --- Expert functions --- */
      /// evaluate pattern
      T evaluate( const C& part ) const;
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      //      virtual ClProperty<C,T>* propertyClone() const;
      /// used by operator<<()
      virtual std::ostream& print( std::ostream& ) const;
      ///
      virtual string getString() const {
	return string(name);
      }

    private:
      T (C::*func)() const;
      const char* name;       // name of member function func
#ifdef USE_ROOT
      ClMyDefT2(ClMember<C,T>,1);
#endif
  }; // ClMember

  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMember<T,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMember<T,unsigned int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMember<T,float>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMember<T,double>& p )
    { return p.print(os); }

  /** Unary expression template.
      @memo class ClMemberOnePar
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClMemberOnePar
  //==============================================================================
  template <class C, class T, class P>
    class ClMemberOnePar: public ClProperty<C,T> {
    public:
      /** @name --- Constructors ---  */
      ///
      ClMemberOnePar() {}
#ifndef __CINT__
      ///
      ClMemberOnePar( T (C::*f)( P ) const, P par );
#endif
      ///
      ~ClMemberOnePar();
      
      /** @name --- Expert functions --- */
      /// evaluate pattern
      T evaluate( const C& part ) const;
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      //      virtual ClProperty<C,T>* propertyClone() const;
      /// used by operator<<()
      virtual std::ostream& print( std::ostream& ) const;
      /// 
      virtual string getString() const {
	ostrstream puffer;
	puffer << "(" << par1 << ")";
	return string(name) + string(puffer.str());
      }

    private:
      T (C::*func)( P par ) const;
      P par;
      
#ifdef USE_ROOT
      ClMyDefT3(ClMemberOnePar<C,T,P>,1);
#endif
  }; // ClMemberOnePar


  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberOnePar<T,int,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberOnePar<T,float,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberOnePar<T,float,float>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberOnePar<T,double,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberOnePar<T,double,double>& p )
    { return p.print(os); }


  /** Binary expression template.
      @memo ClMemberTwoPar
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClMemberTwoPar
  //==============================================================================
  template <class C, class T, class P1, class P2>
    class ClMemberTwoPar: public ClProperty<C,T> {
    public:
      /** @name --- Constructors --- */
      ///
      ClMemberTwoPar() {}
#ifndef __CINT__
      ///
      ClMemberTwoPar( T (C::*f)( P1, P2 ) const, P1 par1, P2 par2 );
#endif
      ///
      ~ClMemberTwoPar();
      
      /** @name --- Expert functions --- */
      /// evaluate pattern
      T evaluate( const C& part ) const;
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      //      virtual ClProperty<C,T>* propertyClone() const;

      /// used by operator<<()
      virtual std::ostream& print( std::ostream& ) const;
      ///
      virtual string getString() const {
	ostrstream puffer;
	puffer << "(" << par1 << "," << par2 << ")";
	return string(name) + string(puffer.str());
      }

    private:
      T (C::*func)( P1 par1, P2 par2 ) const;
      P1 par1;
      P2 par2;
      
#ifdef USE_ROOT
      ClMyDefT4(ClMemberTwoPar<C,T,P1,P2>,1);
#endif
  }; // ClMemberTwoPar


  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberTwoPar<T,float,int,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClMemberTwoPar<T,double,int,int>& p )
    { return p.print(os); }


  /** Constant expression template.
      @memo ClConstant
      @author C. Borgmeier, T. Glebe
  */
  //==============================================================================
  // ClConstant
  //==============================================================================
  template <class C, class T>
    class ClConstant: public ClProperty<C,T> {
    public:
      /** @name --- Constructors --- */
      ///
      ClConstant() {}
      ///
      ClConstant( T constant );
      ///
      ~ClConstant();
      
      /** @name --- Expert functions --- */
      /// evaluate pattern
      T evaluate( const C& part ) const;
      ///
      virtual ClProperty<C,T>* propertyClone() const;
      //      virtual ClProperty<C,T>* propertyClone() const;
      /// used by operator<<()
      virtual std::ostream& print( std::ostream& ) const;
      ///
      virtual string getString() const {
	ostrstream puffer;
	puffer << constant;
	return string(puffer.str());
      }
      
    private:
      T constant;
#ifdef USE_ROOT
      ClMyDefT2(ClConstant<C,T>,1);
#endif
    }; // class ClConstant
  

  //==============================================================================
  // operator<<: the second template argument is a problem, therefore hardcoded
  //==============================================================================
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClConstant<T,int>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClConstant<T,float>& p )
    { return p.print(os); }
  template <class T>
    inline std::ostream& operator<<(std::ostream& os, const ClConstant<T,double>& p )
    { return p.print(os); }

#ifdef NAMESPACE
} // namespace CLUE

#ifdef USE_ROOT
MyClMyInstT2(CLUE,ClProperty,C,T);
MyClMyInstT3(CLUE,ClExpression,C,T,X);
MyClMyInstT2(CLUE,ClMember,C,T);
MyClMyInstT3(CLUE,ClMemberOnePar,C,T,P);
MyClMyInstT4(CLUE,ClMemberTwoPar,C,T,P1,P2);
MyClMyInstT2(CLUE,ClConstant,C,T);
#endif
#else  // NAMESPACE
#ifdef USE_ROOT
ClMyInstT2(ClProperty,C,T);
ClMyInstT3(ClExpression,C,T,X);
ClMyInstT2(ClMember,C,T);
ClMyInstT3(ClMemberOnePar,C,T,P);
ClMyInstT4(ClMemberTwoPar,C,T,P1,P2);
ClMyInstT2(ClConstant,C,T);
#endif
#endif // NAMESPACE

// Load the implementation file
#ifndef __CINT__
#include "pattern/Property.icc"
#endif
#endif // PROPERTY_PATTERN_HH
