// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: HERA-B track covariance matrix class
//              Attention: Consturctor makes transformation
//              sigma(Q/p) -> sigma(p) (a Vt-requirement).
//
// changes:
// 21 Aug 2000 (TG) creation
// 06 Sep 2000 (TG) include VtUtil.hh added
// 06 Sep 2000 (TG) transform sigma(Q/p) into sigma(p) already in constructor
// 26 Okt 2000 (TG) #ifndef NO_ARTE added (for integration into CLUE)
// 14 Nov 2000 (TG) added constructor code for CLUE (RecoSegment, ArtePointer<RecoTrack>)
// 20 Nov 2000 (TG) added using namespace CLUE
// 24 Nov 2000 (TG) removed using namespace CLUE
// 24 Jan 2001 (TG) removed obsolete include directives
// 04 Sep 2001 (TG) SMatrix CTOR added
// 13 Sep 2001 (TG) RecoSegment/RecoTrack CTOR: changed cov-matrix calls
// 17 Okt 2001 (TG) added/improved #defines for ARTE
//
// *****************************************************************************

#include <iostream>
#include <assert.h>
#include "vt++/CMatrix.hh"
#include "vt++/VtUtil.hh"

#if defined NO_ARTE && defined USE_ROOT
#include "clue/cluearte.hh"
#include "clue/RecoTrack.hh"
#endif

using namespace VERTEX;

namespace MATRIX {
  //============================================================================
  // CMatrix Constructors
  //============================================================================
  CMatrix::CMatrix(const CMatrix& rhs):VtSymMatrix(rhs) {}

#if defined NO_ARTE && defined USE_ROOT
  CMatrix::CMatrix(const SMatrix<double,5>& rhs) :
    VtSymMatrix(5) {

    set_x      (rhs(0,0));
    set_x_y    (rhs(1,0));
    set_y      (rhs(1,1));
    set_x_tx   (rhs(0,2));
    set_y_tx   (rhs(1,2));
    set_tx     (rhs(2,2));
    set_x_ty   (rhs(0,3));
    set_y_ty   (rhs(1,3));
    set_tx_ty  (rhs(2,3));
    set_ty     (rhs(3,3));
    set_x_p    (rhs(0,4));
    set_y_p    (rhs(1,4));
    set_tx_p   (rhs(2,4));
    set_ty_p   (rhs(3,4));
    set_p      (rhs(4,4));
  }
#endif

#ifndef NO_ARTE
  CMatrix::CMatrix(const ArtePointer<RTRA>& rtra):VtSymMatrix(5) {
    double p2      = sqr(1./rtra->pf);
    const int sign = sgn(rtra->pf);

    set_x      (rtra->cvf[0]);
    set_x_y    (rtra->cvf[1]);
    set_y      (rtra->cvf[2]);
    set_x_tx   (rtra->cvf[3]);
    set_y_tx   (rtra->cvf[4]);
    set_tx     (rtra->cvf[5]);
    set_x_ty   (rtra->cvf[6]);
    set_y_ty   (rtra->cvf[7]);
    set_tx_ty  (rtra->cvf[8]);
    set_ty     (rtra->cvf[9]);
    set_x_p    (rtra->cvf[10] * -sign * p2);
    set_y_p    (rtra->cvf[11] * -sign * p2);
    set_tx_p   (rtra->cvf[12] * -sign * p2);
    set_ty_p   (rtra->cvf[13] * -sign * p2);
    set_p      (rtra->cvf[14] * p2 * p2);

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || p() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RTRA " << rtra
	   << ": neg. or zero diagonal element!!!" << endl;
    }
  }

  CMatrix::CMatrix(const ArtePointer<RSEG>& rseg):VtSymMatrix(5) {
    double p2      = sqr(1./rseg->pf);
    const int sign = sgn(rseg->pf);

    set_x      (rseg->cvf[0]);
    set_x_y    (rseg->cvf[1]);
    set_y      (rseg->cvf[2]);
    set_x_tx   (rseg->cvf[3]);
    set_y_tx   (rseg->cvf[4]);
    set_tx     (rseg->cvf[5]);
    set_x_ty   (rseg->cvf[6]);
    set_y_ty   (rseg->cvf[7]);
    set_tx_ty  (rseg->cvf[8]);
    set_ty     (rseg->cvf[9]);
    set_x_p    (rseg->cvf[10] * -sign * p2);
    set_y_p    (rseg->cvf[11] * -sign * p2);
    set_tx_p   (rseg->cvf[12] * -sign * p2);
    set_ty_p   (rseg->cvf[13] * -sign * p2);
    set_p      (rseg->cvf[14] * p2 * p2);

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || p() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RSEG " << rseg
	   << ": neg. or zero diagonal element!!!" << endl;
    }

  }

  CMatrix::CMatrix(const ArtePointer<MTRA>& mtra):VtSymMatrix(5) {
    set_x      (1e-9);
    set_x_y    (0.);
    set_y      (1e-9);
    set_x_tx   (0.);
    set_y_tx   (0.);
    set_tx     (1e-9);
    set_x_ty   (0.);
    set_y_ty   (0.);
    set_tx_ty  (0.);
    set_ty     (1e-9);
    set_x_p    (0.);
    set_y_p    (0.);
    set_tx_p   (0.);
    set_ty_p   (0.);
    set_p      (1e-9);
  }
#endif // NO_ARTE

#if defined NO_ARTE && defined USE_ROOT
  CMatrix::CMatrix(const RecoSegment& r):VtSymMatrix(5) {
    set_x      (r.cov_x());
    set_x_y    (r.cov_x_y());
    set_y      (r.cov_y());
    set_x_tx   (r.cov_x_tx());
    set_y_tx   (r.cov_y_tx());
    set_tx     (r.cov_tx());
    set_x_ty   (r.cov_x_ty());
    set_y_ty   (r.cov_y_ty());
    set_tx_ty  (r.cov_tx_ty());
    set_ty     (r.cov_ty());
    set_x_p    (r.cov_x_p());
    set_y_p    (r.cov_y_p());
    set_tx_p   (r.cov_tx_p());
    set_ty_p   (r.cov_ty_p());
    set_p      (r.cov_p());

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || p() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RecoSegment " << r
	   << ": neg. or zero diagonal element!!!" << endl;
    }
  }


  CMatrix::CMatrix(const ArtePointer<RecoTrack>& r):VtSymMatrix(5) {
    set_x      (r->cov_x());
    set_x_y    (r->cov_x_y());
    set_y      (r->cov_y());
    set_x_tx   (r->cov_x_tx());
    set_y_tx   (r->cov_y_tx());
    set_tx     (r->cov_tx());
    set_x_ty   (r->cov_x_ty());
    set_y_ty   (r->cov_y_ty());
    set_tx_ty  (r->cov_tx_ty());
    set_ty     (r->cov_ty());
    set_x_p    (r->cov_x_p());
    set_y_p    (r->cov_y_p());
    set_tx_p   (r->cov_tx_p());
    set_ty_p   (r->cov_ty_p());
    set_p      (r->cov_p());

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || p() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RecoTrack " << r
	   << ": neg. or zero diagonal element!!!" << endl;
    }
  }
#endif

  //============================================================================
  // CMatrix::propagate: propagate Matrix by dz
  //============================================================================
  void CMatrix::propagate(const double dz) {
    set_x      ( x_prop(dz)    );
    set_x_y    ( x_y_prop(dz)  );
    set_x_tx   ( x_tx_prop(dz) );
    set_x_ty   ( x_ty_prop(dz) );
    set_x_p    ( x_p_prop(dz)  );
    set_y      ( y_prop(dz)    );
    set_y_tx   ( y_tx_prop(dz) );
    set_y_ty   ( y_ty_prop(dz) );
    set_y_p    ( y_p_prop(dz)  );

    // tx     is invariant
    // tx_ty  is invariant
    // ty     is invariant
    // tx_p   is invariant
    // ty_p   is invariant
    // p      is invariant
    return;
  }

} // namespace MATRIX
